/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1987, 1988, Patrick Powell
 ***************************************************************************
 * MODULE: banner.c
 ***************************************************************************/

#include "lp.h"
#include "library/errormsg.h"
#include "LPD/banner.h"

/*
 * Print a banner The large character fonts have been stolen from the BANNER program,
 * which stole them from an old card deck...
 * 
 * banner(): print the banner
 * 1. form feed if needed; 2. print banner; 3. another form feed.
 */

static int linecnt;
static int fail;
extern char chartable[][HEIGHT], chartable2[][HEIGHT+3];
static char *bline;
static unsigned int blen;
static int fd;
static char SEQNUMBER[4];

static void shortbanner (void), longbanner (void), Out_line (void),
    breakline (char), bigprint (char *, char *), longbanner_PLP (void),
    longbanner_EUCS (void), longendbanner_EUCS (void), longendbanner_PLP (void);

int
banner (int pr, struct plp_queue *q) {
    /*
     * print the banner
     */

    linecnt = 0;
    fd = pr;
    fail = False;

    if (bline == NULL) {
	blen = BUFSIZ;
	malloc_or_die (bline, blen);
    }
    if (blen < ((unsigned int) PW) + 1) {
	blen = PW + 2;
	if ((bline = (char *) realloc (bline, blen)) == NULL) {
	    logerr_die (XLOG_INFO, Malloc_failed_msg);
	}
    }

    Queue_tmp = q;

    (void) sprintf (SEQNUMBER, "%-3d", Queue_tmp->q_num);

    if (SB) {
	DBGLOCAL4 ("banner: short banner (sequence number is %s)", SEQNUMBER);
	shortbanner ();
    } else {
	DBGLOCAL4 ("banner: long banner (sequence number is %s)", SEQNUMBER);
	longbanner ();
    }

    DBGLOCAL4 ("banner: %s", fail ? "failed" : "succeeded");

    if (fail) {
	return (JFAIL);
    } else {
	return (JSUCC);
    }
}

int 
endbanner (int pr) {
    /*
     * print the banner
     */

    linecnt = 0;
    fd = pr;
    fail = False;

    if (bline == NULL) {
	blen = BUFSIZ;
	malloc_or_die (bline, blen);
    }
    if (blen < ((unsigned int) PW) + 1) {
	blen = PW + 2;
	if ((bline = (char *) realloc (bline, blen)) == NULL) {
	    logerr_die (XLOG_INFO, Malloc_failed_msg);
	}
    }
    
    DBGLOCAL4 ("endbanner: long endbanner");
    if (EUCS_banner_format) {
	longendbanner_EUCS ();
    } else {
	longendbanner_PLP ();
    }
    DBGLOCAL4 ("endbanner: %s", fail ? "failed" : "succeeded");
    
    if (fail) {
	return (JFAIL);
    } else {
	return (JSUCC);
    }
}

/*
 * seqinfo gets the sequence number for the banner page 
 */
static inline void
seqinfo (void) 
{
    (void) sprintf (SEQNUMBER, "%-3d", Queue_tmp->q_num);
}
 

/*
 * shortbanner single line, has either of these formats:
 *
 * 0....+....1....+....2....+....3....+....4....+....5....+....6....+....7....+....
 * (BSD-style)
 * adobe:shore  Job: test.data  Date: Tue Sep 18 16:22:33 1984
 *
 * (PLP-style)
 * adobe:shore  Job: test.data  Seq: 678  Date: Tue Sep 18 16:22:33 1984
 */
static void
shortbanner (void) {
    time_t t;               /* date string */
    char * time_buf;
    char c;

    t = time ((time_t *) NULL);     /* initialize date */
    printed_time = t;
    malloc_or_die (time_buf, strlen (ctime (&t) + 1));
    time_buf = ctime (&t);
    time_buf[24] = '\0';   /* strip the \n character */
    
#ifdef STEF_BANNER
    /* hostname begins with a capitalized letter */
    c = FROMHOST[0];
    if (c > 96) {
        c = c - 32;
	FROMHOST[0] = c;
    }
#endif /* STEF_BANNER */
    seqinfo ();
    
    if (EUCS_banner_format || BK) {
	/* make short banner look like BSD-style (Sun, etc.) short banner.
	 * if "bk" is set in the printcap, use a BSD-style short banner
	 * as well. This may be necessary for use with an OF that parses
	 * the banner (as lprps does).
	 */
	if (strstr (CLASSNAME, MAGIC_STRING) == NULL) {
	    (void) sprintf (bline, "%s:%s  Job: %s  Date: %s",
		CLASSNAME, LOGNAME, JOBNAME, time_buf);
	} else {
	    (void) sprintf (bline, "%s:%s  Job: %s  Date: %s",
		FROMHOST, LOGNAME, JOBNAME, time_buf);
	}

    } else {
	(void) sprintf (bline, "%s:%s  Job: %s  Seq: %s  Date: %s",
	    FROMHOST, LOGNAME, JOBNAME, SEQNUMBER, time_buf);
    }

    Out_line ();
}

/*
 * userinfo: just printf the information
 */
static void
userinfo (void) {
    char *time_buf;             /* temporary time buffer string */
    char s[26];                 /* actual time */
    time_t tvec;
    char c;

    s[0] = '\0';

#ifdef STEF_BANNER
    /* hostname begins with a capitalized letter */
    c = FROMHOST[0];
    if (c > 96) {
        c = c - 32;
	FROMHOST[0] = c;
    }
#endif /* STEF_BANNER */

    (void) sprintf (bline, "User:          %s (%s@%s)", BNRNAME, LOGNAME, FROMHOST);
    Out_line ();
    (void) sprintf (bline, "Machine:       %s", FROMHOST);
    Out_line ();
    (void) sprintf (bline, "Job:           %s", JOBNAME);
    Out_line ();

    /*
     * put the name of the printer on the banner page 
     */
    if (CM && *CM) {
       (void) sprintf (bline, "Output device: %s", CM);
    } else {
       (void) sprintf (bline, "Output device: Printer '%s'", Printer);
    }
    Out_line ();

    /* Get the "spooled" time from the controlfile */
    malloc_or_die (time_buf, strlen (ctime (&Queue_tmp->q_time) + 1));
    time_buf = ctime (&Queue_tmp->q_time);
    time_buf[24] = '\0';   /* strip the \n character */

    (void) sprintf (bline, "Spooled at:    %s", time_buf);
    Out_line ();

    /* Get the actual "printing" time */
    tvec = time ((time_t *) NULL);
    printed_time = tvec;
    (void) strcpy (s, (char *) ctime (&tvec));
    s[24] = '\0';   /* strip the \n character */
    (void) sprintf (bline, "Printed at:    %s", s);
    Out_line ();

    (void) sprintf (bline, "Class:         %s", CLASSNAME);
    Out_line ();
    (void) sprintf (bline, "Options:       %s", ZOPTS);
    Out_line ();
}


static void
longbanner (void) {
    if (EUCS_banner_format) {
	longbanner_EUCS ();
    } else {
	longbanner_PLP ();
    }
    return;
}

extern unsigned long jobsize;		/* set in Printjob () */

static void
longbanner_EUCS (void) {
    char files[FILES_LENGTH + 1];
    char username[USER_INFO_LENGTH + 1];
    char delivery[DELIVERY_LENGTH + 1];
    char options[OPTIONS_LENGTH + 1];
    char erccreg[ERCCREG_LENGTH + 1];
    char account[ACCOUNT_LENGTH + 1];
    char c;

    printed_time = time ((time_t *) NULL); /* get the actual time */

    *username = *delivery = *options = *erccreg = *account = *files = '\0';
    interpret_job (JOBNAME, files, username, delivery, options, erccreg, account);
    if (strstr (options, "a4") != NULL)
	PW = 80;

    breakline ('*');
    (void) sprintf (bline, "*%*c*", PW - 2, ' ');
    Out_line ();

#ifdef STEF_BANNER
    /* hostname begins with a capitalized letter */
    c = FROMHOST[0];
    if (c > 96) {
        c = c - 32;
	FROMHOST[0] = c;
    }
#endif /* STEF_BANNER */

    if (PW <= 80) {
	int remaining = PW - 34;
	(void) sprintf (bline, "* %-14.14s %-*.*s %*.*s %14.14s *",
			FROMHOST, 
			remaining / 2, remaining / 2, 
			(username[0] != '\0') ? username : LOGNAME,
			remaining - (remaining / 2 + 1), 
			remaining - (remaining / 2 + 1),
			LOGNAME, FROMHOST);

	Out_line ();
	(void) sprintf (bline, "* %-14.14s %-*.*s %14.14s *",
			FROMHOST, remaining, remaining, Time_str (), FROMHOST);
	Out_line ();
	(void) sprintf (bline, "* %-14.14s DELIVER TO - %*.*s %14.14s *",
			FROMHOST, PW - 47, PW - 47, delivery, FROMHOST);
	Out_line ();
	(void) sprintf (bline, "* %-14.14s %-*.*s %14.14s *",
			FROMHOST, 
			remaining, 
			remaining, 
			(files[0] != '\0') ? files : JOBNAME, FROMHOST);
	Out_line ();
	(void) sprintf (bline, "* %-14.14s %9ldk%*s %14.14s *",
			FROMHOST, jobsize, remaining - 10, "", FROMHOST);
	Out_line ();

    } else {
	(void) sprintf (bline,
		 "** %-15.15s %-15.15s %-20.20s %-11.11s %-29.29s %-15.15s %-15.15s **",
			FROMHOST, 
			FROMHOST, 
			(username[0] != '\0') ? username : LOGNAME, 
			LOGNAME, 
			Time_str (), 
			FROMHOST, 
			FROMHOST);
	Out_line ();
	(void) sprintf (bline,
		      "** %-15.15s %-15.15s DELIVER TO - %-49.49s %-15.15s %-15.15s **",
			FROMHOST, FROMHOST, delivery, FROMHOST, FROMHOST);
	Out_line ();
	(void) sprintf (bline, "** %-15.15s %-15.15s %-62.62s %-15.15s %-15.15s **",
			FROMHOST, 
			FROMHOST, 
			(files[0] != '\0') ? files : JOBNAME, 
			FROMHOST, FROMHOST);
	Out_line ();
	(void) sprintf (bline, "** %-15.15s %-15.15s %10ldk%51s %-15.15s %-15.15s **",
			FROMHOST, FROMHOST, jobsize, "", FROMHOST, FROMHOST);
	Out_line ();
    }

    (void) sprintf (bline, "*%*c*", PW - 2, ' ');
    Out_line ();
    breakline ('*');
}

static void
longendbanner_EUCS (void) {
    char files[FILES_LENGTH + 1];
    char username[USER_INFO_LENGTH + 1];
    char delivery[DELIVERY_LENGTH + 1];
    char options[OPTIONS_LENGTH + 1];
    char erccreg[ERCCREG_LENGTH + 1];
    char account[ACCOUNT_LENGTH + 1];
    char c;

    *username = *delivery = *options = *erccreg = *account = *files = '\0';
    interpret_job (JOBNAME, files, username, delivery, options, erccreg, account); 
    if (strstr (options, "a4") != NULL)
	PW = 80;

    breakline ('*');
    (void) sprintf (bline, "*%*c*", PW - 2, ' ');
    Out_line ();

#ifdef STEF_BANNER
    /* hostname begins with a capitalized letter */
    c = FROMHOST[0];
    if (c > 96) {
        c = c - 32;
	FROMHOST[0] = c;
    }
#endif /* STEF_BANNER */

    if (PW <= 80) {
	int remaining = PW - 34;
	(void) sprintf (bline, "* %-14.14s %-*.*s %*.*s %14.14s *",
			FROMHOST, 
			remaining / 2, remaining / 2, 
			(username[0] != '\0') ? username : LOGNAME,
			remaining - (remaining / 2 + 1), 
			remaining - (remaining / 2 + 1),
			LOGNAME, FROMHOST);

	Out_line ();
	(void) sprintf (bline, "* %-14.14s %-*.*s %14.14s *",
			FROMHOST, remaining, remaining, Time_str (), FROMHOST);
	Out_line ();
	(void) sprintf (bline, "* %-14.14s - END OF JOB - %*.*s %14.14s *",
			FROMHOST, PW - 49, PW - 49, " ", FROMHOST);
	Out_line ();
	(void) sprintf (bline, "* %-14.14s %-*.*s %14.14s *",
			FROMHOST, 
			remaining, 
			remaining, 
			(files[0] != '\0') ? files : JOBNAME, 
			FROMHOST);
	Out_line ();
	(void) sprintf (bline, "* %-14.14s %9ldk%*s %14.14s *",
			FROMHOST, jobsize, remaining - 10, "", FROMHOST);
	Out_line ();

    } else {
	(void) sprintf (bline,
		 "** %-15.15s %-15.15s %-20.20s %-11.11s %-29.29s %-15.15s %-15.15s **",
			FROMHOST, 
			FROMHOST, 
			(username[0] != '\0') ? username : LOGNAME, 
			LOGNAME, 
			Time_str (), 
			FROMHOST, 
			FROMHOST);
	Out_line ();
	(void) sprintf (bline,
		      "** %-15.15s %-15.15s - END OF JOB - %47.47s %-15.15s %-15.15s **",
			FROMHOST, FROMHOST, " ", FROMHOST, FROMHOST);
	Out_line ();
	(void) sprintf (bline, "** %-15.15s %-15.15s %-62.62s %-15.15s %-15.15s **",
			FROMHOST, 
			FROMHOST, 
			(files[0] != '\0') ? files : JOBNAME, 
			FROMHOST, 
			FROMHOST);
	Out_line ();
	(void) sprintf (bline, "** %-15.15s %-15.15s %10ldk%51s %-15.15s %-15.15s **",
			FROMHOST, FROMHOST, jobsize, "", FROMHOST, FROMHOST);
	Out_line ();
    }

    (void) sprintf (bline, "*%*c*", PW - 2, ' ');
    Out_line ();
    breakline ('*');
    return;
}

static void
longbanner_PLP (void) {
    int i;			/* ACME integers, INC */
    char c;

    seqinfo ();      /* gets the sequence number from Queue_tmp->q_num */

    for (i = 0; i < BLAST; ++i) {
	breakline ('*');
    }
    for (i = 0; i < TOP_SEP; ++i) {
	breakline (0);
    }
    /*
     * print the Name, Host and Jobname in BIG letters
     * allow some of them to be dropped if there isn't enough
     * room.
     */

    if (PL > 57) { 
        bigprint ("# ", SEQNUMBER); 
    }
    bigprint ("o ", LOGNAME);

#ifdef STEF_BANNER
    /* hostname begins with a capitalized letter */
    c = FROMHOST[0];
    if (c > 96) {
        c = c - 32;
	FROMHOST[0] = c;
    }
#endif /* STEF_BANNER */

    if (PL > 48) { 
        bigprint ("o ", FROMHOST); 
    }
    bigprint ("* ", JOBNAME);

    /*
     * Prints the information about the user, jobname, etc...
     */
    userinfo ();

    for (i = 0; i < BOTTOM_SEP; ++i) {
	breakline (0);
    }

    if (PL > 0) {
	int my_blast = BLAST;
	
	linecnt--; 
	/* use PL-1 here, so we can use a formfeed without worrying about
	 * producing a blank sheet of paper. Also, if the paper is smaller
	 * than 60 lines, try to reduce the blast lines.
	 */
	while (((linecnt + my_blast) > (PL - 2)) && (my_blast > 0)) {
	    my_blast--;
	}

	if (my_blast > 0) {
	    while ((linecnt + my_blast) % (PL - 2)) {
		breakline (0);		/* separator from banner */
	    }
	    while (linecnt % (PL - 2)) {
		breakline ('*');		/* page bottom marker */
	    }
	}

	breakline ('*');		/* page bottom marker */
    }
    return;
}

static void
longendbanner_PLP (void) {
    int i;                    /* ACME integers, INC */
    
    seqinfo ();      /* gets the sequence number from Queue_tmp->q_num */
 
    for (i = 0; i < BLAST; ++i) {
	breakline ('*');
    }
    for (i = 0; i < TOP_SEP; ++i) {
	breakline (0);
    }
    /*
     * print the Name, Host and Jobname in BIG letters
     * allow some of them to be dropped if there isn't enough
     * room.
     */
    
    if (PL > 57) { 
       bigprint ("# ", SEQNUMBER); 
    }
    bigprint ("o ", LOGNAME);
    if (PL > 48) { 
       bigprint ("o ", FROMHOST); 
    }
    bigprint ("* ", JOBNAME);
    
    for (i = 0; i < BOTTOM_SEP; ++i) {
	breakline (0);
    }
    
    if (PL > 0) {
	int my_blast = BLAST;
	
	/* use PL-1 here, so we can use a formfeed without worrying about
	 * producing a blank sheet of paper. Also, if the paper is smaller
	 * than 60 lines, try to reduce the blast lines.
	 */
	while (((linecnt + my_blast) > (PL - 1)) && (my_blast > 0)) {
	    my_blast--;
	}
	
	if (my_blast > 0) {
	    while ((linecnt + my_blast) % (PL - 2)) {
		breakline (0);          /* separator from banner */
	    }
	    while (linecnt % (PL - 2)) {
		breakline ('*');                /* page bottom marker */
	    }
	}
	
	breakline ('*');                /* page bottom marker */
    }
    return;
}


static void
breakline (char c) {
    int i;

    if (c) {

#ifdef STEF_BANNER
	/* put the sequence number in the break lines. I don't like
	 * the look of it, but Stefano does -- what difference does
	 * ~5 lines make anyway? ;)
	 */
	for (i = 0; i < PW - 12; i++) {
	    bline[i] = c;
	}
	if (Queue_tmp->q_num < 10) {
	    (void) sprintf (&(bline[i]), "** SEQ NR. %1.1s", SEQNUMBER);
	} else if (Queue_tmp->q_num < 100) {
	    (void) sprintf (&(bline[i]), "* SEQ NR. %2.2s", SEQNUMBER);
	} else {
	    (void) sprintf (&(bline[i]), " SEQ NR. %3.3s", SEQNUMBER);
	}
	/* sprintf will append the '\0'. */

#else
	for (i = 0; i < PW; i++)
	    bline[i] = c;
	bline[i] = '\0';
#endif

    } else {
	bline[0] = '\0';
    }
    Out_line ();
    return;
}

/***************************************************************************
 * do_char (fgbg, c, line, buf)
 *  print out the raster line corresponding to character c, line l
 *  If a character is not a descender, you can use the direct mapping.
 *  If it is a descender,  you have to shift down by the descender amount.
 *
 *  1.  for lines above the "top" of a descender line
 *      if the character is a descender, print a space
 *      otherwise print normally
 *  2.  for lines at and below the top of a descender
 *      if the character is a descender, print line
 *      otherwise print normally
 *  3.  for lines below the baseline
 *      if a descender, print line
 *      otherwise print blank
 *
 * fgbg is a 2-element string consisting of one character for "on" pixels,
 * and one character for "off" pixels (foreground, background).
 ***************************************************************************/
static inline int
is_descender (char c) {
    switch (c) {
    case '_':
    case ';':
    case ',':
    case 'g':
    case 'p':
    case 'q':
    case 'y':
	return True;
    default:
	break;
    }
    return False;
}

static inline void
do_char (char *fgbg, char c, int line, char *buf) {
    int i;			/* ACME Integer, Inc. */
    int position;		/* index into chartable */
    int rep;			/* representation of character scan line */

    assert (fgbg && fgbg[0] && fgbg[1]);

    if (!isprint (c)) {
	c = ' ';
    }
    if (c == 'j') {
        position = 0;
    } else {
        position = c - ' ';
    }
    if (is_descender (c)) {
	if (line >= DROP) {
	    line = line - DROP;
	} else {
	    position = 0;
	    line = 0;
	}
    }

    if (c == 'j') {
        if (line >= HEIGHT+3) {
	    position = 0;
	    line = 0;
	}
    } else {
      if (line >= HEIGHT) {
	  position = 0;
	  line = 0;
      }
    }

    if (c == 'j') {
        rep = chartable2[0][line];
    } else {
        rep = chartable[position][line];
    }
    for (i = 6; i >= 0; --i) {
	if (rep & (1 << i)) {
	    buf[6 - i] = fgbg[0];
	} else {
	    buf[6 - i] = fgbg[1];
	}
    }
    for (i = 7; i < WIDTH; ++i) {
	buf[i] = ' ';
    }
    return;
}

/***************************************************************************
 * bigprint (char * fgbg, char * line)
 * print the line in big characters
 * for i = topline to bottomline do
 *  for each character in the line do
 *    get the scan line representation for the character
 *    foreach bit in the string do
 *       if the bit is set, print X, else print ' ';
 *	  endfor
 *  endfor
 * endfor
 *
 * fgbg is a 2-element string consisting of one character for "on" pixels,
 * and one character for "off" pixels (foreground, background).
 ***************************************************************************/

static void
bigprint (char *fgbg, char *line) {
    int i, j;			/* ACME Integers, Inc. */
    int cnt;			/* number of characters */

    DBGLOCAL5 ("bigprint: '%s'", line);
    cnt = strlen (line);
    i = PW / WIDTH;
    if (cnt > i) {
	cnt = i;
    }
    for (i = 0; i < HEIGHT + DROP; ++i) {
	for (j = 0; j < cnt; ++j) {
	    do_char (fgbg, line[j], i, bline + (j * WIDTH));
	}
	bline[j * WIDTH] = 0;
	Out_line ();
    }
    return;
}

/***************************************************************************
 * write the buffer out to the file descriptor.
 * don't do if fail is invalid.
 ***************************************************************************/

static void
Out_line (void) {
    int i = strlen (bline);

    if (i > PW) {
	if (!LS) {                              /* lprps package needs 
						 * more than 80
						 * characters for the
						 * banner line
						 */ 

	    bline[PW - 1] = '\0';		/* Bobbitise. */
	}
    }

    if (fail == False) {
	if ((i > 0 && write (fd, bline, i) != i)
		|| (write (fd, "\n", 1) != 1))
	{
	    logerr (XLOG_INFO, "Out_line: write failed");
	    fail = True;
	}
    }
    ++linecnt;
    return;
}
