/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1987, 1988, Patrick Powell
 * version 3.3.0 Justin Mason July 1994
 ***************************************************************************
 * MODULE: localprinter.c - local Printer queue job handler
 *
 * (enhancements from dimakop@cti.gr:)
 * Variable *names[] is added so that it keeps the names of the
 * files if there are more than one in a print job.
 * This is necessary because, when cover pages are printed for every file
 * printout in a many file print job the file names should be printed
 * correctly on each of them.
 ***************************************************************************/

#include "lp.h"
#include "LPD/localprinter.h"
#include "LPD/print_support.h"
#include "library/errormsg.h"
#include "library/utils.h"

static int suppress_banner;		/* flag: don't print a banner */
static int read_file_as_submitter;	/* flag */
static int jobs_printed;		/* numbers of jobs processed */
static int files_in_job;		/* number of files in current job */
static int endbanner_printed = False;   /* flag if the endbanner has been printed */

#define NAMES_ARY_LEN		99	/* max number of files in one job */
static char *names[NAMES_ARY_LEN];
static int name_index = 0;
static int writing_account_done; /* for writing in the account file only once */
unsigned long jobsize;		 /* global so that longbanner() can access it */

static int queuejob (FILE *cfp, struct plp_queue *q);
static int Banner_program (char *prog);
static int printfile (int format, char *file);

extern int update_history (void);
extern int create_history_entry (int, struct plp_queue *);

void setup_ep_hl (struct plp_queue *), reset_ep_hl (struct plp_queue *);

/* friendly messages for the error mails. */

#define NO_MORE_PAGES_MSG \
"Your job %d [%s] was cancelled -- no more pages\n\
are allowed on the printer '%s'."

#define NO_PERMISSION_MSG \
"Your job %d [%s] was cancelled -- you don't have permission\n\
to use the printer '%s'."

#define UNSUPPD_FILE_FORMAT_MSG	\
"Your job %d [%s] was cancelled -- the '%c' file format\n\
is not supported by the printer '%s'."

#define OVER_MX_MSG \
"Your job %d [%s] was cancelled -- it was too big\n\
(the size limit for '%s' is %ldk, and your job was %ldk)."

extern char Mail_log_message[];

static int DoneLD = False;

/**********************************************************************
 * 1. set up the state to indicated 0 completed jobs
 * 2. get the printer ready
 **********************************************************************/

int
Local_setup (void) {
    int n;

    DBGLOCAL3 ("Local_setup");
    jobs_printed = 0;

    n = Print_ready (True);
    if (n != JSUCC) 
	return n;

    if (LD && *LD && !DoneLD) {
	DBGLOCAL3 ("Local_setup: LD on open: '%s'", LD);
	n = Print_string (LD);
	DoneLD = True;
    }
    return n;
}

/**********************************************************************
 * called when there is an error in the job handling
 * 1. set up the state to indicated 0 completed jobs
 * 2. close the printer
 **********************************************************************/

void
Local_error (void) {
    DBGLOCAL3 ("Local_error");
    jobs_printed = 0;

    Print_close ();
    DoneLD = False;
}

void
Local_finish (void) {
    DBGLOCAL3 ("Local_finish: called with jobs_printed %d", jobs_printed);
    /*
     * print out trailer on close if a job has been printed
     */
    if (jobs_printed) {
	/*
	 * FQ set means FF needed on close (ie. quit)
	 */
	if (FQ && FF && *FF) {
	    DBGLOCAL3 ("Local_finish: FF on final");
	    (void) Print_string (FF);
	}
	if (TR && *TR) {		/* TR is sent on close */
	    DBGLOCAL2 ("Local_finish: TR '%s'", TR);
	    (void) Print_string (TR);
	}
	jobs_printed = 0;
    }
    Print_close ();
    DoneLD = False;
}

/* RFC-1179 has provision for checking a symlink in the spool dir versus
 * what's pointed to. If they fail, the user has done something like this:
 * 
 *   ln -s /etc/motd x; lpr -s x;rm -f x; ln -s /etc/shadow x
 *
 * which is pretty obviously a security hole. Here's what a news posting
 * from Keith Bostic (bostic@vangogh.CS.Berkeley.EDU) has to say on
 * a similar matter (local mail delivery):
 *
 * >A better defense (credit to Eric Allman) is that the local delivery
 * >agent does the lstat, the open and the fstat on the returned file
 * >descriptor, and then *compares* the values of the stat calls, most
 * >notably the dev/ino numbers.  This is what the current 4BSD local mail
 * >delivery agent does.  This should be absolutely safe on most systems,
 * >and statistically safe on systems with multiple NFS mounted filesystems.
 *
 * That's what we're doing here.
 */

int
is_race_symlink (char *filename, struct stat *stbuf) {
    struct stat lstbuf;
    unsigned long Sdev;
    unsigned long Sino;
    char *spc;

    if (lstat (filename, &lstbuf) < 0) {
	logerr (XLOG_INFO, "is_race_symlink: cannot lstat file %s", filename);
	return True;
    }
    if (!S_ISLNK(lstbuf.st_mode) ||	/* not a symlink so we needn't worry */
    	(SLINKDATA[0] == '\0')) 	/* no valid info, we have no choice */
    {
	return False;
    }
    if (stat (filename, &lstbuf) < 0) {
	logerr (XLOG_INFO, "is_race_symlink: cannot stat file %s", filename);
	return True;
    }
    if ((spc = strchr (SLINKDATA, ' ')) == NULL) {
	return False;			/* not a valid format */
    }
    
    *spc = '\0';
    /* need to use strtoul(); dev_t is unsigned long on SVR4, and if we
     * overflow, we'll produce an incorrect error.
     */
    Sdev = strtoul (SLINKDATA, (char **) NULL, 10);
    Sino = strtoul (spc + 1, (char **) NULL, 10);
    *spc = ' ';

    if ((stbuf->st_dev != Sdev) || (stbuf->st_ino != Sino)) {
	return True;		/* it's a different file; shout about it */
    }
    return False;			/* it's the same file. */
}

static int
print_FF (char *current_state) {
    int jstatus = JSUCC;		/* job status */

    if (FF && *FF && !SF) {
	DBGLOCAL3 ("print_FF: printing FF %s", current_state);
	jstatus = Print_string (FF);
	if (jstatus != JSUCC) {
	    log (XLOG_INFO, "printing FF %s failed", current_state);
	}
    }
    return jstatus;
}

static int
do_checkperms (char *host, char *user, char *queue, int *perms) {
    int result;
    struct plp_queue *q_for_msg;

    result = Checkperm (host, user, queue, perms, (int *) NULL, files_in_job);

    q_for_msg = Queue;
    if (result == -2) {		/* -2 means "out of pages" */
	sprintf (Mail_log_message, NO_MORE_PAGES_MSG, 
		 q_for_msg->q_num, q_for_msg->q_data->name, Printer);
	log (XLOG_INFO, "%s: %s@%s: no more pages allowed", queue, user, host);
	return result;

    } else if (result <= 0) {
	sprintf (Mail_log_message, NO_PERMISSION_MSG, 
		 q_for_msg->q_num, q_for_msg->q_data->name, Printer);
	log (XLOG_INFO, "%s: %s@%s: permission denied", queue, user, host);
	return result;

    } else {
	return 1;
    }
}

/*
 * print banner if it has not been disabled by (a) the SH flag being
 * set, or (b) the lpr -h flag was used and the AB (always banner)
 * flag is clear.
 */
static int
print_banner (struct plp_queue *q) {
    int jstatus;		/* job status */

    /*
     * use banner printing program, if available
     */
    if (BP && *BP) {
	if ((jstatus = Banner_program (BP)) != JSUCC) {
	    log (XLOG_WARNING, "print_banner: banner program \"%s\" failed", BP);
	    return jstatus;
	}
    } else {
	if ((jstatus = Print_banner (q)) != JSUCC) {
	    log (XLOG_WARNING, "print_banner: failed to print banner");
	    return jstatus;
	}
    }
    return JSUCC;
}

/*
 * print endbanner if it has not been disabled by (a) the SH flag being
 * set, or (b) the lpr -h flag was used and the AB (always banner)
 * flag is clear.
 */
static int
print_endbanner (struct plp_queue *q) {
    int jstatus;		/* job status */

    /*
     * use banner printing program, if available
     */
    if (EP && *EP) {
	if ((jstatus = Banner_program (EP)) != JSUCC) {
	    log (XLOG_WARNING, "print_endbanner: endbanner program \"%s\" failed", EP);
	    return jstatus;
	}
    } else {
	if ((jstatus = Print_endbanner (q)) != JSUCC) {
	    log (XLOG_WARNING, "print_endbanner: failed to print endbanner");
	    return jstatus;
	}
    }
    return JSUCC;
}


/* See if we can rewrite the hostname into an FQDN.
 * Note that this makes no effort to make sure that the hostname
 * will fit into the buffer supplied (it assumes the buffer is
 * big enough, as it is from Read_controlfile()).
 */

char *
attempt_to_qualify (char *hostname) {
    char *s = NULL;

    if (strchr (hostname, '.') != NULL)
	return hostname;

    if (FQDN && ((s = strchr (FQDN, '.')) != NULL)) {
	/* check to see if the hostname that this controlfile
	 * came from is in the same domain as the hostname of
	 * the host that initiated the current connection.
	 */
	if ((s = host_in_domain (hostname, s + 1)) != NULL) {
	    DEBUG2 ("attempt_to_qualify: hostname %s qualified to %s", hostname, s);
	    (void) strcpy (hostname, s);
	    return hostname;
	}

    } else {
	/* check to see if the hostname that this controlfile
	 * came from is in the same domain as this host.
	 */
	if ((DomainNameUnset == False) && Host_domain && *Host_domain
		&& (s = host_in_domain (hostname, Host_domain)))
	{
	    DEBUG2 ("attempt_to_qualify: hostname %s qualified to %s", hostname, s);
	    (void) strcpy (hostname, s);
	    return hostname;
	}
    }
    DEBUG2 ("attempt_to_qualify: hostname %s could not be qualified", hostname, s);
    return NULL;
}

void
parse_symlink_line (char *arg) {
    if (strsame (arg, "X")) {	/* old PLP compat; support it (urgh) */
	suppress_banner = True;
    } else {			/* symlink authentication data */
	(void) strcpy (SLINKDATA, arg);
	read_file_as_submitter = True;
    }
}

int
Read_controlfile (FILE *cfp, struct plp_queue *q, int printer_is_local) {
    char parm[BUFSIZ];		/* holds a line read in */
    char *arg;			/* control line argument */
    char opt;			/* control line option */
    int i;
    struct plp_queue *q_for_msg;

    /* initialize the CFparm array, which holds values read from file */
    empty_cfparms ();
    read_file_as_submitter = False;
 
    /*
     * read the control file and extract user information
     */
    while (fgets (parm, sizeof (parm), cfp)) {
	if ((arg = strchr (parm, '\n')) == NULL) {
	    log (XLOG_INFO, "bad controlfile (%s), no end-of-line", q->q_name);
	    return JABORT;
	}
	*arg = '\0';
	opt = parm[0];
	arg = parm + 1;
	if ((int) strlen (arg) > MAXPARMLEN) {
	    DEBUG0 ("Read_controlfile: control_file: line too long");
	    continue;		/* ignore it and continue */
	}

	if (!isascii (opt) || !isalnum (opt)) {
	    DEBUG0 ("Read_control_file: controlfile: skipping unknown line: \"%s\"", 
		    parm);
	    continue;		/* ignore it and continue */
	}

	/*
	 * copy it into the appropriate place if needed
	 */
	if (isupper (opt)) {
	    switch (opt) {
	    case 'Q':
		strcpy (Printer, arg);
		break;

	    case 'U':
		goto ignore_previous;

	    case 'N':
                if (name_index > NAMES_ARY_LEN) {
		    log(XLOG_INFO, "Printjob: Too many files to print: %d", name_index);
		    return JABORT;
                }
		names[name_index] = strdup (arg); name_index++;
                names[name_index] = NULL;
		goto ignore_previous;

	    case 'S':
		parse_symlink_line (arg);
		break;

add_to_cfparm:
	    default:
		if (CFparm[opt - 'A'][0]) {
		    log (XLOG_INFO,
			 "duplicate in controlfile: '%c%s' (previous: '%s')",
			 opt, arg, CFparm[opt - 'A']);
		    return JABORT;
		}

ignore_previous:
		if ((int) strlen (arg) >= MAXPARMLEN) {
		    log (XLOG_INFO,
			 "controlfile %s: line too long: '%c%s'",
			 q->q_name, opt, arg);
		    return JABORT;
		}
		(void) strcpy (CFparm[opt - 'A'], arg);
		break;
	    }

	} else if (islower (opt)) {
	    /* check to make sure the name format is valid */
	    if (Job_match (q->q_name, arg) == 0) {
		log (XLOG_INFO, "Read_controlfile: bad file name format: '%s'", parm);
		continue;
	    }

	    /* does the datafile exist? */
	    if (stat (arg, &LO_statb) < 0) {
		if (SS && *SS) {
		    /* this is a multiple-server queue; chances are that
		     * some other server has nabbed the job before we
		     * got to it.
		     */
		    
		    return JSUCC;

		} else if (SLINKDATA[0] && UNLNKFILE[0]) {
		    
		    return JSUCC;

		} else {
		    logerr (XLOG_INFO, "Read_control_file: cannot stat file %s", arg);
		    continue;
		}
	    }

	    if (printer_is_local && !QH) {
		/* does it use a funny format, and do we have a filter? */
		switch (opt) {
		    case 'f':		/* uses the IF filter */
		    case 'l':
		    case 'p':
			break;

		    default:
			if (Filter_name[opt - 'a'] == '\0') {
			    q_for_msg = q;
			    sprintf (Mail_log_message, UNSUPPD_FILE_FORMAT_MSG, 
				     q_for_msg->q_num, q_for_msg->q_data->name, 
				     opt, Printer);
			    log (XLOG_INFO, "no filter for format '%c'!", opt);
			    return JABORT;
			}
			break;
		}
	    }

	    /* the file is okay: add it to the list of jobs to print. */
	    i = Add_name (arg);
	    jobsize = jobsize + (LO_statb.st_size + 1023) / 1024;
	    files_in_job++;

	} else {
	    DEBUG0 ("Read_control_file: controlfile: skipping unknown line: \"%s\"", 
		    parm);
	    continue;           /* ignore it and continue */
	}
    }

    /* Set up any parameters that were not provided to default values */

    if (PWIDTH[0] == '\0') {
	(void) sprintf (PWIDTH, "%d", PW);
    }

    if (CLASSNAME[0] == '\0') {
	CLASSNAME[0] = DEFPRIORITY;
	CLASSNAME[1] = '\0';
    }
    if (Parmcount > 1) {
	(void) strcat (JOBNAME, "...");
    }

    return JSUCC;
}

/* unlink file after printing:
 * we must check to see if this job came from this host, or if we've got
 * NW set. If it did, we can check to see if it exists in our filespace...
 * if it does, and the ownership matches, then we can unlink it.
 */
void
unlink_datafile (char *filename) {
    struct passwd *pwd;

    if (*filename != '/') {
        DEBUG4 ("unlink_datafile: not a full path, ignoring.");
        return;
    }

    if (!NW && hostcmp (ShortHost, FROMHOST)) {
        DEBUG4 ("unlink_datafile: on a different host, ignoring.");
        return;
    }

    pwd = getpwnam (LOGNAME);
    if (pwd == NULL) {
        DEBUG4 ("unlink_datafile: user %s doesn't exist, ignoring.", LOGNAME);
        return;
    }

    /* doing this as the submitter's user id will definitely
     * eliminate all security hole race condition risks, and PLP
     * can do it. The only risk here now is if the controlfile
     * has been forged somehow...
     */
    user_to_root ();
    root_to_uid (pwd->pw_uid);

    unlink (filename);
    DEBUG4 ("unlink_datafile: unlinked '%s'", filename);

    uid_to_root (pwd->pw_uid);
    root_to_user ();
    return;
}

/*
 * we must try and read this datafile using the job submitter's
 * user ID and privileges -- this cuts down on the security holes...
 */
int
open_as_submitter (char *filename, int opts, int args) {
    struct passwd *pwd = NULL;
    int dfd;

    if (read_file_as_submitter) {
	if (hostcmp (ShortHost, FROMHOST)) {
	    /* we're not on the submitter's host -- don't bother trying */
	    read_file_as_submitter = False;

	} else {
	    DBGLOCAL2 ("open_as_submitter: reading data file as '%s'", LOGNAME);

	    if ((pwd = getpwnam (LOGNAME)) == NULL) {
		logerr (XLOG_WARNING, "open_as_submitter: user '%s' doesn't exist!");
		return -1;
	    }
	    user_to_root ();
	    root_to_uid (pwd->pw_uid);
	}
    }

    dfd = open (filename, opts, args);
    
    if (read_file_as_submitter) {
	uid_to_root (pwd->pw_uid);
	root_to_user ();
    }
    return dfd;
}

/**********************************************************************
 *   1.  First scan extracts information which controls printing
 *   2.  Set any default values not passed
 *   3.  Second scan does the printing.
 * Returns:  JBUSY, JFAIL, JSUCC, JABORT
 * Side effects:  sets information vector  CFparm[].
 **********************************************************************/

int
Local_dojob (FILE *cfp, struct plp_queue *q) {
    int jstatus;		  /* job status */
    static int jobs_printed_succ; /* number of jobs printed successfully */
    int perms = 'R';		  /* file perms */
    int need_separating_ff;	  /* flag: have we printed from this job? */
    static int banner_print_done; /* flag for printing the banner only 
				   * once if you have multiple files in 
				   * one job 
				   */
    int tmp_seqnumber, store_seqnumber; /* store temp. seqencenumber */
    struct plp_queue *q_for_msg;
    struct plp_queue *q_for_hist;
    char user_name[32], tmp_host[64];
    char user_name_file[BUFSIZ];
    FILE *user_name_file_fp;

    /********************************************************************
     * syntactic sugar macro -- if this command fails, catch the
     * return status and jump to the error-handling section.
     */
#   define job_relies_on(cmd)	if ((jstatus = cmd) != JSUCC) goto error;
    /********************************************************************/

    if (fseek (cfp, 0L, 0) < 0)
	logerr_die (XLOG_INFO, "Local_dojob: initial fseek failed");

    suppress_banner = False;
    jobsize = 0;
    Parmcount = 0;
    files_in_job = 0;
    jobs_printed_succ = 0;
    banner_print_done = False;
    endbanner_printed = False;

    /* True here means this is a local printer */
    job_relies_on (Read_controlfile (cfp, q, True));

    /* See if there are any files to print */
    if (Parmcount == 0) {
	DBGLOCAL4 ("Local_dojob: no files to print!");
	jstatus = JSUCC;
	goto error;
    }

    /* check to see if the job is over the limit... */
    if (MX > 0 && jobsize > MX) {
        q_for_msg = q;
	sprintf (Mail_log_message, OVER_MX_MSG, 
		 q_for_msg->q_num, q_for_msg->q_data->name, Printer, 
		 (long) MX, (long) jobsize);
	log (XLOG_INFO, "job %s: larger than mx: %ldk.", q->q_name, (long) jobsize);
	goto error;
    }

    /* check permissions */

    if (hostcmp (&q->q_from, FROMHOST)) {
	log (XLOG_INFO,
	     "Warning: ctrl file origin '%s', H entry '%s' do not match (user=%s)",
	     &q->q_from, FROMHOST, LOGNAME);
    }

    if ((FQDN != NULL) && hostcmp (FROMHOST, FQDN)) {
	/* TODO: deal with forwarded jobs using host%host in printer_perms file.
	 * in the meantime, check both hosts for permission and refuse if either
	 * one is not permitted to print.
	 */
       if (do_checkperms (FQDN, LOGNAME, First_name, &perms) <= 0) {
	   sendmail (q, 5);
	   goto error;
       }
    }

    if (do_checkperms (FROMHOST, LOGNAME, First_name, &perms) <= 0) {
        sendmail (q, 5); 
	goto error;
    }
    /*
     * Pass 2: Do the printing now
     */
    if (fseek (cfp, 0L, 0) < 0) {
	/* this is impossible */
	logerr (XLOG_WARNING, "Local_dojob: fseek on %s failed", q->q_name);
	goto error;
    }

    if (Log_LPRs) {
	log_request ("local_dojob: %s: printing job %d for %s@%s, %d bytes", Printer,
		     q->q_num, LOGNAME, FROMHOST, q->q_size);
	if (!Log_no_filenames) {
	    int idx;

	    for (idx = 0; idx < q->q_entries; idx++)
		log_request ("%s: %d cop%s of %s, %d bytes", Printer,
			     q->q_data[idx].copies,
			     q->q_data[idx].copies == 1 ? "y" : "ies",
			     q->q_data[idx].name,
			     q->q_data[idx].size);
	}
    }
    if (QH) {
	/*
	 * queuejob returns one of the Job Status settings
	 */
	jstatus = queuejob (cfp, q);
	goto error;
    }

    /* check _now_ to see if banner should be suppressed */
    if (SH || ((suppress_banner || !BNRNAME[0]) && !AB)) {
	DBGLOCAL3 ("local_do_job: banner is suppressed");
	suppress_banner = True;
    }

    /*
     * CL: close printer between jobs. This implies that FFs should
     * not be sent either on open or between jobs.
     */
    if (CL) {
	FO = False;		/* FF on open = no */
	FQ = False;		/* FF at end = no */
	SF = True;		/* suppress separating FF = yes */
	Local_setup ();
    }

    /* Put out FF on open and between jobs, if necessary */
    if (FO || jobs_printed)
	job_relies_on (print_FF ("before starting loop"));

    name_index = 0;

    /* Issue only a debug message...*/
    if (NB) {
        DBGLOCAL3 ("Local_do_job: endbanner will be suppressed."); 
    } else {
        DBGLOCAL3 ("Local_do_job: endbanner will be printed %s.", 
		   HL ? "first" : "at the end of job");
    }

    /* print each job in the file */
    {
	char parm[BUFSIZ];		/* holds a line read in */
	char *arg;			/* control line argument */
	char opt;			/* control line option */

	need_separating_ff = False;

	while (fgets (parm, sizeof (parm), cfp)) {
	    arg = strchr (parm, '\n');	/* we've already checked this */
	    *arg = '\0';

	    /* setting the actual value of q to the one needed for seq.number et al.*/
	    Queue_tmp = q; 

	    opt = parm[0];
	    arg = parm + 1;

	    if (!islower (opt)) {
                if (opt == 'U')
                    unlink_datafile (arg);
                if (opt == 'S')
                    parse_symlink_line (arg);
                continue;
            }

	    /* Before printing each of the files of the job,
	     * print its associated banner and/or endbanner.
	     */
            if (names[name_index] != NULL) {
		/* we can just do this -- if the name was long enough to
		 * flood our buffer, we wouldn't get this far. */
               (void) strcpy (CFparm['N' - 'A'], names[name_index++]);
	    }

	    /* do we need to print an FF after the previous file? */
	    if (need_separating_ff) {
		job_relies_on (print_FF ("between jobs"));
		if (CL) {
		    Local_finish ();
		    Local_setup ();
		}
	    }

	    if (NB) {   /* suppress endbanner */
	        endbanner_printed = True;
	    }

	    no_bb_banner = False;

	    if ((!suppress_banner) && (BB)) {
		/* if BB (print no banner page between jobs of the
		 * same user) is set and if a user submits more than
		 * one job one by one, no extra banner  
		 * is required; we will generate only one banner. This will be done 
		 * by writing the user name and the sequence number in a short file 
		 * in the spool dir. This file will be read and the username 
		 * compared with the username of the next job.
		 */

		/* construct the filename */
		(void) sprintf (user_name_file, "%s/.user_name", SD);
		if ((user_name_file_fp = fopen_daemon (user_name_file, "r")) == NULL) {
		    /* no file present, send banner and actualize the file */
		    if ((user_name_file_fp = fopen_daemon (user_name_file, "w")) == NULL) {
			log (XLOG_WARNING ,"Local_do_job: user_name file cannot be opened");
		    } else { /* create an entry */
			fprintf (user_name_file_fp, "%s %d %s %s\n%s", 
				 q->q_user, 
				 q->q_num,
				 &q->q_from,
				 BNRNAME,
				 JOBNAME);
			fclose (user_name_file_fp);
			endbanner_printed = True;
		    }
		} else { /* file exists, read the content */
		    (void) fscanf (user_name_file_fp, "%s", user_name);
		    (void) fscanf (user_name_file_fp, "%d", &tmp_seqnumber);
		    (void) fscanf (user_name_file_fp, "%s", tmp_host);
		    if ( (strcmp (user_name, q->q_user) == 0 ) &&
			 (strcmp (tmp_host, &q->q_from) == 0) ){
			/* user and host the same, no new banner required */
			fclose (user_name_file_fp);
			suppress_banner = True;
			endbanner_printed = True;
			no_bb_banner = True;
			DEBUG3 ("Local_dojob: BB set: same user, no banner");
		    } else { /* new user, banner is required, but we have to
			      * check first if an endbanner has to be printed */
			DEBUG3 ("Local_do_job: BB set: user %s different from %s, banner required", user_name, q->q_user);
			(void) fgets (bnrname_bb, 127, user_name_file_fp);
			(void) fgets (jobname_bb, 127, user_name_file_fp);
			/* strip the "\n" */
			bnrname_bb[strlen (bnrname_bb)-1] = '\0';
			{ 
			    int l;
			    for (l = 1; l < strlen (bnrname_bb); l++) {
				bnrname_bb[l-1] = bnrname_bb[l];
			    }
			    bnrname_bb[l-1] = '\0';
			}
			/* use temporary buffers, save values */
			store_seqnumber = Queue_tmp->q_num;
			(void) sprintf (seqnumber_bb, "%3d", tmp_seqnumber);
			(void) strcpy (username_bb, user_name);
			(void) strcpy (hostname_bb, tmp_host);
			/* print banner */
			if (HL) { /* end banner has been already printed */
			    setup_ep_hl (q);
			    if (BP && *BP) {
				job_relies_on (print_FF ("before BP"));
				job_relies_on (Banner_program (BP));
			    } else { /* use builtin banner program */
				job_relies_on (print_banner (q));
				job_relies_on (print_FF ("after banner"));
			    }
			    reset_ep_hl (q);
			    banner_print_done = True;
			} else {
			    if (!NB) {
				if (EP && *EP) {
				    job_relies_on (Banner_program (EP));
				    job_relies_on (print_FF ("after EP"));
				} else { /* use builtin banner program */
				    job_relies_on (print_endbanner (q));
				    job_relies_on (print_FF ("after endbanner"));
				}
				endbanner_printed = True;
			    }
			}
			/*
			(void) strcpy (LOGNAME, store_user_name);
			Queue_tmp->q_num = store_seqnumber;
			(void) strcpy (JOBNAME, store_jobname);
			(void) strcpy (&Queue_tmp->q_from, store_host);*/
		    }
		    fclose (user_name_file_fp);
		    user_name_file_fp = fopen_daemon (user_name_file, "w");
		    fprintf (user_name_file_fp, "%s %d %s %s\n%s", 
			     Queue_tmp->q_user, 
			     Queue_tmp->q_num, 
			     &Queue_tmp->q_from,
			     BNRNAME,
			     JOBNAME);
		    fclose (user_name_file_fp);
		}
	    }

	    DBGLOCAL3 ("Local_do_job: number of files in job: %d", files_in_job);
	    if (files_in_job > 1) { /* multiple copies or more than one file in job */
		if ( (! banner_print_done) && (! endbanner_printed) ){
		    if (!suppress_banner) {
			if ( (HL) && (!endbanner_printed) ) { 
			    /* hl: banner after job, but endbanner first */
			   DBGLOCAL3 ("Local_do_job: endbanner %s printed.", 
				      endbanner_printed ? "has been already"  : "will be now");
			    if (EP && *EP) {
			        job_relies_on (Banner_program (EP));
				job_relies_on (print_FF ("after EP"));
			    } else { /* use builtin endbanner program */
			        job_relies_on (print_endbanner (q));
				job_relies_on (print_FF ("after endbanner"));
			    }			        
			    job_relies_on (printfile (opt, arg));
			    endbanner_printed = True;
			} else { /* banner before job, but endbanner last */
			    DBGLOCAL3 ("Local_do_job: banner is %s printed.", 
				       banner_print_done ? "already"  : "now");
			    job_relies_on (print_banner (q));
			    job_relies_on (print_FF ("before job"));
			    job_relies_on (printfile (opt, arg));
			    banner_print_done = True;
			}
		    } else {
			job_relies_on (printfile (opt, arg));
			need_separating_ff = True;
		    }
		} else {
		    job_relies_on (printfile (opt, arg));
		    need_separating_ff = True;
		}
	    } else { /* single copy */
		if (!suppress_banner) {
		    if (HL)  {	/* hl: banner after job, endbanner first if desired*/
		        if (!NB  && (!endbanner_printed) ) { /* endbanner should be printed */
			    if (EP && *EP)  { /* "end banner printing" program */
			        job_relies_on (print_FF ("before EP"));
				job_relies_on (Banner_program (EP));
			    } else {
			        job_relies_on (print_endbanner (q));
				job_relies_on (print_FF ("after endbanner"));
			    }
			    endbanner_printed = True;
			}
			/* check to see if there is the BB option set;
			 * if so then the banner of the last user should
			 * be printed 
			 */
			if ((BB) && (!endbanner_printed && !banner_print_done)) {
			    setup_ep_hl (q);
			    if (BP && *BP) {
				job_relies_on (print_FF ("before BP"));
				job_relies_on (Banner_program (BP));
			    } else {
				job_relies_on (print_banner (q));
				job_relies_on (print_FF ("before next job"));
			    }				
			    reset_ep_hl (q);
			    banner_print_done = True;
			}
			if (BB && !NB && !endbanner_printed && !banner_print_done) 
			    /* print endbanner of the current job */
			    {
			    if (EP && *EP) {
			        job_relies_on (Banner_program (EP));
				job_relies_on (print_FF ("after EP"));
			    } else { /* use builtin endbanner program */
			        job_relies_on (print_endbanner (q));
				job_relies_on (print_FF ("after endbanner"));
			    }			        
			    }	
			job_relies_on (printfile (opt, arg));
			endbanner_printed = True;
		    } else { /* regular: banner first, then job(s), then endbanner */
		        if (BP && *BP) {   /* "banner printing" program */
			    job_relies_on (Banner_program (BP));
			    job_relies_on (print_FF ("after BP"));
			} else { /* builtin banner program */    
			    job_relies_on (print_banner (q));
			    job_relies_on (print_FF ("before job"));
			}
			job_relies_on (printfile (opt, arg));
			banner_print_done = True;
		    }
		} else {
		    job_relies_on (printfile (opt, arg));
		}
	    }
	    jobs_printed++;
	    jobs_printed_succ++;
	    need_separating_ff = True;
	    DEBUG5 ("Local_dojob: %d job(s) printed out of %d", 
		    jobs_printed_succ, files_in_job);
	    if ((jobs_printed_succ >= files_in_job) && (!suppress_banner) && (!SB) && (!banner_print_done)) {
	      /* all files in one job has been printed, now print endbanner */
   	        if (HL) { /* HL set, print banner now */
 		   if (BP && *BP) {
		       if (BB) { /* no banner between jobs, set values */
			   setup_ep_hl (q);
		       }
		       job_relies_on (print_FF ("before BP"));
		       job_relies_on (Banner_program (BP));
		       if (BB) { /* no banner between jobs, reset values */
			   reset_ep_hl (q);
		       }
		   } else {
		       job_relies_on (print_FF ("after job"));
		       job_relies_on (print_banner (q));
		   }
		   banner_print_done = True;
		} else {
		  if ((NB) || (endbanner_printed)) { /* no endbanner */
		      DEBUG5 ("Local_do_job: endbanner is suppressed.");
		  } else {
		      if (EP && *EP) {  
		          job_relies_on (print_FF ("before EP"));
			  job_relies_on (Banner_program (EP));
		      } else {
			  job_relies_on (print_FF ("before endbanner"));
			  job_relies_on (print_endbanner (q));
		      }
		      endbanner_printed = True;
		  }
		} 
	    } else {
	        DEBUG5 ("Local_dojob: no endbanner now, more files are to be printed or endbanner already printed.");
	    } 
	}
	if (jobs_printed == files_in_job) {
	    jobs_printed_succ = files_in_job = 0;
	    
	}
    }
    /*
     * error and status reporting
     */
error:
    /* don't log a serious (but vague) msg, it just clutters
     * the logs and confuses the users when they get mail.
     * A _useful_ message has already been logged.
     */
    DBGLOCAL3 ("Local_dojob: '%s' finished with status %d", q->q_name, jstatus);
    if (CL)
	Local_finish ();

    /* update history files */
    if (Use_history) {
	DEBUG3 ("Local_do_job: calling create_history_entry with error = %d", jstatus == 1 ? 0 : jstatus);
	q_for_hist = q;
	if (create_history_entry (jstatus, q_for_hist) ) {
	    log (XLOG_WARNING, "Local_do_job: errors during create_history_entry");
	}
	DEBUG3 ("Local_do_job: calling update_history");
	if (update_history () ) {
	    log (XLOG_WARNING, "Local_do_job: errors during update_history");
	}
    }
    
    return (jstatus);
    }

/***********************************************************************
 *   1. determine the format and corresponding filter command
 *   2. if 'p' format, set up pr process to format output
 *   3. output to Printer
 * Returns: JFAIL if retry, JSUCC if success, JABORT if ugly
 ***********************************************************************/

static int
printfile (int format, char *file) {
    int dfd;			/* data file fd */
    int status;			/* return status */
    char *filter = NULL;	/* program Name */
    char buf[BUFSIZ];		/* hold the Printer command */
    int p[2];			/* pipe file descriptors */
    int prchild;		/* Printer process */
    struct stat stbuf;
    char SEQNUMBER[4];          /* guess what...*/
    char filename_for_acct[MAXPARMLEN];
    time_t t;

    DBGLOCAL3 ("printfile: format %c, file %s", format, file);
    /*
     * Open input file
     */
    Setuid_debug ("localprinter");

    /*
     * Get the sequence number and copy it 
     */
    (void) sprintf (SEQNUMBER, "%-3d", Queue_tmp->q_num);
 
    dfd = open_as_submitter (file, O_RDONLY, 0);
    if (dfd < 0) {
	logerr (XLOG_NOTICE, "printfile: open failed '%s'", file);
	return (JABORT);
    }

    status = fstat (dfd, &stbuf);
    if (status < 0) {
	logerr (XLOG_NOTICE, "printfile: fstat failed '%s'", file);
	return (JABORT);
    }

    if (is_race_symlink (file, &stbuf)) {
	log (XLOG_CRIT,
	    "POSSIBLE CRACK ATTEMPT: 'lpr -s' symlink has been rerouted: %s@%s",
	    LOGNAME, FROMHOST);
	return JABORT;
    }

    t = time ((time_t *) NULL);
    printed_time = t;

    /*
     * find filter chain to be generated
     */
    switch (format) {
	/*
	 * 'f' and 'l' formats use the IF filter, or the OF if the IF is
	 * not present
	 */
    case 'f':			/* default */
    case 'l':			/* ignore control */
    case 'p':			/* print file using 'pr' */
	if (IF == 0 || *IF == 0) {
	    filter = NULL;

	} else if (format == 'l') {
	    filter = Setup_filter ('l', IF);

	} else {
	    filter = Setup_filter ('f', IF);
	}
	if (format != 'p') {
	    break;
	}
	/*
	 * create a process to invoke 'pr' first, set up the pr command
	 * if there are any options specified with the OP flag use them
	 */
	if (OP && *OP) {
	    (void) sprintf (buf, "%s %s", PR, OP);
	} else {
	    (void) sprintf (buf, "%s -w%d -l%d %s '%s'",
			    PR, PW, PL, PRTITLE[0] ? "-h" : "", PRTITLE);
	}
	DBGLOCAL3 ("printfile: pr command '%s'", buf);
	/*
	 * create PR process
	 */
	if (pipe (p) < 0) {
	    logerr_die (XLOG_NOTICE, "printfile: cannot make pipe");
	}
	/*
	 * make a pipe, fork PR process Run this process as daemon
	 */
	if ((prchild = fork ()) == 0) {	/* child */
	    if (dup2 (dfd, 0) < 0	/* file is stdin */
		|| dup2 (p[1], 1) < 0)	/* pipe is stdout */
	    {
		logerr_die (XLOG_NOTICE, "printfile: dup2 failed, PR process");
	    }
	    setup_close_on_exec ();	/* this'll close the unused pipe fds */
	    full_daemon_perms ();
	    secure_exec (buf);	/* we are "daemon" */

	} else if (prchild < 0) {
	    logerr_die (XLOG_NOTICE, "printfile: fork for pr failed");
	}
	(void) close (p[1]);	/* close output side */
	(void) close (dfd);
	dfd = p[0];		/* use pipe for input */
	break;

	/*
	 * use the filter Name tagged by 'Xf', where X is format
	 */
    default:
	filter = Setup_filter (format, Filter_name[format - 'a']);
	break;
    }

    if (LA && AF && *AF) {      /* local logging */
	FILE *f;

	if (! writing_account_done) {
	    DBGLOCAL3 ("printfile: writing acct entry");
	} else {
	    DBGLOCAL3 ("printfile: skipping writing acct entry, already done");
	}

	/*
	 * writing account entry
	 */
	if (Parmcount > 1) {
	    if (!writing_account_done) {
		(void) strcpy (filename_for_acct, Queue_tmp->q_data->name);
		(void) strcat (filename_for_acct, " ...");
		if ((f = fopen (AF, "a")) != NULL) {
		    fprintf (f, "%s\t%s\t%s\t%d %c %s\t%s %s",
			     (LOGNAME && *LOGNAME) ? LOGNAME : "NULL",
			     (FROMHOST && *FROMHOST) ? FROMHOST : "NULL",
			     (Printer && *Printer) ? Printer : "NULL",
			     files_in_job, 
			     format, 
			     filename_for_acct, 
			     SEQNUMBER, 
			     ctime (&t));
		    (void) fclose (f);
		    writing_account_done = True;
		} else {
		    logerr (XLOG_INFO,
			    "Can't write acct entry for printer %s", Printer);
		}
	    }
	} else {
	    if ((f = fopen (AF, "a")) != NULL) {
		fprintf (f, "%s\t%s\t%s\t%d %c %s\t%s %s",
			 (LOGNAME && *LOGNAME) ? LOGNAME : "NULL",
			 (FROMHOST && *FROMHOST) ? FROMHOST : "NULL",
			 (Printer && *Printer) ? Printer : "NULL",
			 files_in_job, 
			 format, 
			 Queue_tmp->q_data->name, 
			 SEQNUMBER, 
			 ctime (&t));
		(void) fclose (f);
		writing_account_done = 0;
	    } else {
		logerr (XLOG_INFO,
			"Can't write acct entry for printer %s", Printer);
	    }
	}
    }
    /*
     * start filter
     */
    if (filter == NULL) {
	status = Print_copy (dfd);

    } else {
	DBGLOCAL0 ("printfile: format %c, file %s, filter %s", format, file, filter);

	of_stop ();
	status = Print_filter (dfd, filter, True);
	of_start ();
#ifndef IS_LINUX
	(void) free (filter);
#endif
    }

    (void) close (dfd);
    return (status);
}

/********************************************************************
 * Print the banner using the user specified program
 * 1. setup the filter
 * 2. start program
 * 3. wait for completion.
 ********************************************************************/
static int
Banner_program (char *prog) {
    char *filter;
    int status;

    filter = Setup_filter ('f', prog);

    DBGLOCAL3 ("Banner_program: filter %s", filter);
    status = Print_filter (0, filter, 0);
#ifndef IS_LINUX
    (void) free (filter);
#endif
    return (status);
}

/********************************************************************
 * use the user specified Queue Handler.
 * exec'd
 *  progname arguments \   <- from qh options
 *      -Ccontrol_file \
 *      -PPrinter -wwidth -llength -xwidth -ylength [-c] [-iindent] \
 *              [-Zoptions] [-Cclass] [-Jjob] [-Raccntname] -nlogin -hHost
 *      -FX [affile]
 *
 *  exit codes are used to indicate the different success:
 *   0 - successful; 1 retry; anything else- abandon
 * Returns: JSUCC, etc.
 * NOTE: the queue handler is invoked with the control file as
 *  stdin and the Printer as stdout.
 ********************************************************************/

static int
queuejob (FILE *cfp, struct plp_queue *q) {
    plp_status_t status;
    int ret;
    pid_t id, pid;
    char buf[BUFSIZ];
    char *filter;

    /*
     * set up arguments
     */
    (void) sprintf (buf, "%s -X%s", QH, q->q_name);
    DBGLOCAL0 ("queuejob: %s", buf);
    filter = Setup_filter ('X', buf);
    /*
     * Open the Printer
     */
    Print_open ();

    if ((pid = fork ()) == 0) {	/* child */
	if (dup2 (fileno (cfp), 0) < 0) {	/* stdin is cf */
	    logerr_die (XLOG_NOTICE, "queuejob: dup2 failed");
	}
	if (dup2 (Print_fd, 1) < 0) {		/* stdout is printer */
	    logerr_die (XLOG_NOTICE, "queuejob: dup2 failed");
	}
	setup_close_on_exec (); /* this'll close the unused pipe fds */
	full_daemon_perms ();
	secure_exec (filter);	/* we are "daemon" */

    } else if (pid < 0) {
	logerr_die (XLOG_INFO, "queuejob: fork failed");
    }
#ifndef IS_LINUX
    (void) free (filter);
#endif

    /*
     * wait for process
     */
    DBGLOCAL0 ("queuejob: waiting for queuejob %d", pid);
    qh_pid = pid;       /* Remember the pid in case job is removed */
    id = plp_waitpid (pid, &status, WUNTRACED);
    qh_pid = 0;         /* Pid has gone now */
	DBGLOCAL0 ("queuejob: pid %d status %s", id,
	     Decode_status (&status));
    if (id < 0) {
	logerr (XLOG_INFO, "waitpid");
    }

    if (PLP_WIFSTOPPED (status) || (unsigned) PLP_WEXITSTATUS (status) > 1)
    {
	log (XLOG_INFO, "queuejob process %d failed, status (%s)",
		id, Decode_status (&status));

	ret = JABORT;
    } else if (PLP_WEXITSTATUS (status) == 1) {
	ret = JFAIL;
    } else {
	ret = JSUCC;
    }
    return (ret);
}

/********************************************************************
 * Setup values for printing the banner with an extrenal banner program
 * 
 ********************************************************************/
void
setup_ep_hl (struct plp_queue *q)
{
    struct plp_queue *q_ep_tmp;
    char q_info_file[BUFSIZ];
    int q_info_file_fd;
    int bytes_written;
    /* construct the filename */
    (void) sprintf (q_info_file, "%s/.queue_info", SD);

    /* check if the file exists */
    if ((q_info_file_fd = open_daemon (q_info_file, O_RDONLY, 0640)) < 0) {
	/* file doesn't exist, we have to write the information on it */
	if ((q_info_file_fd = open_daemon (q_info_file, O_WRONLY | O_CREAT, 0640)) < 0) {
	    log (XLOG_INFO, "setup_ep_hl: open (w) for %s failed", q_info_file);
	    return;
	} else {
	    bytes_written = write (q_info_file_fd, q, sizeof (*q));
	    if (bytes_written != sizeof (*q)) {
		log (XLOG_INFO, "setup_ep_hl: write in file %s failed", q_info_file);
	    }
	    (void) close (q_info_file_fd);
	    return;
	}
    } else { 
	/* read the content and compare it with the value stored in q_ep */
	malloc_or_die (q_ep_tmp, sizeof (struct plp_queue));
	if (read (q_info_file_fd, q_ep_tmp, sizeof (struct plp_queue)) < 0) {
	    log (XLOG_INFO, "setup_ep_hl: read from file %s failed", q_info_file);
	    (void) close (q_info_file_fd);
	    return;
	} else { 
	    if ( (strcmp (q_ep_tmp->q_user, q->q_user) == 0) &&
		 (strcmp (&q_ep_tmp->q_from, &q->q_from) == 0) ) {
		/* same user, same host, no action required */
		(void) close (q_info_file_fd);
	    } else {
		/* not same user or same host, action required */
		
		(void) close (q_info_file_fd);
		/* rewrite information */
		if ((q_info_file_fd = open_daemon (q_info_file, 
						   O_WRONLY | O_CREAT, 
						   0640)) < 0) {
		    log (XLOG_INFO, "setup_ep_hl: open (w) for %s failed", 
			 q_info_file);
		    return;
		} else {
		    bytes_written = write (q_info_file_fd, q, sizeof (*q));
		    if (bytes_written != sizeof (*q)) {
			log (XLOG_INFO, "setup_ep_hl: write in file %s failed", 
			     q_info_file);
		    }
		    (void) close (q_info_file_fd);
		    q = q_ep_tmp;
		    (void) strcpy (LOGNAME, q->q_user);
		    (void) strcpy (JOBNAME, jobname_bb);
		}
	    }
	}
    }
    (void) free (q_ep_tmp);
    return;
}

void
reset_ep_hl (struct plp_queue *q)
{
    struct plp_queue *q_ep_tmp;
    char q_info_file[BUFSIZ];
    int q_info_file_fd;
    /* construct the filename */
    (void) sprintf (q_info_file, "%s/.queue_info", SD);

    /* check if the file exists */
    if ((q_info_file_fd = open_daemon (q_info_file, O_RDONLY, 0640)) < 0) {
	return;
    } else {
	malloc_or_die (q_ep_tmp, sizeof (struct plp_queue));
	if (read (q_info_file_fd, q_ep_tmp, sizeof (struct plp_queue)) < 0) {
	    log (XLOG_INFO, "setup_ep_hl: read from file %s failed", q_info_file);
	} else {
	    q = q_ep_tmp;
	    (void) free (q_ep_tmp);
	}
    }
    (void) close (q_info_file_fd);
    return;
}
