/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1987, 1988, Patrick Powell
 * version 3.3.0 Justin Mason July 1994
 ***************************************************************************
 * MODULE: Print_support.c
 * handle the actual output to the Printer
 ***************************************************************************
 * void Print_open(): opens the Printer
 * static int Print_of_fd(): makes an 'of' filter if needed
 * int Print_close(): closes the Printer
 * int Print_ready(): combines Print_open() and Print_of_fd()
 * int of_stop(): stops the 'of' filter
 * int Print_string (str) : prints a string through 'of' or to Printer
 * int Print_copy (fd) : copies a file through 'of' or to Printer
 * int Print_filter (file, cmd) : makes a filter and copies file
 * int Print_banner(): prints a banner through 'of' or to Printer
 * Note: the above procedures which return values return JSUCC on success,
 *   and JFAIL or JABORT on failure
 ***************************************************************************/

#include "lp.h"
#include "library/errormsg.h"
#include "LPD/startprinter.h"
#include "LPD/print_support.h"
#include "LPD/setproctitle.h"
#include "common/setstatus.h"

#ifndef _FCNTL_H
#include <fcntl.h>
#endif

/***************************************************************************
 * Print_open()
 * 	Open the Printer, and set the Print_fd variables
 * 	If the RW printcap flag is set, output is opened RW, otherwise
 * 	opened writeonly in append mode.
 *
 * Side Effect:
 * 	sets the Print_fd variable to a non-zero value
 *      terminates if Printer is unable to be opened
 ****************************************************************************/

static pid_t lppipe_pid;	/* LP-pipe pid */
static int lppipe_running;	/* lp-pipe running */
static int is_lp_pipe;		/* is this LP a subprocess? */

static pid_t of_pid;		/* ofilter process pid */
static int of_running;		/* of filter running */
static int of_fd;		/* pipe to ofilter process */

static int Open_lp_pipe (void), rw_pipe (int[]);
static void Print_of_fd (void), LockDevicefile (int, char *);

static char filter_stop[] = "\031\001";		/* magic string to stop ofilter */

#ifndef O_NOCTTY	
	/* O_NOCTTY = don't make this tty our controlling terminal */
#define O_NOCTTY	0	/* just fake it. */
#endif

void
Print_open (void) {
    /*
     * check to see if it is open
     */
    if (Print_fd) {
	return;
    }
    if (LP == NULL || *LP == '\0') {
	fatal (XLOG_WARNING, "Missing LP value for local Printer");
    }
    if (!called_from_one_job) {
	setstatus ("Waiting for '%s' to become ready since %s (offline?).",
		   Printer, Time_str ());
    }
    setproctitle ("waiting for '%s' to become ready.", Printer);

    if (*LP == '|') {
	is_lp_pipe = True;
	Print_fd = Open_lp_pipe ();

    } else {
	is_lp_pipe = False;
	Print_fd = open (LP,
		(RW ? O_RDWR : O_WRONLY) | O_NOCTTY | O_APPEND, 0644);

	if (!strsame (LP, "/dev/null")) {
	    LockDevicefile (Print_fd, LP);
	}
    }

    assert (Print_fd != 0);

    if (Print_fd < 0) {
	setstatus ("Cannot open %s! (%s)", LP, Errormsg (errno));
	logerr_die (XLOG_WARNING, "Print_open: cannot open %s", LP);
    }

    /*
     * if it is a tty, set the baud rates and control information; also detach from it if
     * TIOCNOTTY permits.
     */
    if (!is_lp_pipe && isatty (Print_fd)) {
	Do_stty (Print_fd);

#ifdef TIOCNOTTY
	if ((ttyfd = open ("/dev/tty", O_WRONLY | O_APPEND, 0644)) != -1) {
	    if (ioctl (ttyfd, TIOCNOTTY, 0) < 0) {
		log (XLOG_WARNING, "cannot detach from controlling terminal");
	    }
	    (void) close (ttyfd);
	}
#endif
    }
    DBGLOCAL7 ("Print_open: %s is fd %d", LP, Print_fd);
}

/*
 * Start up an output filter, if needed.
 */
static void
Print_of_fd (void) {
    int p[2];
    char *cmd;

    if (OF == NULL || *OF == '\0' || of_pid) {
	return;
    }
    /*
     * set ofilter command line
     */
    cmd = Setup_filter ('o', OF);
    /*
     * create pipe and fork
     */
    if (pipe (p) < 0) {
	logerr_die (XLOG_NOTICE, "pipe failed");
    }
    assert ((p[0] >= 3) && (p[1] >= 3));
    if ((of_pid = fork ()) == 0) {		/* child */
	if (dup2 (p[0], 0) < 0			/* pipe is stdin */
	    || dup2 (Print_fd, 1) < 0)		/* Printer is stdout */
	{
	    logerr_die (XLOG_NOTICE, "Print_of_fd: dup2 failed");
	}
	DBGLOCAL2 ("starting ofilter: %s", cmd);
	setup_close_on_exec ();	/* this'll close the unused pipe fds */
	full_daemon_perms ();
	secure_exec (cmd);	/* run as "daemon" */

    } else if (of_pid < 0) {
	logerr_die (XLOG_NOTICE, "Print_of_fd: fork failed");
    }
    (void) close (p[0]);	/* close input side */
    (void) free (cmd);
    of_fd = p[1];
    of_running = of_pid;
    DBGLOCAL3 ("started ofilter process %d, fd %d", of_pid, of_fd);
}

static int wait_for_process (pid_t pid, char *errtag) {
    pid_t i;
    plp_status_t statb;

    i = plp_waitpid (pid, &statb, WUNTRACED);
    DBGLOCAL3 ("wait for %s: finished (%s)", errtag, Decode_status (&statb));

    if (i < 0) {
	logerr (XLOG_WARNING, "waitpid (%s)", errtag);
	return (JABORT);
    }

    if (PLP_WIFSTOPPED(statb)) {
	log (XLOG_WARNING, "%s didn't restart properly -- stop queue (%s)",
	    errtag, Decode_status (&statb));
	return (JABORT);

    } else if (PLP_WEXITSTATUS(statb) > 1) {
	log (XLOG_WARNING, "%s died seriously -- stop queue (%s)",
	    errtag, Decode_status (&statb));
	return (JABORT);

    } else if (PLP_WEXITSTATUS(statb) != 0) {
	if (RT < 0) {
	    /* if infinite retries are specified, this is expected. */
	    DBGLOCAL1 ("%s died (%s)", errtag, Decode_status (&statb));
	} else {
	    log (XLOG_WARNING, "%s died (%s)", errtag, Decode_status (&statb));
	}
	return (JFAIL);
    }

    /* else, success */
    return (JSUCC);
}

/***************************************************************************
 * Print_close()
 * 1. Close the pipes and outputs.
 * 2. Signal the of process that it is to do something.
 * 3. Signal the lp-pipe process that it is to do something.
 * 4. Wait for the lp-pipe process.
 ***************************************************************************/

int
Print_close (void) {
    int ret;

    ret = JSUCC;

    if (of_fd) {
	DBGLOCAL2 ("Print_close: closing of filter (pid %d, fd %d)",
	   		 of_pid, of_fd);
	(void) close (of_fd);
	of_fd = 0;
    }
    if (Print_fd) {
	DBGLOCAL2 ("Print_close: closing printer (fd %d)", Print_fd);
	(void) close (Print_fd);	/* close Printer */
	Print_fd = 0;
    }

    if (of_pid) {
	if (of_running == False && FB) {
	    DBGLOCAL2 ("Print_close: sending SIGCONT to ofilter (pid %d)",
			of_pid);
	    (void) kill (of_pid, SIGCONT);
	}
	ret = wait_for_process (of_pid, "ofilter");
    }
    if (lppipe_pid) {
	ret = wait_for_process (lppipe_pid, "LP-pipe");
    }
    lppipe_pid = of_pid = 0;
    return (ret);
}

/***************************************************************************
 * Print_ready()
 * 1. open the Printer if neccessary
 * 2. start up the output filter
 * 3. if the filter has been stopped, start it if the argument says
 *    we should (so we can inhibit this for datafiles)
 *
 * Return: JSUCC if started, JFAIL if not
 ***************************************************************************/
int
Print_ready (int start_of) {
    int jstatus = JSUCC;

    Print_open ();
    Print_of_fd ();

    if (start_of) {
	jstatus = of_start ();
    }
    return (jstatus);
}

int
of_start (void) {
    if (!of_pid || of_running)
	return JSUCC;
    if (!FB)		/* we don't do filters BSD-style */
	return JSUCC;

    DBGLOCAL2 ("of_start: sending SIGCONT to ofilter (pid %d)", of_pid);
    if (kill (of_pid, SIGCONT) < 0) {
	logerr (XLOG_WARNING, "cannot restart output filter");
	if (Print_close () == JABORT) { return (JABORT); }
	return (JFAIL);
    }
    of_running = of_pid;
    return JSUCC;
}

/***************************************************************************
 * of_stop()
 * wait for the output filter to stop itself
 * Return: JSUCC if stopped, JFAIL if it died
 ***************************************************************************/

int
of_stop (void) {
    plp_status_t statb;
    pid_t pid;

    Print_open ();
    Print_of_fd ();

    if (!FB)		/* we don't do filters BSD-style */
	return JSUCC;

    if (of_pid && of_running) {
	/* stop the ofilter filter */
	DBGLOCAL3 ("of_stop: writing ofilter-stop string");
	if (Print_string (filter_stop) != JSUCC) {
	    if (Print_close () == JABORT) { return (JABORT); }
	    return (JFAIL);
	}

	DBGLOCAL3 ("of_stop: waiting for ofilter %d", of_pid);
	if ((pid = plp_waitpid (of_pid, &statb, WUNTRACED)) < 0)
	    logerr (XLOG_WARNING, "waitpid");

	if (!PLP_WIFSTOPPED (statb)) {
	    logerr (XLOG_WARNING, "of_stop: output filter died (%s)",
		    Decode_status (&statb));
	    if (Print_close () == JABORT) { return (JABORT); }
	    return (JFAIL);
	}

	DBGLOCAL3 ("of_stop: output filter stopped");
	of_running = False;
    }

    return (JSUCC);
}

/***************************************************************************
 * Print_string (char *str)
 * print a string through a filter
 * 1. Enable the line Printer
 * 2. get the filter or device fd;
 * 3. put out the string
 * 4. if unsuccessful, close the Printer
 * Return: JSUCC if successful, JFAIL otherwise
 ***************************************************************************/
int
Print_string (char *str) {
    int f;
    unsigned int l;

    if (Print_ready (True) != JSUCC) {
	return (JFAIL);
    }
    if (of_fd) {
	f = of_fd;
    } else {
	f = Print_fd;
    }
    l = strlen (str);
    if (BPWrite (f, str, l) != l) {
	logerr (XLOG_WARNING, "Print_string: write error");
	if (Print_close () == JABORT) { 
	    return (JABORT); 
	}
	return (JFAIL);
    }
    return (JSUCC);
}


/***************************************************************************
 * Print_copy (int fd)
 * copy a file through a filter
 * 1. ready the Printer
 * 2. generates a file writing the percentage of printed data in it
 * 3. copy the file to the appropriate output device
 * 4. if an error, close the Printer
 ***************************************************************************/
int
Print_copy (int fd) {
    int f;
    long cnt;
    char buf[BUFSIZ];
    int in;			/* bytes read */
    int out;			/* bytes written out */
    long sum = 0, job_size = 0; 
    int percent = 0;            /* percentage of printed bytes */
    FILE *fp_progress;          /* file pointer for the progress file */
    char file_progress[BUFSIZ]; /* progress file name */

    cnt = 0;
    if (Print_ready (False) != JSUCC) {
	return (JFAIL);
    }
    if (of_fd) {
	f = of_fd;
    } else {
	f = Print_fd;
    }
    job_size = (long) Queue->q_size;
    /* Name of the progress file */
    (void) sprintf (file_progress, "%s/progress", SD);
    DBGLOCAL4 ("Print_copy: name of the progress-file: %s", file_progress);
    if ((fp_progress = fopen_daemon (file_progress, "w")) == NULL) {
	log (XLOG_WARNING, "Print_copy: file %s couldn't be opened.", file_progress);
    }
    /* in case that this file is not owned by daemon...*/
    (void) chown (file_progress, DaemonUID, DaemonGID);
    DBGLOCAL4 ("Print_copy: progress-file %s successfully openend for writing",
	       file_progress);
    DBGLOCAL4 ("Print_copy: job_size = %ld", job_size);
    while ((in = read (fd, buf, sizeof (buf))) > 0) {
	out = BPWrite (f, buf, in);
	if (fp_progress) {
 	    sum += out;
	    percent = (int)((sum * 100) / job_size);
	    if (fseek (fp_progress, 0L, SEEK_SET) < 0) {
		log (XLOG_WARNING, "Print_copy: error during positioning in progess-file.");
	    }
	    (void) fprintf (fp_progress, "%d", percent);
	    (void) fflush (fp_progress);
	    DBGLOCAL8 ("Print_copy: printed %d%%", percent);
	}
	if (in != out) {
	    logerr (XLOG_WARNING, "Print_copy: write error");
	    if (Print_close () == JABORT) { 
		/* delete the progress file, even if it doesn't exists */
		(void) fclose (fp_progress);
		(void) unlink (file_progress);
		return (JABORT); 
	    }
	    (void) fclose (fp_progress);
	    (void) unlink (file_progress);
	    return (JFAIL);
	}
	cnt = cnt + out;
    }
    /*
     * completed the reading
     */
    if (in < 0) {
	logerr (XLOG_WARNING, "Print_copy: read error");
	if (Print_close () == JABORT) { 
	    /* delete the progress file, even if it doesn't exists */
	    (void) fclose (fp_progress);
	    (void) unlink (file_progress);
	    return (JABORT); 
	}
	(void) fclose (fp_progress);
	(void) unlink (file_progress);
	return (JFAIL);
    }
    DBGLOCAL3 ("Print_copy: printed %d bytes", cnt);
    /* Now the progress file is no longer needed and will be deleted */
    if (fp_progress) {
	(void) fclose (fp_progress);
	(void) unlink (file_progress);
    DBGLOCAL4 ("Print_copy: progress-file %s successfully closed and deleted", 
	       file_progress);
    }
    return (JSUCC);
}

/***************************************************************************
 * Print_filter (int fd, char *cmd, int writing_to_device)
 *  spawn a subprocess to do the printing
 * 1. stop the Printer
 * 2. fork a process
 * 3. wait for process to complete
 * 4. restart the process
 * Return: JSUCC if successful, JFAIL otherwise
 * if writing_to_device is set, Print_open(), Print_close() etc. can be
 * called.
 ***************************************************************************/
int
Print_filter (int file, char *cmd, int writing_to_device) {
    pid_t filter_pid;		/* filter process */
    pid_t pid;			/* daughter pid */
    plp_status_t status;	/* daughter status */

    DBGLOCAL2 ("Print_filter: writing to %s", writing_to_device ? "device" : "file");
    if (writing_to_device && (Print_ready (False) != JSUCC)) {
	return (JFAIL);
    }
    /*
     * fork a process,  and connect file to fd 0, Printer to fd 1.
     */
    if ((filter_pid = fork ()) == 0) {	/* daughter */
	int to_this_fd;
	/*
	 * dup input file to standard in and Printer to stdout.
	 * Note that we ignore an ofilter if it is present and the "fb"
	 * flag is set -- this is for BSD compatibility, and is
	 * needed by eg. lprps.
	 */
	if (FB || of_fd == 0) {
	    to_this_fd = Print_fd;
	} else {
	    to_this_fd = of_fd;
	}

	DBGLOCAL2 ("Print_filter: fd %d", to_this_fd);

	if (dup2 (file, 0) < 0 || dup2 (to_this_fd, 1) < 0) {
	    logerr_die (XLOG_NOTICE, "Print_filter: dup2 failed filter %s", cmd);
	}
	setup_close_on_exec ();		/* this'll close the unused pipe fds */
	full_daemon_perms ();
	secure_exec (cmd);

    } else if (filter_pid < 0) {
	logerr (XLOG_NOTICE, "Print_filter: fork failed");
	if (writing_to_device) {
	    if (Print_close () == JABORT) { 
		return (JABORT); 
	    }
	}
	return (JFAIL);
    }

    /* Wait for filter to complete */
    DBGLOCAL2("Print_filter: waiting for pid %d", filter_pid);
    pid = plp_waitpid (filter_pid, &status, WUNTRACED);
    DBGLOCAL2("filter finished (%s)", Decode_status (&status));

    /*
     * Check to see how filter terminated
     */
    if (pid < 0) {
	if (!Job_removed ()) {
	    /* if the job is removed, the filter exit status doesn't matter. */
	    logerr (XLOG_WARNING, "waitpid");
	    return (JABORT);
	} else {
	    return (JSUCC);
	}

    } else if ((PLP_WIFEXITED (status)) && (PLP_WEXITSTATUS (status) > 1)) {
	/* exited with a status > 1, don't run this again */

	log (XLOG_WARNING, "Filter malfunctioned (%s): %s",
	    Decode_status (&status), cmd);

	if (writing_to_device) {
	    if (Print_close () == JABORT) { 
		return (JABORT); 
	    }
	}
	return (JABORT);

    } else if ((!PLP_WIFEXITED (status)) || (PLP_WEXITSTATUS (status) != 0)) {
	/*
	 * either died from a signal, coredumped, or exited
	 * with an LP_RETRY status, so try again.
	 * 
	 * RT == 0 implies many retries expected - eg padprinters; no point
	 * logging such retries.
	 */
	if (RT > 0) {
	    log (XLOG_INFO, "Filter wants retry: %s", cmd);

	    if (writing_to_device) {
		if (Print_close () == JABORT) { 
		    return (JABORT); 
		}
	    }
	}
	return (JFAIL);
    }
    if (writing_to_device && CL) {
        return (Print_close ());
    } else {
	return (JSUCC);
    }
}

/*
 * Print_banner() 1. get the Printer ready 2. call the banner() routine 
 * with the correct parameter
 */

int
Print_banner (struct plp_queue *q) {
    int to_this_fd;

    DBGLOCAL3 ("Print_banner: printing banner");
    if (Print_ready (True) != JSUCC) {
	return (JFAIL);
    }
    if (of_fd) {
	to_this_fd = of_fd;
    } else {
	to_this_fd = Print_fd;
    }
    if (banner (to_this_fd, q) != JSUCC) {
	if (Print_close () == JABORT) { return (JABORT); }
	return (JFAIL);
    }
    return (JSUCC);
}

/*
 * Print_banner() 1. get the Printer ready 2. call the banner() routine 
 * with the correct parameter
 */

int
Print_endbanner (struct plp_queue *q) {
    int to_this_fd;

    DBGLOCAL3 ("Print_endbanner: printing endbanner");
    if (Print_ready (True) != JSUCC) {
	return (JFAIL);
    }
    if (of_fd) {
	to_this_fd = of_fd;
    } else {
	to_this_fd = Print_fd;
    }
    if (endbanner (to_this_fd) != JSUCC) {
	if (Print_close () == JABORT) { return (JABORT); }
	return (JFAIL);
    }
    return (JSUCC);
}

/* called when the active job is suddenly lprm'ed. */
void
Print_intr (void) {
    if (Print_fd <= 0) {
	DBGLOCAL3 ("Print_intr: printer is not open");
    }
    /* send a ^C^D in case it's a postscript printer. */
    Print_string ("\003\004");
    Print_close ();
}

/***************************************************************************
 * LockDevice(fd, char *devname)
 * Tries to lock the device file so that two or more queues can work on
 * the same print device. First does a non-blocking lock, if this fails,
 * puts a nice message in the status file and blocks in a second lock.
 * (contributed by Michael Joosten <joost@cadlab.de>)
 ***************************************************************************/

static void 
LockDevicefile (int fd, char *devname) {
    int lock;

    if ((lock = do_devlock (fd, devname, 1)) < 0) {
	logerr_die (XLOG_CRIT, "LockDevice: test lock '%s' failed", devname);
    } else if (lock == 0) {
	DBGLOCAL3 ("LockDevice: device '%s' already locked", devname);
	setstatus ("Printer busy, waiting for lock on %s.", devname);
        setproctitle ("Waiting for '%s': blocking lock on device %s.", Printer, devname);

	/* must wait for other lpd incarnation to finish the job */
	lock = do_devlock (fd, devname, 0);	/* blocking lock */
	/* waiting here... */
	if (lock < 0) {
	    logerr_die (XLOG_CRIT, "LockDevice: lock '%s' failed", devname);
	}
    }
    DBGLOCAL3 ("LockDevice: device '%s' - got lock", devname);

    /* got the lock, proceed with printing */
}

/* Given an LP string as follows:
 * LP = "| /usr/local/etc/lp-pipe port=x host=y whatever",
 * set up an fd which pipes into this command (which gets the usual
 * filter args as well). The command's fds are as follows:
 * 0 = the input, 1 = closed, 2 = the logfile.
 */

static int
Open_lp_pipe (void) {
    char *lppipe, *pipecmd, *oAF;
    int fds[2];

    lppipe = LP + 1;

    DBGLOCAL1 ("opening lp-pipe: '%s'", lppipe);

    /* set it up with the std filter arguments. */

    oAF = AF;
    AF = ST;
    pipecmd = Setup_filter ('f', lppipe);
    AF = oAF;
    if (pipecmd == 0 || *pipecmd == 0) {
        fatal (XLOG_WARNING, "cannot open lp-pipe '%s'", lppipe);
    }

    if (RW) {
	if (rw_pipe (fds) < 0) {
	    logerr_die (XLOG_CRIT, "can't create r/w pipe for lp-pipe");
	}
    } else {
	if (pipe (fds) < 0) {
	    logerr_die (XLOG_CRIT, "can't create pipe for lp-pipe");
	}
    }

    if ((lppipe_pid = fork ()) == 0) {		/* daughter */
	if (dup2 (fds[0], 0) < 0) {
	    logerr_die (XLOG_CRIT, "can't dup lp-pipe input to stdin");
	}
	setup_close_on_exec ();
	full_daemon_perms ();
        secure_exec (pipecmd);

    } else if (lppipe_pid < 0) {
        logerr_die (XLOG_NOTICE, "lp-pipe: fork failed");
    }
    (void) free (pipecmd);

    DEBUG3 ("lp-pipe started, pid %d, fd %d", lppipe_pid, fds[1]);
    
    close (fds[0]);
    lppipe_running = lppipe_pid;
    return (fds[1]);
}

/*
 * socketpair is not supported on some OSes, including SVR4 (according
 * to "Advanced Programming in the UNIX Environment", Stevens); however,
 * SVR4 uses pipe() to do the same thing, so we can support that here.
 * if we're on anything else, log a warning message but try pipe() anyway.
 */
static int
rw_pipe (int fds[]) {
#ifdef HAVE_SOCKETPAIR
    return (socketpair (AF_UNIX, SOCK_STREAM, 0, fds));
#else

#ifndef SVR4
    /* log a warning message if we're not on SVR4; pipe() may not
     * do the job.
     */
    log (XLOG_WARNING, "warning: 'rw' LP-pipes may not be supported for this OS");
#endif

    return (pipe (fds));
#endif
}
