/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1987, 1988, Patrick Powell
 * version 3.3.0 Justin Mason July 1994
 ***************************************************************************
 * MODULE: recvfiles.c
 * Receive files from a remote site
 ***************************************************************************/

#include "lp.h"
#include "library/errormsg.h"
#include "LPD/setproctitle.h"
#include "LPD/statfs.h"
#include "LPD/recvfiles.h"

/**************************************************************************/

#define ACK_SUCCESS	0	/* succeeded; delete remote copy */
#define ACK_STOP_Q	1	/* failed; stop the remote queue */
#define ACK_RETRY	2	/* failed; retry later */

static int send_ack (int), client_cmd (int *, unsigned long *, char *),
    add_to_tmpfiles (char *), get_file (unsigned long, int, char *),
    Validate_cf (FILE *, char *, char *), check_free_space (unsigned long);

static long minfree;	/* minimum number of blocks to leave free */
static unsigned long jobsize;	/* size of the incoming job */
static int files_recvd;		/* the number of files received so far */

static void get_minfree (void);

extern void empty_cfparms (void);
extern char *attempt_to_qualify (char *);

/***************************************************************************
 * This module implements the file transfer protocol at the receiving
 * site. For each job:
 * 1.  The data files are transferred.
 * 2.  The control files are transferred.
 *     If the transfer is unsuccessful,  all job files are removed.
 *
 * Individual files are transferred using the following protocol.
 *
 * 1.  The sending site sends a line with the size and name of the file,
 * 	and a command code indicating if it is a data or control file.
 * 2.  The receiving site will acknowledge reception.
 * 3.  The remote site will send the file.
 * 4.  The receiving site will acknowledge reception.
 *
 * The transfer protocol is implemented by a simple FSM:
 * INIT:  no files transferred
 * DATA:  data file transferred
 * CONTROL: control file transferred
 *
 * A list of all files associated with the job is maintained,
 * and if any errors result,  the list of files will be deleted.
 *
 ***************************************************************************/

static int
get_file_from_remote (unsigned long size, char *tmpf) {
    int i, succ, fd;

    i = add_to_tmpfiles (tmpf);		/* add tmpf to list of tmp files */

    /*
     * check to see that file size does not exceed the total
     */
    Parms[i].num = 1;
    Parms[i].size = ((size + 1023) / 1024);
    jobsize = jobsize + Parms[i].size;

    if (MX > 0 && jobsize > MX) {
	log (XLOG_INFO, "incoming job exceeds file size limit (mx=%ld)",
		tmpf, (long) MX);
	send_ack (ACK_RETRY);	/* non-crit; won't stop the remote queue. */
	Link_close (); return -1;
    }

    if (!check_free_space (size)) {
	log (XLOG_INFO, "disk space in \"%s\" is below minfree (mi)", SD);
	send_ack (ACK_RETRY);	/* what Sun's lpd sends in this situation. */
	Link_close (); return -1;
    }

    fd = Exlockcf (tmpf);              /* we lock the file (die otherwise) */

    if (((succ = get_file (size, fd, tmpf)) != JSUCC) ||	/* get file */
    		(size && (send_ack (ACK_SUCCESS) != JSUCC)))	/* say thanks */
    {
	log (XLOG_INFO, "recvfiles: failed to get file");
	(void) close (fd); (void) unlink (tmpf);

	/*@@@ Fix from Dave Atkin */
	return -1; /* Avoid assertion failure */
	/* end fix */
	/* return succ; */
    }
    return fd;
}

static void
protocol_error (void) {
    setproctitle ("%s: %s: incoming: protocol violation", Printer, From);
    log (XLOG_INFO, "recvfiles: protocol violation");
    rm_tmpfiles ();
    (void) send_ack (ACK_STOP_Q);
    Link_close (); return;
}

static void
remote_error (void) {
    setproctitle ("%s: %s: incoming: remote error", Printer, From);
    DBGREM0 ("recvfiles: remote error");
    rm_tmpfiles ();
    (void) send_ack (ACK_STOP_Q);
    Link_close (); return;
}

/***************************************************************************
 * recvfiles()
 *
 * Gets a job from the remote end.
 *
 * The old Berkeley protocol was to send data files and then the control
 * file.  The PLP protocol will send the control file first, then the data
 * files, followed by the name of the control file.
 *
 * If there is an error during transfer, all the jobs associated with the
 * file are removed.
 *
 * In order to be compatible with the Berkeley LPD, old Berkeley format is
 * also supported.
 ***************************************************************************/

#define BSD_PROTO	0	/* BSD (std) version of LPD protocol */
#define PLP_PROTO	1	/* PLP (extended) */

void
recvfiles (void)
{
    char tmpf[CFNAMELEN + 1];	/* file name */
    char rewritten[CFNAMELEN + 1]; /* tmp: rewritten cfile name */
    unsigned long size;		/* file size */
    int fd;			/* file descriptor */
    char cfile[CFNAMELEN + 1];	/* controlfile name */
    int ctrl_fd;		/* controlfile fd */
    FILE *cfp;			/* controlfile FILE* */
    int i, succ, cmd_code;
    int Protocol_variant;

    /*
     * get the printcap entry
     */
    if (Log_LPRs) {
	log_request ("job received from %s for printer '%s'", From, Printer);
    }
    if (Get_pc_entry (Printer, Status_pc_vars, Status_pc_len) == 0) {
	log (XLOG_INFO, "recvfiles: non-existent printer: %s", Printer);
    }
    /*
     * set Debug Level from Printcap
     */
    if (DB && DbgRem < DB) {
	DbgRem = DB;
    }
    if ((SD == NULL) || !*SD) {
	/*
	 * no spooling directory, not a Printer
	 */
	DBGREM0 ("recvfiles: not a printer (no sd)");
	return;
    }
    chdir_SD ();
    (void) Checklockfile (LO, (pid_t *) 0, (char *) 0, 0, &LO_statb);

    if (LO_statb.st_mode & DISABLE_QUEUE) {
	DBGREM3 ("recvfiles: spooling disabled");
	send_ack (ACK_STOP_Q);
	return;
    }
    get_minfree ();

    Setuplog (LF, 1);

    setproctitle ("%s: receiving job from %s", Printer, From);

    DBGREM3 ("recvfiles: starting transfer");
    if (send_ack (ACK_SUCCESS) != JSUCC) {
	DBGREM3 ("recvfiles: confirm failed");
	return;
    }

    for (;;) {
	/*
	 * set up the file transfers
	 */
	Protocol_variant = BSD_PROTO;	/* default bsd-style xfer */
	jobsize = (unsigned long) 0;
	fd = ctrl_fd = -1;
	cfp = (FILE *) NULL;
	files_recvd = 0;

	empty_cfparms ();
	while (((succ = client_cmd (&cmd_code, &size, tmpf)) == JSUCC)
		    && (cmd_code != 0))
	{
	    if (cmd_code == CFILE || cmd_code == CNAME) {
		if (cmd_code == CNAME) {
		    /* CNAME indicates PLP. */
		    Protocol_variant = PLP_PROTO;
		    DBGREM3 ("recvfiles: using fj (PLP) protocol");
	 	}
		if (ctrl_fd != -1) {
		    log (XLOG_INFO, "recvfiles: duplicate ctrl! (%s, %s)", cfile, tmpf);
		    protocol_error (); return;
		}
		(void) strcpy (cfile, tmpf); tmpf[0] = 't';
		DBGREM3 ("recvfiles: getting cfile %s as %s", cfile, tmpf);

		ctrl_fd = get_file_from_remote (size, tmpf);
		DBGREM5 ("recvfiles: files_recvd is now %d", files_recvd);
		if (ctrl_fd < 0) { remote_error (); return; }
		if ((files_recvd > 1) && (Protocol_variant != PLP_PROTO)) {
		    break;
		}

	    } else if (cmd_code == DFILE) {
		DBGREM3 ("recvfiles: getting datafile %s", tmpf);
		(void) unlink (tmpf);
		fd = get_file_from_remote (size, tmpf);
		if (fd < 0) { protocol_error (); return; }
		(void) close (fd);

		if ((ctrl_fd != -1) && (Protocol_variant != PLP_PROTO)) {
		    break;
		}
	    } else if (cmd_code == CEND) {
		DBGREM3 ("recvfiles: got PLP end-of-job marker");
		break;
	    }
	}

	if (cmd_code == 0) {
	    /* that's all the jobs we're going to get */
	    break;
	}
	/* otherwise, check the files for consistency and reenter the loop */

	if (succ != JSUCC) {
	    DBGREM3 ("recvfiles: error, cmd_code=%d", cmd_code);
	    remote_error (); return;
	}

	/* now that we've got all the files, let's check them out.
	 * ctrl_fd -> open fd into the control file
	 * cfile -> control file name
	 * Parms -> all files (including the control file!)
	 */

	if (ctrl_fd == -1) {
	    log (XLOG_INFO, "no control file in job!");
	    protocol_error (); return;
	}

	if (files_recvd == 1) {		/* only the control file */
	    log (XLOG_INFO, "no data files in job!");
	    protocol_error (); return;
	}

	for (i = 0; i < files_recvd; ++i) {
	    if (Job_match (cfile, Parms[i].filename) == 0) {
		log (XLOG_INFO, "recvfiles: file with bad format %s", tmpf);
		protocol_error (); return;
	    }
	}

	/* scan the controlfile for entries, and check to see that all
	 * the datafiles are present.
	 */
	if ((cfp = fdopen (ctrl_fd, "r")) == NULL) {
	    logerr_die (XLOG_CRIT, "recvfiles: fdopen failed");
	}

	/* if ctrlfile-logdir is set, copy the control file to it.
	 */
	if (Ctrlfile_logdir && *Ctrlfile_logdir) {
	    char buff[128];
	    (void) sprintf (buff, ".lpd-in.%s", ShortHost);
	    Log_cf (cfile, buff, fileno (cfp));
	}

	/* these must be before Validate_cf () */

	(void) strcpy (tmpf, cfile);
	tmpf[0] = 't';			/* the just-received controlfile */

	(void) strcpy (rewritten, cfile);
	rewritten[0] = 't'; rewritten[1] = 'r';		/* r->"rewritten" */

	if ((Validate_cf (cfp, cfile, rewritten)) == False) {
	    (void) fclose (cfp);
	    (void) unlink (tmpf);
	    (void) unlink (rewritten);
	    (void) unlink (cfile);
	    send_ack (ACK_RETRY); Link_close (); return;
	}

	(void) unlink (tmpf);  /* we're done with the old version of the cfile */

	/* the controlfile is valid -- rename it and reset files_recvd.
	 * This way, even if the current lpd is killed, (1) the job
	 * will be printed when another lpd is started, and (2) the
	 * current copies of the files won't get deleted by rm_tmpfiles().
	 *
	 * Note that this MUST happen AFTER Validate_cf() -- it now checks
	 * for incoming non-PLP control files and renames them so they don't
	 * take a high priority anymore.
	 */
	if (rename (rewritten, cfile) < 0) {
	    logerr (XLOG_WARNING, "recvfiles: cannot rename %s to %s", tmpf, cfile);
	    rm_tmpfiles ();
	    send_ack (ACK_RETRY); Link_close (); return;
	}
	files_recvd = 0;

        if (close (ctrl_fd)) {
	    log (XLOG_INFO, "recvfiles: closing of controlfile %s (fd %d) failed", 
		 cfile, ctrl_fd);
        } else {
		DEBUG4 ("recvfiles: closed controlfile %s (fd %d)", 
		     cfile, ctrl_fd);
        }

       /*
        * Better to do a chmod 600 to the control-file in case of it's in mode 666
        */
       if (chmod (cfile, S_IRUSR | S_IWUSR) < 0) {
           logerr_die (XLOG_CRIT, "recvfiles: cannot chmod 600 file '%s'", cfile);
       } else {
           DBGREM3 ("recvfiles: mode of file '%s' changed to 600", cfile);
       }   

	/*
	 * We can confirm transfer to the other end
	 */
	if (send_ack (ACK_SUCCESS) != JSUCC) {
	    remote_error (); return;
	}
    }

    setproctitle ("%s: %s: starting printer", Printer, From);
    DBGREM0 ("recvfiles: starting printer");

    Link_close (); 
    Startprinter (); 
    return;
}

/***************************************************************************
 * add_to_tmpfiles(char * Name)
 * add a file Name to the received files
 ***************************************************************************/

static int
add_to_tmpfiles (char *filename)
{
    int i;

    DBGREM6 ("add_to_tmpfiles: temp file \"%s\"", filename);
    /*
     * add a control file name to the parameter list entries;
     * if there is nothing in the list, clobber the entry.
     */
    if (files_recvd == 0) {
	Parmcount = 0;
    }
    i = Add_name (filename);
    files_recvd = Parmcount;
    return (i);
}

/***************************************************************************
 * rm_tmpfiles()
 * remove the files that are in the receive list; do this as daemon
 ***************************************************************************/

void
rm_tmpfiles (void)
{
    int i;			/* ACME Integer, Inc. */

    DBGREM6 ("rm_tmpfiles: removing temporary files");

    for (i = 0; i < files_recvd; ++i) {
	(void) unlink (Parms[i].filename);
    }
    files_recvd = 0;
    return;
}

/***************************************************************************
 * client_cmd(int *cmd_code; unsigned long *size; char *name)
 * 1. read a line from the far end
 * 2. the line has the format <cmd_code><size> <name>\n
 *      "%c%d %s\n" format
 * 3. extract the information from the line
 * 4. acknowledge the information
 ***************************************************************************/
static int
client_cmd (int *cmd_code, unsigned long *size, char *name)
{
    char buf[256];		/* small buffer for reading */
    int i;			/* ACME Integers, Inc. */
    char *cp;			/* ACME Pointers, Inc. */

    /*
     * read a line from the remote end use Bomb-proof Read (bpread), which reads up to
     * the first \n
     */
    DBGREM3 ("client_cmd: waiting for command", buf);

    if ((i = bpread (1, buf, sizeof (buf))) < 0) {
	logerr (XLOG_INFO, "client_cmd: read error from remote");
	*cmd_code = 0;
	remote_error ();
	return (JFAIL);

    } else if (i == 0) {
	DBGREM3 ("client_cmd: end of input");
	*cmd_code = 0;
	return (JSUCC);
    }

    i = buf[0];
    if (i != DFILE && i != CFILE && i != CNAME && i != CEND) {
	log (XLOG_INFO, "client_cmd: bad first char (%d)", i);
	protocol_error (); return JFAIL;
    }
    *cmd_code = i;

    /*
     * now pull off the size information
     */
    *size = 0;
    for (cp = &buf[1]; isdigit (*cp); ++cp) {
	*size = 10 * (*size) + ((*cp) - '0');
    }

    if (*cp != ' ') {
	log (XLOG_INFO, "client_cmd: no space separator (%d)", *cp);
	protocol_error (); return JFAIL;
    }
    while (*cp == ' ')		/* skip whitespace */
	++cp;

    if (strlen (cp) >= CFNAMELEN) {
	log (XLOG_INFO, "client_cmd: job name too long '%s'", cp);
	protocol_error (); return JFAIL;
    }

    if (Job_match (cp, cp) == 0) {
	log (XLOG_INFO, "client_cmd: bad job name '%s'", cp);
	protocol_error (); return JFAIL;
    }

    /*
     * must be a data or control file only
     */
    (void) strcpy (name, cp);

    /*
     * send a confirmation
     */
    if (send_ack (0) != JSUCC) {
	remote_error (); return JFAIL;
    }
    DBGREM3 ("client_cmd: command '\\%03o', size %ld, name \"%s\"",
	     *cmd_code, (long) *size, name);
    return JSUCC;
}

/***************************************************************************
 * get_file (unsigned long size; int fd; char *name)
 * 1. copy size bytes from fd 1 to fd
 * 2. the next byte read should be 0; if not, then you have error
 ***************************************************************************/

static int
get_file (unsigned long size, int fd, char *name) {
    char buf[BUFSIZ];		/* buffer for reading */
    unsigned long cnt;		/* number of bytes */
    long i;			/* ACME Integers, Inc. */
    int unknownSize;

    /*
     * we simply copy the file from the remote end
     */
    DBGREM3 ("get_file: get '%s' (%ld bytes) from %s", name, (long) size, From);

    if (size == 0) {
	/* as per protocol */
	unknownSize = True;
	cnt = 0;

    } else {
	unknownSize = False;
	cnt = size;
    }
    
    while (cnt > 0 || unknownSize) {
	if (cnt > sizeof (buf)) {
	    i = sizeof (buf);
	} else {
	    i = cnt;
	}
	i = read (1, buf, i);
	if (i < 0) {
	    logerr (XLOG_INFO, "get_file (from %s): read error", From);
	    remote_error (); 
	    return JFAIL;
	}
	if (i == 0) {
	    if (unknownSize) {
		return JSUCC;		/* connection closed */
	    } else {
		log (XLOG_INFO, "get_file (from %s): nothing received", From);
		remote_error (); 
		return JFAIL;
	    }
	}
	if (write (fd, buf, i) != i) {
	    logerr (XLOG_INFO, "get_file (from %s): write error", From);
	    remote_error (); 
	    return JFAIL;
	}
	cnt = cnt - i;
    }
    i = read (1, buf, 1);
    if (i < 0) {
	logerr (XLOG_INFO, "get_file (from %s): post-file ack: error", From);
	remote_error (); 
	return JFAIL;
    }
    if (i == 0) {
	log (XLOG_INFO, "get_file (from %s): post-file ack not recieved", From);
	remote_error (); 
	return JFAIL;
    }
    if (buf[0] != 0) {
	log (XLOG_INFO, "get_file (from %s): post-file ack: bad (%d)", From, buf[0]);
	remote_error (); 
	return JFAIL;
    }
    return (JSUCC);
}

/***************************************************************************
 * send_ack (int c)
 * acknowledge by sending back single char c
 ***************************************************************************/

static int
send_ack (int c) {
    char buf[1];

    buf[0] = (char) c;
    if (write (1, buf, 1) != 1) {
	if ((errno != EBADF) &&		/* the fd has already been closed */
		(errno != EPIPE))	/* the remote process has exited */
	{
	    logerr (XLOG_INFO, "send_ack: '\\%03o' to %s failed", c, From);
	}
	return JFAIL;
    }
    if (c == 0) {
	DBGREM5 ("sent acknowledge byte");
    } else {
	DBGREM5 ("sent bad result byte");
    }
    return JSUCC;
}

/***************************************************************************
 * Validate_cf (fp, tmpf, rewritten)
 * Check to ensure that the files contained in this control file
 * were actually sent as part of the control file.
 * 1. fseek to the start of the file
 * 2. read the file, looking for data file entries
 * 3. Check that the names of the data files are consistent with the
 *    name of the control file
 * 4. Check that the file was sent
 * returns: 0 if error, 1 if successful, -1 if use of printer is not permitted
 ***************************************************************************/

static int
Validate_cf (FILE *fp, char *tmpf, char *rewritten) {
    char buf[BUFSIZ];		/* Buffer, Inc. */
    int c, l;			/* ACME Aluminum Siding and Integers, Inc. */
    char *bp = NULL;		/* Buffer */
    FILE *tfp;
    int sender_is_PLP;

    if (DbgRem > 6) {
	log (XLOG_DEBUG, "Validate_cf: %d files to check...", Parmcount);
	for (c = 0; c < Parmcount; ++c) {
	    log (XLOG_DEBUG, ", %d '%s'(%d)", c,
			    Parms[c].filename, Parms[c].num);
	}
    }
    
    if ((tfp = fopen (rewritten, "w")) == NULL) {
	logerr_die (XLOG_CRIT, "recvfiles: fopen %s", rewritten);
    }
    
    if (fseek (fp, 0L, 0) < 0) {
	logerr_die (XLOG_CRIT, "Validate_cf %s: fseek", tmpf);
    }

    sender_is_PLP = False;	/* assume it's not */

    while (fgets (buf, sizeof (buf), fp)) {
	l = strlen (buf);
	/* make sure that they can't overrun our buffers */
	if (l > MAXPARMLEN) {
	    log (XLOG_INFO, "Validate_cf %s: line too long", tmpf);
	    return False;
	}

	if (buf[l - 1] != '\n') {
	    log (XLOG_INFO, "Validate_cf %s: no newlines!", tmpf);
	    /* @@@ Fix from Dave Atkin */
	    if (l < BUFSIZ-1)  /* Be more resilient */
		strcat (buf,"\n");
            else 
		/* end fix */
		return False; 
	}

	if (*buf == '\0') {
	    continue;		/* blank line -- FTP's DOS lpr does this */
	}

	buf[l - 1] = 0;
	c = *buf;
	bp = buf + 1;

	/* Check to see if it's a valid PLP controlfile
	 * option, and weed out unknown values...
	 */
        if (!isascii (c) || !isalnum (c)) {
            DEBUG0 ("Validate_cf %s: skipping unknown line: \"%s\"", tmpf, buf);
            continue;           /* ignore it and continue */
        }

	if (isupper (c)) {
	    switch (c) {
	    case 'U': 	/* unlink this file - ignored here */
		break;

	    case 'H':	/* host */
		/* if the hostname is non-qualified, this will try and
		 * qualify it using either (1) the connecting host's
		 * domain, and (2) this host's domain.
		 */
		(void) attempt_to_qualify (bp);
		goto add_to_cfparm;

	    case 'X':	/* obsolete PLP option */
		c = 'T';
		goto add_to_cfparm;

	    case 'C':	/* class */
		if ((isupper (bp[0])) && (bp[1] == '\0') &&
			    (tmpf[STARTPR] == bp[0]))
		{
		    sender_is_PLP = True;
		}
		/* fall through */

	    case 'P':	/* person */
	    case 'M':	/* mail address */
	    case 'L':	/* gecos stuff / login name (bsd) */
	    case 'J':	/* job name */
	    case 'T':	/* title */

	    case 'N':	/* name of file */
	    case 'S':	/* symlink data */
	    case 'W':	/* page width */
	    case 'I':	/* indentation */
	    case 'Q':   /* printername */

	    case 'D':	/* PLP: when lpr'ed */
	    case 'R':	/* PLP: accounting info */
	    case 'Z':	/* PLP: Z-options */
add_to_cfparm:
		/* put it into CFparm for later use. */
		(void) strcpy (CFparm[c - 'A'], bp);
		break;

	    default:
		DEBUG0 ("Validate_cf %s: skipping unknown line: \"%s\"", tmpf, buf);
		continue;
	    }

	} else if (islower (c)) {
	    /*
	     * if we fail any of these tests, the file in question is skipped.
	     */
	    if (Job_match (tmpf, bp) == 0) {
		log (XLOG_INFO, "Validate_cf: %s@%s, file %s, bad data file '%s'",
		     LOGNAME, FROMHOST, tmpf, bp);
		continue;
	    }
	    
	    if ((l = Find_name (bp)) < 0) {
		log (XLOG_INFO,
		     "Validate_cf: %s@%s, file %s, data file '%s' not in list",
		     tmpf, bp);
		continue;

	    } else if (Parms[l].num == 0) {
		log (XLOG_INFO,
		     "Validate_cf: %s@%s, file %s, data file '%s' not transferred",
		     tmpf, bp);
		continue;
	    }
	    /* we've already checked the size of the incoming file-set,
	     * so don't do it again. */

	} else {
	    DEBUG0 ("Validate_cf %s: skipping unknown line: \"%s\"", tmpf, buf);
	    continue;
	}

	/* okay -- we're happy with this line. Put it in the rewritten file. */
	(void) fputs (buf, tfp);
	(void) fputc ('\n', tfp);
    }

    /* add a line to control file for queue name
     * line is 'Qqueuename'  -- baba buehler
     */
    sprintf(buf,"Q%s\n",Printer);
    (void) fputs (buf, tfp);   

    if (sender_is_PLP == False) {
	/* we don't have a PLP-style priority field, so this
	 * is not a PLP-generated control file. Assign the
	 * default priority to the new controlfile's
	 * filename so it is sorted correctly.
	 */
	tmpf[STARTPR] = DEFPRIORITY;

	/* set it for CFparm... */
	bp[0] = DEFPRIORITY; bp[1] = '\0';
	(void) strcpy (CFparm['C' - 'A'], bp);

	    DEBUG3 ("sender is not using PLP cfile format");
    }

    /*
     * check to see if the remote submitter has valid authorizations
     */
    if (LOGNAME[0] == '\0') {
	if (AnonUsername && *AnonUsername) {
	    log (XLOG_INFO,
		"incoming job %s from %s: missing username, assuming '%s'",
		tmpf, FROMHOST, AnonUsername);
	    (void) strcpy (LOGNAME, AnonUsername);

	} else {
	    log (XLOG_INFO,
		"incoming job %s from %s: missing username, refused",
		tmpf, FROMHOST);
	    return False;
	}
    }

    (void) fclose (tfp);

    if (FD) {
	/*
	 * check to see if you will accept forwarded files; if FD is set, then no
	 * forwarded files are allowed
	 */
	if (hostcmp (FROMHOST, FQDN)) {
	    log (XLOG_INFO, "Validate_cf: %s@%s, forwarded job '%s' refused",
		 LOGNAME, FROMHOST, tmpf);
	    return False;
	}
    }

    /* jmason: important note!
     * 
     * if we check the printer_perms file and refuse a job here, the user will not be
     * notified and the remote queue will be stopped. This is not good, so just pass the
     * job on to the unspooling stage; it'll be stopped there.
     */
    return True;
}

static int
check_free_space (unsigned long size) {
    /*
     * returns 1 if enough space left in spool partition, 0 if out of space
     */
    plp_struct_statfs fsb;
    unsigned long space;

    if (plp_statfs (SD, &fsb) == -1) {
	logerr (XLOG_WARNING, "failed to statfs for %s", SD);
	return True;
    }
    size = (size + 1023) / 1024;
    space = (plp_fs_free_bytes (fsb) + 1023) / 1024;
    if (minfree + size > space)
	return False;

    return True;
}

static void
get_minfree (void) {
    FILE *fp;

    minfree = 0;

    if (!MI || !*MI) {			/* minfree is not being used */
	return;
    }

    if (*MI >= '0' && *MI <= '9') {
        minfree = (long) strtoul (MI, (char **) NULL, 10); /* value is numeric */
	return;
    }

    /* we don't bother caching the value, the lookup
     * only happens once per invocation.
     */
    if ((fp = fopen (MI, "r")) != NULL) {
	int retval;
	retval = fscanf (fp, "%ld", &minfree);
	if (retval != 1) {
	    log (XLOG_INFO, "couldn't read value from minfree file '%s'", MI);
	    minfree = 0;
	}
	fclose (fp);
    }

    DBGREM2 ("minfree value is set to %ld", (long) minfree);
    return;
}
