/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1987, 1988, Patrick Powell
 * version 3.3.0 Justin Mason July 1994
 ***************************************************************************
 * MODULE: remoteprinter.c
 * send a job to a remote Printer
 ***************************************************************************/

#include "lp.h"
#include "library/errormsg.h"
#include "LPD/setproctitle.h"
#include "LPD/print_support.h"
#include "LPD/startprinter.h"
#include "LPD/localprinter.h"
#include "common/setstatus.h"

extern void unlink_datafile (char *);
extern void parse_symlink_line (char *);
extern int open_as_submitter (char *, int, int);
extern int Confirm_timeout;		/* from link_support.c */
extern int update_history (void);
extern int create_history_entry (int, struct plp_queue *);

static int sendfile (FILE *, char *, int);
static int sendname (FILE *, char *, int);
static int senddatafile (char *, int);
static int Print_remotebanner (struct plp_queue *, FILE *);
static int Print_remoteendbanner (struct plp_queue *, FILE *);
static int Remote_Banner_program (char *, FILE *);
static int Remote_EndBanner_program (char *, FILE *);
static int remote_print_FF (char *);

static int banner_printed = False; 
static int jobs_printed = 0;
static int jobs_prepared = 0;
static int files_per_job_sent = 0;
static int error_occurred = 0;

struct plp_queue *q_tmp;

static void
log_spool_full (void) {
    setstatus ("Spool area full on %s@%s at %s, sleeping.", RP, RM, Time_str ());
    log (XLOG_INFO, "no space left on %s@%s", RP, RM);
    setproctitle ("%s: spool area full for %s@%s", RP, RM);
}

/***************************************************************************
 * int Remote_setup()
 * sends the "\02Printer" message to the remote Host
 * Returns: successful JSUCC, otherwise JFAIL
 ***************************************************************************/
int
Remote_setup (void) {
    char buf[BUFSIZ];
    int sleeptime, retry;

    if (RP == 0 || *RP == 0) {
	fatal (XLOG_INFO, "no RP specified for RM (%s)", RM);
    }
    (void) sprintf (buf, "%c%s\n", REQ_RECV, RP);
    retry = 1;
    sleeptime = 15;

    /* Do retries: WSLPD, JetDirect are single-connection
     * LPD implementations, and they need the capacity for configurable
     * retries, as if they were printer devices.
     */
    while ((RT > 0) ? retry < RT : 1) {
	setstatus ("Attempt %d to open link to %s at %s.", retry, RM, Time_str ());

	if (JSUCC == Link_line (1, buf)) {
	    DBGREM3 ("connect to %s: waiting for confirm", RM);
	    if (JSUCC == Link_confirm ()) {
		DBGREM3 ("connect to %s: confirmed RECV request", RM);
		return (JSUCC);
	    } else {
		DBGREM3 ("connect to %s: RECV request not confirmed", RM);
	    }

	} else {
	    DBGREM3 ("connect to %s: remote daemon busy", RM);
	}
	setstatus ("Remote daemon on %s busy, retry %d.", RM, retry);
	setproctitle ("%s: remote daemon on %s busy, retry %d", Printer,
		RM, retry);

	Link_close ();
	sleep (sleeptime);
	retry++;
    }
    return (JFAIL);
}

/***************************************************************************
 * Remote_dojob (FILE *cfp; struct plp_queue *q)
 *   1.  First scan extracts information which controls printing
 *       and finds the names of the files to send.
 *   2.	 The data files are then sent.
 *   3.	 Control file is then sent.
 *   4.  Job files are removed
 *   5.  Control file is removed
 * Returns:  JBUSY, JFAIL, JSUCC, JABORT
 * NOTE: this code was based on a description of the LPD file transfer
 *    protocol.
 *
 * File transfer protocol:
 * Sender: sends line of form 'flag' length filename
 * Receiver: acknowledges with 0 (single byte)
 * Sender: sends file (length bytes of information), then a teminating 0
 * Receiver: acknowledges with 0 (single byte)
 *
 * Tne end is signalled by closing the stream, which results in a read error.
 * Side effects:  sets information vector  CFparm[].
 *************************************************************************/
int
Remote_dojob (FILE *cfp, struct plp_queue *q) {
    int jstatus;		/* job status */
    int success;
    time_t t;
    struct plp_queue *q_for_hist;

    /********************************************************************
     * syntactic sugar macro -- if this command fails, catch the
     * return status and jump to the error-handling section.
     */
#   define job_relies_on(cmd)   if ((jstatus = cmd) != JSUCC) goto error;
    /********************************************************************/

    t = time ((time_t *) 0);
    printed_time = t;

    if (!called_from_one_job) {
	setstatus ("Sending job to %s@%s, started at %s.", RP, RM, Time_str ());
    }
    setproctitle ("%s: sending job to %s@%s", Printer, RP, RM);

    /* if we're logging the control files, do the business */
    if (Ctrlfile_logdir && *Ctrlfile_logdir) {
	char buff[128];
	(void) sprintf (buff, ".lpd-out.%s", ShortHost);
	Log_cf (q->q_name, buff, fileno (cfp));
    }

    jobsize = 0;
    Parmcount = 0;

    /* False here means not a local printer */
    job_relies_on (Read_controlfile (cfp, q, False));

    /* See if there are any files to print */
    if (Parmcount == 0) {
	DBGLOCAL4 ("Remote_dojob: no files to print!");
        jstatus = JSUCC;
        goto error;
    }

    /*
     * We send the control file first if we are using the PLP protocol
     */
    if (FJ) {
	jstatus = sendfile (cfp, q->q_name, CNAME);
	/* no need to check for WSLPD -- it won't be using FJ. */

	if (jstatus == JFULL) {
	    log_spool_full ();
	    goto error;

	} else if (jstatus != JSUCC) {
	    setstatus ("Control file transfer to %s@%s failed at %s.", RP, RM, Time_str ());
	    /* don't log anything here -- it's already logged */
	    setproctitle ("%s: transfer %s to %s failed.", 
			  Printer, q->q_name, RM);
	    goto error;
	}
    }

    /*
     * pass 2: send the data files listed in the control file
     */
    (void) fseek (cfp, 0L, 0);

    {
	char parm[BUFSIZ];              /* holds a line read in */
	char *arg;                      /* control line argument */
	char opt;                       /* control line option */
	char parm_cp[BUFSIZ];           /* copy of parm */

	parm_cp[0] = '\0';              /* initialize the parm_cp array */
	files_per_job_sent = 1;
	while (fgets (parm, sizeof (parm), cfp)) {
	    arg = strchr (parm, '\n');  /* we've already checked this */
	    *arg = '\0';
	
	    opt = parm[0];
	    arg = parm + 1;
	    if ( (strcmp (parm_cp, parm) == 0) && (jstatus == JSUCC) ) { 
                 /* check if parm_cp and parm are the same and the 
		  * data file was successfully transferred; if so, 
		  * then multiple copies of the same job are 
		  * requested, but the datafile will be transferred 
		  * only once to avoid problems on the remote site 
		  * @@@ Stef, 02.06.1995 
		  */
		DBGREM3 ("Remote_dojob: Mulitple copies of one file req., DATAFILE was already transferred");
		goto datafile_transferred;
	    }
	    
            DBGREM3 ("Remote_dojob: contents of parm: %s", parm);
            DBGREM3 ("Remote_dojob: contents of parm_cp: %s", parm_cp);
	    if (!islower (opt)) {
                if (opt == 'U')
                    unlink_datafile (arg);
                if (opt == 'S')
		    parse_symlink_line (arg);
                continue;
            }

            /* if the data filename is not in Parms, it should be skipped. */
            if (Find_name (arg) < 0)
                continue;

	    DBGREM3 ("Remote_dojob: STARTING DATA FILE TRANSFER: %s", arg);

	    if (RB) {
	       q_tmp = q;
	    }
	    
	    Queue_tmp = q;
	    jstatus = senddatafile (arg, DFILE);
	    if (jstatus == JNOCONF) {
		error_occurred = True;
		/* one of the data files failed to get through -- need to retry */
		jstatus = JFAIL;
	    }
	    if (jstatus == JFULL) {
		log_spool_full ();
		DBGREM3 ("Remote_dojob: FAILED (spool full) DATA FILE TRANSFER: %s", arg);
		error_occurred = True;
		goto error;

	    } else if (jstatus != JSUCC) {
		setstatus ("Data file transfer to %s@%s failed at %s.", RP, RM, Time_str ());
		setproctitle ("%s: transfer to %s@%s failed", Printer, RP, RM);
		DBGREM3 ("Remote_dojob: FAILED DATA FILE TRANSFER: %s", arg);
		error_occurred = True;
		goto error;
	    }
	    DBGREM3 ("Remote_dojob: SUCCESSFUL DATA FILE TRANSFER: %s", arg);
	    error_occurred = False;

	 datafile_transferred:
	    strcpy (parm_cp, parm); /* copy parm -> parm_cp */
	    DEBUG3 ("Remote_dojob: number of files per job transferred: %d", 
		    files_per_job_sent);
	    files_per_job_sent++;
	}
    }

    /*
     * send the control file. Also, extend the timeout for the confirm to
     * an hour, as WSLPD won't send a confirm until the job is printed.
     */
    Confirm_timeout = 3600;
    DBGREM3 ("Remote_dojob: STARTING CONTROL FILE TRANSFER: %s", q->q_name);
    if (FJ) {
	jstatus = sendname (cfp, q->q_name, CEND);
    } else {
	jstatus = sendfile (cfp, q->q_name, CFILE);
    }
    Confirm_timeout = 30;

    if (jstatus == JNOCONF) {
	/* WSLPD does this; it's in violation of the RFC, but we may as well
 	 * support it. "be liberal in what you accept", as the saying goes.
	 * Basically, WSLPD starts printing the file straightaway, instead of
	 * confirming the transfer; the Link_confirm() call times out, but
	 * the file _is_ printed, so don't retry it. (ugh)
	 */
	DBGREM1 ("Remote_dojob: SUCCESSFUL CONTROL FILE TRANSFER: %s", q->q_name);
	jstatus = JSUCC;

    } else if (jstatus == JFULL) {
	log_spool_full ();
	DBGREM1 ("Remote_dojob: FAILED (spool full) CONTROL FILE TRANSFER: %s", q->q_name);
	goto error;

    } else if (jstatus != JSUCC) {
	setstatus ("Control file transfer to %s@%s failed at %s.", RP, RM, Time_str ());
	DBGREM3 ("Remote_dojob: FAILED CONTROL FILE TRANSFER: %s", q->q_name);
	goto error;

    } else {
	DBGREM3 ("Remote_dojob: SUCCESSFUL CONTROL FILE TRANSFER: %s", q->q_name);
	jstatus = JSUCC;			/* finished! */
    }

    /* update history files */
    if (Use_history) {
	if (jstatus == JSUCC) {
	    success = 4;
	} else {
	    success = jstatus;
	}
	DEBUG3 ("Local_do_job: calling create_history_entry with error = %d", success);
	q_for_hist = q;
	if (create_history_entry (success, q_for_hist) ) {
	    log (XLOG_WARNING, "Local_do_job: errors during create_history_entry");
	}
	DEBUG3 ("Local_do_job: calling update_history");
	if (update_history () ) {
	    log (XLOG_WARNING, "Local_do_job: errors during update_history");
	}
    }

    jobs_prepared = 0;
    
    /*
     * Edinburgh University remote transfer logging addition:
     */
    if (AR && AF && *AF) {	/* remote logging */
	FILE *f;

	if ((f = fopen (AF, "a")) != NULL) {
	    DBGREM3 ("Remote_dojob: Value of Parmcount: %d", Parmcount);
	    DBGREM3 ("Remote_dojob: writing acct entry for remote transfer");
	    fprintf (f, "%s\t%s\t%s\t%d %s %s\t%d %s",
		     (q->q_user && *q->q_user) ? q->q_user : "NULL", /* user */
#ifdef SHORTHOSTNAME
		     (ShortHost && *ShortHost) ? ShortHost : "NULL", /* shorthost */
#else 
		     (Host && *Host) ? Host : "NULL",                /* host */
#endif /* SHORTHOSTNAME */
		     (Printer && *Printer) ? Printer : "NULL",       /* printer */
		     q->q_entries,                                   /* npages */
		     "remote",                                       /* format */
		     JOBNAME,                                        /* name of job */ 
		     q->q_num,                                       /* seq number */ 
		     ctime (&t));                                    /* spooled time */
	    (void) fclose (f);
	} else {
	    logerr (XLOG_INFO,
		    "can't write acct file '%s' for remote printer %s",
			AF, Printer);
	}
    }

error:
    if (jstatus != JSUCC) {
        Link_close ();
	DEBUG3 ("Remote_dojob: DATAFILE TRANSFER FAILED, sleeping for 5 seconds");
	(void) sleep (5);
    } else {
	error_occurred = False;
    }
    files_per_job_sent = 0;
    return (jstatus);
}

/***************************************************************************
 *    all files have been sent; send a closing 0 and shut down link
 ***************************************************************************/
void
Remote_finish (void) {
    Link_close ();		/* then close link and exit */
}

/***************************************************************************
 * Called when you have retried a couple of times
 ***************************************************************************/
void
Remote_error (void) {
    /* use debug so as to avoid this message taking the place of the
     * more descriptive one we've previously logged. (hopefully ;)
     */
    log (XLOG_DEBUG, "job cannot be sent from %s to %s", Host, RM);
    Link_close ();
}

/***************************************************************************
 * sendfile (FILE* fp; char *name; int flag)
 *   sendfile protocol:
 * 1. line containing the flag, number of bytes, and file name is sent.
 *     This has the format:
 *        "%c%d %s", flag, filesize, name
 *    DFILE (data file) flag is 03,  a CFILE (control file) flag is 02
 * 2. the remote end responds by sending back a single 0 character;
 *    anything else is an error condition and terminates transfer
 * 3. the file is then copied to the remote end and terminated with 0
 * 4. if an error is detected, we send a non-zero character
 *    and close the link.
 ***************************************************************************/

static int
sendfile (FILE *fp, char *name, int flag) {
    int succ;			/* ACME Integer, Inc. */
    struct stat statb;		/* status buffer */

    if (Job_removed ()) {
	/* the current job has been removed; fake out the xfer */
	succ = Link_ack (2);	/* send "skip job" code */
	Link_close ();
	return (JSUCC);
    }
    /*
     * stat the file
     */
    if (fstat (fileno (fp), &statb) < 0) {
	logerr (XLOG_INFO, "sendfile: could not stat %s", name);
	return (JFAIL);
    }
    /*
     * rewind the file
     */
    if (fseek (fp, 0L, 0) < 0) {
	logerr_die (XLOG_INFO, "sendfile: fseek failed '%s'", name);
    }
    succ = sendname (fp, name, flag);
    if (succ == JSUCC) {
	succ = Link_copy (fp, (long) statb.st_size, name);
    }
    if (succ == JSUCC) {
	succ = Link_ack (0);
    }
    if (succ == JSUCC) {
	succ = Link_confirm ();
    }
    if (succ != JSUCC) {
	Link_close ();		/* in case it isn't already closed */
    }
    return (succ);
}

static inline int 
Open_tmpfile (char *buf) {
    int fd;

    fd = Exlockcf (buf);
    DBGREM3 ("Open_tmpfile: file %s, fd %d", buf, fd);
    return (fd);
}

static int
senddatafile (char *name, int flag) {
    int succ, succ_banner, succ_endbanner;	/* ACME Integer, Inc. */
    int suppress_banner = False; 
    int short_banner = False;
    int no_endbanner = False;
    int ld_string = False;
    int tr_string = False;
    int dfd;
    int c, status_pg, status_epg;
    FILE *raw, *raw_filtered, *banner_raw, *banner_end_raw;
    char *filter, tmpf[BUFSIZ];
    struct stat stbuf;
    struct stat st_pgbuf, st_epgbuf;

    if (Job_removed ()) {
	/* the current job has been removed; fake out the xfer */
	succ = Link_ack (2);	/* send "skip job" code */
	Link_close ();
	return (JSUCC);
    }

    dfd = open_as_submitter (name, O_RDONLY, 0);

    if (dfd < 0) {
	logerr (XLOG_INFO, "senddatafile: cannot open data file %s", name);
	setstatus ("Cannot open data file %s!", name);
	return (JABORT);
    }

    succ = fstat (dfd, &stbuf);
    if (succ < 0) {
        logerr (XLOG_NOTICE, "senddatafile: fstat failed '%s'", name);
	setstatus ("Cannot fstat data file %s!", name);
        return (JABORT);
    }

    if (is_race_symlink (name, &stbuf)) {
        log (XLOG_CRIT,
            "POSSIBLE CRACK ATTEMPT: 'lpr -s' symlink has been rerouted: %s@%s",
            LOGNAME, FROMHOST);
        return JABORT;
    }

    if (SB) {
        short_banner = True;
	DBGLOCAL3 ("senddatafile: remote shortbanner will be printed");
    }

    if (NB) {
        no_endbanner = True;
	DBGLOCAL3 ("senddatafile: remote endbannerprinting is suppressed");
    }

    if (SH || ((suppress_banner || !BNRNAME[0]) && !AB)) {
	DBGLOCAL3 ("senddatafile: remote bannerprinting is suppressed");
	suppress_banner = True;
    }

    /*
     * impossible combination!! That makes absolutly NO sense!!
     */
    if (RB && SH && HL) {
        logerr (XLOG_ERR, "RB and SH and HL set simultaneously! Invalid! Contact Administratator to change this");
	sendmail (q_tmp, 5);
	return (JABORT);
	
    }

    if ((RB) && 
	((files_per_job_sent == 1) || (files_per_job_sent == Parmcount))) {
      /* 
       * constructs the bannerpage for sending to the remote printer; 
       * temporary files will be used 
       */
        if (HL) { /* end banner first */
	    banner_raw = fopen_daemon ("banner_epg", "w");
	    if (banner_raw == NULL) {
		logerr_die (XLOG_INFO, "senddatafile: open file (w) banner_epg for banner failed!");
	    }
	    status_epg = stat ("banner_epg", &st_epgbuf);
	    if (status_epg < 0) {
		logerr (XLOG_NOTICE, 
			"senddatafile: fstat failed on file 'banner_epg'");
		return (JABORT);
	    }   
	} else { /* normal banner first */
	    banner_raw = fopen_daemon ("banner_pg", "w");
	    if (banner_raw == NULL) {
		logerr_die (XLOG_INFO, "senddatafile: open file (w) banner_pg for banner failed!");
		 }
	    status_pg = stat ("banner_pg", &st_pgbuf);
	    if (status_pg < 0) {
		logerr (XLOG_NOTICE, 
			"senddatafile: fstat failed on file 'banner_pg'");
		return (JABORT);
	    }   
	}
	if (HL) { /* end banner first */
	    banner_end_raw = fopen_daemon ("banner_pg", "w");
	    if (banner_end_raw == NULL) {
		logerr_die (XLOG_INFO, "senddatafile: open file (w) banner_epg for banner failed!");
	    }
	    status_pg = stat ("banner_pg", &st_pgbuf);
	    if (status_pg < 0) {
		logerr (XLOG_NOTICE, 
			"senddatafile: fstat failed on file 'banner_pg'");
		return (JABORT);
	    }   
	} else { /* normal banner first */
	    banner_end_raw = fopen_daemon ("banner_epg", "w");
	    if (banner_end_raw == NULL) {
		logerr_die (XLOG_INFO, "senddatafile: open file (w) banner_epg for endbanner failed!");
	    }
	    status_epg = stat ("banner_epg", &st_epgbuf);
	    if (status_epg < 0) {
		logerr (XLOG_NOTICE, 
			"senddatafile: fstat failed on file 'banner_epg'");
		return (JABORT);
	    }   
	}
	if (HL) { /* endbanner first, then job, then banner page */
	    if (LD) {
	        ld_string = True;
		(void) fprintf (banner_end_raw, LD);
		if (fseek (banner_end_raw, (long) strlen (LD), SEEK_SET) < 0) {
		    logerr (XLOG_INFO, "senddatafile: generating of LD string failed");
		}
	    }
	    if ( (!short_banner) && (!no_endbanner)
		 && (!suppress_banner) && (files_per_job_sent == 1) ) {
	        DBGREM2 ("senddatafile: generating endbanner on file banner_pg");
		succ_endbanner = Print_remoteendbanner (q_tmp, banner_end_raw);
		fclose (banner_end_raw);
		if (succ_endbanner != JSUCC) {
		    unlink ("banner_epg");
		    logerr (XLOG_INFO, "senddatafile: generating of endbannwer on file banner_pg failed");
		    return (JFAIL);
		}
	    }
	    if ( (banner_raw) && (!suppress_banner) 
		 && (files_per_job_sent == Parmcount)) { 
	        DBGREM2 ("senddatafile: generating banner on file banner_epg");
		succ_banner = Print_remotebanner (q_tmp, banner_raw);
		fclose (banner_raw);
		if (succ_banner != JSUCC) { /* not successful */
		    unlink ("banner_pg");
		    logerr (XLOG_INFO, "senddatafile: generating of endbannwer on file banner_epg failed");
		    return (JFAIL);
		}
	    }
	    if (TR) {
	        tr_string = True;  
		(void) fclose (banner_raw);
		banner_raw = fopen_daemon ("banner_epg", "a");
		if (banner_raw == NULL) {
		    logerr_die (XLOG_INFO, "senddatafile: open file (a) banner_epg for appending banner failed!");
		 }
		status_epg = stat ("banner_epg", &st_epgbuf);
		if (status_epg < 0) {
		    logerr (XLOG_NOTICE, 
			    "senddatafile: fstat failed on file 'banner_epg'");
		    return (JABORT);
		}   
		(void) fprintf (banner_raw, TR);
		(void) fclose (banner_raw);
	    }
	} else { /* banner first, then job, then endbanner */
	    if (LD) {
	      ld_string = True;
	      (void) fprintf (banner_raw, LD);
	      if (fseek (banner_raw, (long) strlen (LD), SEEK_SET) < 0) {
	          logerr (XLOG_INFO, "senddatafile: generating of LD string failed");
	      }
	    }
	    if ((banner_raw) && (!suppress_banner) && (files_per_job_sent == 1)) { 
	        DBGREM2 ("senddatafile: generating banner on file banner_pg");
		succ_banner = Print_remotebanner (q_tmp, banner_raw);
		fclose (banner_raw);
		if (succ_banner != JSUCC) { /* not successful */
		    unlink ("banner_pg");
		    logerr (XLOG_INFO, "senddatafile: generating of endbannwer on file banner_pg failed");
		    return (JFAIL);
		}
	    }
	    if ( (!short_banner) && (!no_endbanner) 
		 && (!suppress_banner) && (files_per_job_sent == Parmcount)) {
	         DBGREM2 ("senddatafile: generating endbanner on file banner_epg");
		 succ_endbanner = Print_remoteendbanner (q_tmp, banner_end_raw);
		 fclose (banner_end_raw);
		 if (succ_endbanner != JSUCC) {
		     unlink ("banner_epg");
		    logerr (XLOG_INFO, "senddatafile: generating of endbannwer on file banner_epg failed");
		     return (JFAIL);
		 }
	    }
	    if (TR) {
	        tr_string = True;  
		(void) fclose (banner_end_raw);
		banner_end_raw = fopen_daemon ("banner_epg", "a");
		if (banner_end_raw == NULL) {
		    logerr_die (XLOG_INFO, "senddatafile: open file (a) banner_epg for appending banner failed!");
		 }
		status_epg = stat ("banner_epg", &st_epgbuf);
		if (status_epg < 0) {
		    logerr (XLOG_NOTICE, 
			    "senddatafile: fstat failed on file 'banner_epg'");
		    return (JABORT);
		}   
		(void) fprintf (banner_end_raw, TR);
		(void) fclose (banner_end_raw);
	    }
	}
    }
    raw = fdopen (dfd, "r");
    assert (raw != NULL);

    if (IF == 0 || *IF == 0) {	/* non-filtered -- very boring */
	DEBUG3 ("senddatafile: After IF=0: Values of: HL=%d, suppress_banner=%d, NB=%d banner_printed=%d", HL, suppress_banner, NB, banner_printed); 
        if ( (RB) && ((!SH) || (ld_string) || tr_string || short_banner) ) {
	/* constructing the datafile with banner and end-banner pages */
	     if ((LD || !suppress_banner || (files_per_job_sent == 1)) 
		 && (!error_occurred)) {
	         banner_raw = fopen_daemon ("banner_pg", "a");
		 if (banner_raw == NULL) {
		     logerr_die (XLOG_INFO, "senddatafile: open file banner_pg for appending banner failed!");
		 }
	         if (!banner_printed) {
		     banner_printed = True;
		     while ((c = fgetc (raw)) != EOF) {
		            (void) fputc (c, banner_raw);
		     }
		 }
		 fclose (banner_raw);
	     }
	     jobs_printed++;
	     jobs_prepared++;

	     if ( ((!short_banner) && (!no_endbanner) 
		   && (jobs_prepared == Parmcount) && (!error_occurred)) || TR) {
	          banner_end_raw = fopen_daemon ("banner_epg", "r");
		  if (banner_end_raw == NULL) {
		      logerr_die (XLOG_INFO, "senddatafile: open file banner_epg for reading banner failed!");
		  }
		  banner_raw = fopen_daemon ("banner_pg", "a");
		  if (banner_raw == NULL) {
		      logerr_die (XLOG_INFO, "senddatafile: open file banner_pg for appending banner failed!");
		  }
		  if (!banner_printed) {
		      banner_printed = True;
		      while ((c = fgetc (raw)) != EOF) {
			  (void) fputc (c, banner_raw);
		     }
		  }
		  while ((c = fgetc (banner_end_raw)) != EOF) {
		        (void) fputc (c, banner_raw);
		  }
		  fclose (banner_raw);
		  fclose (banner_end_raw);
	     } 

	     close (dfd);
	     DEBUG3 ("senddatafile: Before HL: Values of: HL=%d, suppress_banner=%d, NB=%d banner_printed=%d", HL, suppress_banner, NB, banner_printed); 
	     if ((HL) && (!suppress_banner) && (NB) && (!banner_printed) 
		 && (!error_occurred)) {
	          banner_raw = fopen_daemon ("banner_pg", "a");
		  if (banner_raw == NULL) {
		     logerr_die (XLOG_INFO, "senddatafile: open file banner_pg for appending banner failed!");
		  }
		  banner_end_raw = fopen_daemon ("banner_epg", "r");
		  if (banner_raw == NULL) {
		     logerr_die (XLOG_INFO, "senddatafile: open file banner_epg for reading banner failed!");
		  }
		  if (FF && *FF && !SF) {
		      (void) fputs (FF, banner_raw);
		  }
		  while ((c = fgetc (banner_end_raw)) != EOF) {
   	              (void) fputc (c, banner_raw);
		  }
		  (void) fclose (banner_raw);
		  (void) fclose (banner_end_raw);
		}
	   
	   unlink (name);
	   if ( ((!short_banner) && (!no_endbanner)) || (HL) ) {
	       unlink ("banner_epg");
	   }
	   if (rename ("banner_pg", name)) {
	       log (XLOG_CRIT, "senddatafile: renameing of datafile failed!");
	       banner_printed = False;
	       return (JFAIL);
	   } 
	   DBGREM3 ("senddatafile: Bannered job renamed from banner_pg to %s.", name);
	   dfd = open_as_submitter (name, O_RDONLY, 0);

	   if (dfd < 0) {
	     logerr (XLOG_INFO, "senddatafile: cannot open data file %s", name);
	     setstatus ("Cannot open data file %s!", name);
	     banner_printed = False;
	     return (JABORT);
	   }
	   
	   succ = fstat (dfd, &stbuf);
	   if (succ < 0) {
	     logerr (XLOG_NOTICE, "senddatafile: fstat failed '%s'", name);
	     setstatus ("Cannot fstat data file %s!", name);
	     banner_printed = False;
	     return (JABORT);
	   }
	   
	}
	succ = sendfile (raw, name, flag);
	fclose (raw);
	if (RB) { /* clearing... */
	    (void) unlink ("banner_pg");
	    (void) unlink ("banner_epg");
	}
	banner_printed = False;
	return (succ);
    }

    /* this remote printer has a text filter... thrills 'n' spills! */

    DEBUG3 ("senddatafile: After IF!=0: Values of: HL=%d, suppress_banner=%d, NB=%d banner_printed=%d", HL, suppress_banner, NB, banner_printed); 
    DBGREM1 ("senddatafile: IF: starting filter");
    filter = Setup_filter ('f', IF);
    (void) strcpy (tmpf, name);
    tmpf[0] = 't';		/* indicate tmp file */
    DBGREM1 ("senddatafile: IF: using temporary file %s", tmpf);

    /*
     * set up an fd, supposedly pointing to the printer, but really pointing to a
     * temporary output file.
     */
    if ((Print_fd = Open_tmpfile (tmpf)) < 0) {
	return (JFAIL);		/* already logged, just return */
    }
    
    DBGREM3 ("senddatafile: value of Print_fd (outputfile)= %d, raw (inputfile) = %d.",
	     Print_fd, fileno (raw)); 

    succ = Print_filter (fileno (raw), filter, False);	/* non-local printer mode */

    close (Print_fd);		/* close the output file */
    fclose (raw);		/* close the input file */

    /* if the filter failed, give up here. */
    if (succ != JSUCC) {
      	unlink (tmpf);
	setstatus ("IF filter failed!");
	setproctitle ("%s: if filter failed.", Printer);
	return (succ);
    }
    /* otherwise, send the filtered/bannered  output to the remote end */
    DEBUG3 ("senddatafile: After RB: Values of: HL=%d, suppress_banner=%d, NB=%d banner_printed=%d", HL, suppress_banner, NB, banner_printed); 

    if ((RB) && (!suppress_banner)) { /* "banner" the output file */
	 {
	     if (files_per_job_sent == 1) {
		 int c;
		 if (!banner_printed) {
		     banner_printed = True;
		     banner_raw = fopen_daemon ("banner_pg", "a");
		     if (banner_raw == NULL) {
			 logerr_die (XLOG_INFO, "senddatafile: open file banner_pg for appending banner failed!");
		     }

		     raw_filtered = fopen_daemon (tmpf, "r");
		     if (raw_filtered == NULL) {
			 logerr_die (XLOG_INFO, "senddatafile: open file '%s' for reading failed!", tmpf);
		     }

		     while ((c = fgetc (raw_filtered)) != EOF) {
			 (void) fputc (c, banner_raw);
		     }
		     close (Print_fd);
		     fclose (raw_filtered);
		     fclose (banner_raw);
		     if (rename ("banner_pg", tmpf)) {
			 log (XLOG_CRIT, 
			      "senddatafile: renameing of datafile failed!");
			 return (JFAIL);
		     }
		 }

		 DBGREM3 ("senddatafile: Bannered job renamed from 'banner_pg' to '%s'.", tmpf);
		 jobs_printed++;
		 jobs_prepared++;
	     }
	     if ( (!SB) && (files_per_job_sent == Parmcount) && (!error_occurred)) {
	          banner_end_raw = fopen_daemon ("banner_epg", "r");
		  if (banner_raw == NULL) {
		     logerr_die (XLOG_INFO, "senddatafile: open file banner_epg for reading banner failed!");
		 }
		  raw_filtered = fopen_daemon (tmpf, "a");
		  while ((c = fgetc (banner_end_raw)) != EOF) {
		      (void) fputc (c, raw_filtered);
		  }
		  fclose (banner_end_raw);
		  fclose (raw_filtered);
		  DBGREM3 ("senddatafile: Appended end banner to file %s.", tmpf);
	     }
	 }
    }

    DBGREM1 ("senddatafile: sending file %s, name %s, flag %d",
	     tmpf, name, flag);
    raw = fopen_daemon (tmpf, "r");
    succ = sendfile (raw, name, flag);
    unlink (tmpf);
    if (RB) { /* clearing... */
        (void) unlink ("banner_pg");
        (void) unlink ("banner_epg");
    }
    free (filter);
    banner_printed = False;
    return (succ);
}

/***************************************************************************
 * sendname (FILE *fp; char *name; int flag)
 *   sendname protocol:
 * 1. line containing the flag and file name is sent
 *     This has the format:
 *        "%c%d %s", flag, filesize, name
 *    DFILE (data file) flag is 03,  a CFILE (control file) flag is 02
 * 2. the remote end responds by sending back a single 0 character;
 *    anything else is an error condition and terminates transfer
 ***************************************************************************/
static int
sendname (FILE *fp, char *name, int flag) {
    struct stat statb;		/* status buffer */
    int succ;			/* ACME Integer, Inc. */
    char buf[BUFSIZ];

    if (Job_removed ()) {
	/* the current job has been removed; fake out the xfer */
	succ = Link_ack (1);	/* send "skip job" code */
	Link_close ();
	return (JSUCC);
    }

    /*
     * stat the file
     */
    if (fstat (fileno (fp), &statb) < 0) {
	logerr (XLOG_INFO, "sendname: could not stat %s", name);
	return (JFAIL);
    }

    (void) sprintf (buf, "%c%ld %s\n", flag, (long) statb.st_size, name);
    succ = Link_send (buf);
    if (succ == JSUCC) {
	succ = Link_confirm ();
    }
    return (succ);
}

static int
Print_remotebanner (struct plp_queue *q_tmp, FILE *banner_raw) {
    int jstatus;

    DBGREM3 ("Print_remotebanner: printing banner");
    if (Print_ready (True) != JSUCC) {
        return (JFAIL);
    }

    Print_fd = fileno (banner_raw);
    DBGREM3 ("Print_remotebanner: file descriptor: %d", Print_fd);
    if (BP && *BP) {
        if ((jstatus = Remote_Banner_program (BP, banner_raw)) != JSUCC) {
            log (XLOG_WARNING, "Print_rermotebanner: banner program \"%s\" failed", 
		 BP);
	    close (Print_fd);
            return (jstatus);
        }
    } else {
        if (HL) {
	   (void) remote_print_FF ("after job");
	}
        if ((jstatus = banner (Print_fd, q_tmp)) != JSUCC) {
	     close (Print_fd);
	     return (jstatus);
	}
    }
    (void) remote_print_FF ("before job");
    close (Print_fd);
    DBGREM3 ("Print_remotebanner: file descriptor: %d closed.", Print_fd);
    return (JSUCC);
}

static int
Print_remoteendbanner (struct plp_queue *q_tmp, FILE *banner_end_raw) {
    int jstatus;

    DBGREM3 ("Print_remoteendbanner: printing endbanner");
    if (Print_ready (True) != JSUCC) {
        return (JFAIL);
    }
    Print_fd = fileno (banner_end_raw);
    DBGREM3 ("Print_remoteendbanner: file descriptor: %d", Print_fd);
    if (!HL) {
       (void) remote_print_FF ("after job");
    }
    if (EP && *EP) {
        if ((jstatus = Remote_EndBanner_program (EP, banner_end_raw)) != JSUCC) {
             log (XLOG_WARNING, "Print_rermoteendbanner: endbanner program \"%s\" failed", 
		  EP);
	     close (Print_fd);
	     return (jstatus);
        }
    } else {
        if ((jstatus = endbanner (Print_fd)) != JSUCC) {
	     close (Print_fd);
	     return (jstatus);
	}
    }
    if (HL) {
       (void) remote_print_FF ("before job");
    }
    close (Print_fd);
    DBGREM3 ("Print_remotebanner: file descriptor: %d closed.", Print_fd);
    return (JSUCC);
}

/****************************************************************************
 * Print the banner using the user specified program (remote banner printing)
 * 1. setup the filter
 * 2. start program
 * 3. wait for completion.
 ****************************************************************************/
static int
Remote_Banner_program (char *prog, FILE *banner_raw) {
    char *filter;
    int status;

    filter = Setup_filter ('f', prog);
    DBGLOCAL3 ("Banner_program: filter %s", filter);
    status = Print_filter (Print_fd, filter, False);
#ifndef IS_LINUX
    (void) free (filter);
#endif
    return (status);
}

/*******************************************************************************
 * Print the endbanner using the user specified program (remote banner printing)
 * 1. setup the filter
 * 2. start program
 * 3. wait for completion.
 *******************************************************************************/
static int
Remote_EndBanner_program (char *prog, FILE *banner_end_raw) {
    char *filter;
    int status;

    filter = Setup_filter ('f', prog);
    DBGLOCAL3 ("EndBanner_program: filter %s", filter);
    status = Print_filter (Print_fd, filter, False);
#ifndef IS_LINUX
    (void) free (filter);
#endif
    return (status);
}

static int
remote_print_FF (char *current_state) {
    int jstatus = JSUCC;		/* job status */

    if (FF && *FF && !SF) {
	DBGLOCAL3 ("remote_print_FF: printing FF %s", current_state);
	jstatus = Print_string (FF);
	if (jstatus != JSUCC) {
	    log (XLOG_INFO, "printing FF %s failed", current_state);
	}
    }
    return jstatus;
}
