/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1987, 1988, Patrick Powell
 * version 3.3.0 Justin Mason July 1994
 ***************************************************************************
 * MODULE: sendmail.c
 * send mail to user on completion or on errors of a job
 ***************************************************************************/

/* to get the real name */
#include <pwd.h>


#include "lp.h"
#include "library/errormsg.h"

extern void Make_mail_address (char *, char *, char *, char *);

/* provide a buffer where common user errors can be caught and described
 * in a "pretty" form -- eg. "you don't have permission!".
 */
char Mail_log_message[1024] = "";

static void
mail_error (FILE *mail) {
    FILE *error_fp;
    char buf[BUFSIZ];
    struct stat stb;

    if (!Errors_file || !*Errors_file)
	return;		/* we're not using it */

    if (stat (Errors_file, &stb) != 0)
	return;		/* not there, probably */

    if (stb.st_size < 1)
	return;		/* empty file */

    if ((error_fp = fopen_daemon (Errors_file, "r")) == NULL)
	return;		/* can't open */

    fprintf (mail, "\nFilter errors:\n");
    while (fgets (buf, sizeof (buf), error_fp)) {
	fputs (buf, mail);
    }

    fclose (error_fp);

    if (unlink_daemon (Errors_file) != 0)
	log (XLOG_INFO, "couldn't unlink errors file '%s'", Errors_file);
}

/*
 * sendmail --- tell people about job completion 1. fork a sendmail process 2. if
 * successful, send the good news 3. if unsuccessful, send the bad news
 */
void
sendmail (struct plp_queue *q, int status) {
    static int p[2];		      /* pipe */
    int c;			      /* ACME Integer, Inc. */
    long i;			      /* ACME Integer, Inc. */
    char buf[100];		      /* command line to run */
    char toaddr[511];		      /* To: address */
    char tmp_buf[32];                 /* hostname for mail */
    char Mail_from_address_host[256]; /* sender including hostname */
    FILE *mail;			      /* mail file */
    pid_t pid;			      /* sendmail process */
    plp_status_t statb;		      /* daughter status */
    struct passwd *pwent;             /* to get the real username */
    char username[64];                /* holds the username */
    char hostname[128];


    /* in this cases we have to make the mail addresse by ourself...*/
    DEBUG3 ("sendmail: Value of job_moved %d", job_moved);
    if ((mail_to_user_for_remote_remove) || (called_from_lprm) || (job_moved)) {
	(void) strcpy (hostname, &q->q_name[6]);
        Make_mail_address (MAILNAME, Mail_addressing, q->q_user, hostname);
    }
    DEBUG3 ("sendmail: MAILNAME = %s", (MAILNAME[0]) ? MAILNAME : "unknown");
    DEBUG3 ("sendmail: FROMHOST = %s", FROMHOST);

    /*
     * check to see if the user really wanted a "your file was printed ok" message
     */
    if (!MAILNAME[0]) {
	if (status == JSUCC) {
	    DEBUG3 ("sendmail: success mail is unwanted");
	    return;
	} else if (!Mail_on_errors) {
	    DEBUG3 ("sendmail: failure mail is unwanted");
	    return;
	}
    }
    if (!Mail_command || !*Mail_command) {
	DEBUG3 ("sendmail: mail is turned off");
	return;
    }
    (void) strcpy (buf, Mail_command);
    DEBUG3 ("sendmail: mail command: %s", buf);

    /* clearing the array "username" */
    for (i = 0; i < 64; i++) {
        username[i] = '\0';
    }

    if (MAILNAME && MAILNAME[0]) {
	/* lpr provided it -- thanks lpr! */
	(void) strcpy (toaddr, MAILNAME);
	(void) strncpy (username, q->q_user, sizeof (username) - 1);
    } else {
	/* we have to make it ourselves and hope for the best */
	if (q && q->q_user) {
	    Make_mail_address (toaddr, Mail_addressing, q->q_user, &(q->q_from));
	    (void) strncpy (username, q->q_user, sizeof (username) - 1);
	} else {
	    Make_mail_address (toaddr, Mail_addressing, LOGNAME, FROMHOST);
	    (void) strncpy (username, LOGNAME, sizeof (username) - 1);
	}
    }
    username[sizeof(username)-1] = '\0';

    DEBUG4 ("sendmail: sendmail to: %s", toaddr);

    if (pipe (p) < 0) {
	logerr_die (XLOG_NOTICE, "pipe failed for sendmail");
    }

    /*
     * start up sendmail process
     */
    if ((pid = fork ()) == 0) {
	if (p[0]) {
	    if (dup2 (p[0], 0) < 0) {
		logerr_die (XLOG_NOTICE, "dup2 failed in sendmail child");
	    }
	}
	/* run as "daemon" user so headers are legal */
	full_daemon_perms ();

	setup_close_on_exec (); /* this'll close the unused pipe fds */
	secure_exec (buf);

    } else if (pid < 0) {	/* parent */
	logerr_die (XLOG_NOTICE, "sendmail: fork failed");
    }
    (void) close (p[0]);
    if ((mail = fdopen (p[1], "w")) == NULL) {
	logerr_die (XLOG_NOTICE, "sendmail: fdopen failed");
    }

    /* write the mail headers here. */
    (void) sprintf (tmp_buf, "%s", ShortHost);
    c = toupper ((int) tmp_buf[0]);
    tmp_buf[0] = (char) c;
    (void) sprintf (Mail_from_address_host, "%s", Mail_from_address);
    c = strlen (Mail_from_address_host) - 1;
    Mail_from_address_host[c] = '\0';
    (void) strcat (Mail_from_address_host , " from ");
    (void) strcat (Mail_from_address_host, tmp_buf);
    (void) strcat (Mail_from_address_host, ")");
    (void) fprintf (mail, "From: %s\n", Mail_from_address_host);
    if (status != JSUCC) {
	/* Cc the printer operator (via .forward or alias) in case
	 * of printer error.
	 */
        DEBUG3 ("sendmail: status = %d, CC will be sent to operator", status);
	(void) fprintf (mail, "Cc: %s\n", Mail_from_address);
    }
    (void) fprintf (mail, "To: %s\n", toaddr);
    (void) fprintf (mail, "Subject: Your '%s' printer job %d [%s]\n\n", 
		    Printer, q->q_num, q->q_data->name);

    /* now do the message */

    if (*Mail_log_message != '\0') {
	(void) fprintf (mail, "%s\n", Mail_log_message);
	*Mail_log_message = '\0';

    } else {
        pwent = getpwnam (username);
	if (pwent != NULL) {
	    (void) sprintf (username, pwent->pw_gecos);
	}
	(void) fprintf (mail, "Message vom PLP printer spooling system.\n\nDear user %s,\nyour %s printer job", username, Printer);
	if (q) 
	    (void) fprintf (mail, " %d", q->q_num);
	
	switch (status) {
	case JSUCC:
	    (void) fprintf (mail, " has been printed successfully.\n");
	    break;

	case JFAIL:
	    (void) fprintf (mail, " failed, and retry count was exceeded.\n");
	    goto the_msg_was;

	case JABORT:
	    (void) fprintf (mail, " failed and could not be retried.\n");
the_msg_was:
	    (void) fprintf (mail, "\nCause of death: %s\n", Last_errormsg);
	    mail_error (mail);
	    break;

	default:
	    (void) fprintf (mail, " died a horrible death.\n");
	    break;
	}

	if (q) {
	    int idx;
            char sbuf[64];
	    double dummy = 0.0;

	    if (q->q_size < 1024) {
		(void) sprintf (sbuf, "%ld Bytes", q->q_size);
	    } else if (q->q_size < (1024 * 1000)) {
	               dummy = (double) q->q_size / 1024.0;
		       (void) sprintf (sbuf, "%.3f KBytes", dummy); 
	    } else if (q->q_size < (1024 * 1000 * 1000)) {
	               dummy = (double) q->q_size / (1024.0 * 1000.0);
		       (void) sprintf (sbuf, "%.3f MBytes", dummy);
            } else {
	               dummy = (double) q->q_size / (1024.0 * 1000000.0);
		       (void) printf (sbuf, "%.3f GBytes", dummy); 
	    } 
	    (void) fputs ("\nFiles:\n", mail);

	    for (idx = 0; idx < q->q_entries; idx++) {
		(void) fputc('\t', mail);
		if (q->q_data[idx].copies > 1) {
		    (void) fprintf (mail, "  %d copies of ",
				    q->q_data[idx].copies);
		}
		(void) fprintf (mail, "%s,\t Size: %s\n", 
				q->q_data[idx].name, sbuf);
	    }
	}
    }

    (void) fflush (mail);
    (void) fclose (mail);
    i = plp_waitpid (pid, &statb, WUNTRACED);
    DEBUG3 ("sendmail: %d finished (%s)", i, Decode_status (&statb));
}
