/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1987, 1988, Patrick Powell
 ***************************************************************************
 * MODULE: lpr.c
 * lpr - off line print
 **************************************************************************/
#include "library/errormsg.h"
#include "library/readconf.h"

#include "lpr.h"
#include "lpr_canprint.h"
#include "lpr_filters.h"
#include "lpr_global.h"
#include "lpr_job.h"
#include "lpr_parms.h"
#include "lpr_temp.h"
#include "patchlevel.h"

extern void Displayq (int, int);

int u_option;
int copy_message = 0;
int be_quiet = 0;

/***************************************************************************
 * SYNOPSIS
 *      lpr [ -PPrinter ] [ -#num ] [ -C class ] [ -N ] [ -J job ] [
 *      -RremoteAccount ] [ -m[mailTo] ] [ -T title ] [-i[numcols]]
 *      [ -1234 font ] [ -wnum ] [ -Zzoptions ] [ -Uuser ] [ -HHost
 *      ] [ -Ffilter ] [ -bhrs ] [ -Dn ] [ -S ] [ filename ...  ]
 * DESCRIPTION
 *      Lpr uses a spooling server to print the named files when
 *      facilities become available.  If no Names appear, the stan-
 *      dard input is assumed.
 *      -#num number of copies of each file to be printed.
 *      -C class or priority (A - Z)
 *      -D[n] debug level
 *      -F?  Filter or format specification
 *      -HHost Used by root process to specify a Host
 *      -J jobname  Specify the job name to print on the burst page
 *      -N spool the job, but don't start a server
 *      -PPrinter  Output to the specified Printer
 *      -R remoteAccount
 *      -S copies the job to the "save"-area
 *      -T title specify the title used by pr(1);
 *      -Uuser Used by root process to specify a user
 *      -Z extra options
 *      -b The files are assumed to contain binary data
 *      -c data produced by cifplot(l).
 *      -d output from tex(l) (DVI format from Stanford).
 *      -f same as -Fr, FORTAN carriage control
 *      -g standard plot data as produced by the plot(3X) routines
 *      -i[numcols] Cause the output to be indented
 *      -l (literal) text with control characters to be printed
 *      -m[mailTo] Send mail upon completion
 *      -n output from DITROFF
 *      -p text to be printed using pr(1) to format the files.
 *      -q don't chatter about the jobs spooled in queue
 *      -s Use symbolic links.
 *      -t output from troff(1)
 *      -v a raster image for devices like the Benson Varian.
 *      -wwidth  specify the page width for pr.
 ****************************************************************************
 * Implementation:
 * 	Each time lpr is invoked it creates a "job" entry in the appropriate
 * spool directory.  Each job consists of a control file and zero or more
 * data files.  The control file contains lines that specify the job
 * parameters, such as job Name, etc., and the Names of the data files.
 *      Control file format
 *      First character in the line flags kind of entry, remainder of line is
 *          the argument.
 *
 *		C -- "class Name" on banner page
 *		H -- "Host Name" of machine where lpr was done
 *		I -- "indent" amount to indent output
 *		J -- "job Name" on banner page
 *		L -- "literal" user's Name to print on banner
 *		M -- "mail" to user when done printing
 *		N -- "Name" of file (used by lpq)
 *		P -- "Person" user's login Name
 *		R -- account id  for charging
 *		U -- "unlink" Name of file to remove after we print it
 *		W -- "width" page width for PR
 *		X -- "header" header title for PR
 *		Z -- xtra options to filters
 *
 *	Lower case letters are formats, and are together with the Name
 *      of the file to print
 *		c -- "file Name" cifplot file to print
 *		d -- "file Name" dvi file to print
 *		f -- "file Name" Name of text file to print
 *		g -- "file Name" plot(1G) file to print
 *		l -- "file Name" text file with control chars
 *		n -- "file Name" ditroff(1) file to print
 *		p -- "file Name" text file to print with pr(1)
 *		t -- "file Name" troff(1) file to print
 *		v -- "file Name" plain raster file to print
 ***************************************************************************/

/***************************************************************************
 * remove the temp files
 ***************************************************************************/

plp_signal_t 
cleanup (int signal) 
{
    plp_block_signals ();
    Remove_temp ();
    exit (1);
}

/***************************************************************************
 * 1. get the Host computer Name and user Name
 * 2. get the parameters and other information
 * 3. set up signals to handle abrupt termination.
 * 4. set up the control file for the job
 * 5. if we are spooling from stdin, copy stdin to a file.
 * 6. if we have -p option (use pr), run the input files through pr
 *    and save to a single file.
 * 7. create a job entry
 * 8. start the server
 ****************************************************************************/

int 
main (int argc, char *argv[], char *envp[])
{
    FILE *cfp_save;             /* File pointer used for save copies */
    FILE *afp;                  /* File pointer for attached printers */
    struct passwd *pw_ent;      /* user entry in /etc/passwd */
    int i, j, fd, job_size = 0;
    int Job_number_save = 0;
    char *dest_SD, *src_SD;     /* Temp pointer for spool dirs */
    char s_srcname[BUFSIZ];	/* for the full pathname */
    char s_newname[BUFSIZ];	/* full pathname of new name  */
    char *orig_Printer;
    char attached_Printer[64];  /* name of the attached printer */
    char attached_SD[BUFSIZ];   /* name of the spool dir and the attache file */
    char afname[MAXPARMLEN];	/* attach file name */
    short printer_is_attached = 0;
    struct stat *save_buf;
    char dup_save_cfname[255];
    char cf_buffer[255];
    char new_cf_buffer[10*BUFSIZ]; /* must be big enough to hold all the lines! */
    char Job_number_char[4];
    char Job_number_save_char[4];

    u_option = 0;
    set_user_uid (0);
    root_to_user ();

    Get_parms (argc, argv);	/* read the switches, set debug */
    Std_environ (argc, argv, envp);
    Setup_waitpid ();
    Readconf ();

    FQDN = Host;		/* set up the From information */
    From = ShortHost;

    /* get user name */
    if ((pw_ent = getpwuid (UserUID)) == 0) {
	logerr_die (XLOG_INFO, "getpwuid failed on uid %d", UserUID);
    }
    (void) strcpy (LOGNAME, pw_ent->pw_name);
    Person = LOGNAME;

    if (!UserUID) {
	Is_root = 1;		/* we are being invoked by root */
    }

    /* set up chatty information for user */
    if (BNRNAME[0] == '\0') {
	/*
	 * Add extra check - gecos field may be empty - if so stick with username.  
	 * What we want to avoid is L field being unset - this means 
	 * no banner to BSD systems.
	 */
	if (strsame (LOGNAME, pw_ent->pw_name) && (pw_ent->pw_gecos[0])) {
	    char *cp;
	    (void) strncpy (BNRNAME, pw_ent->pw_gecos, MAXPARMLEN);
	    BNRNAME[MAXPARMLEN] = '\0';
	    if ((cp = strchr(BNRNAME, ',')) != NULL) {
		*cp = '\0'; if (BNRNAME[0] == '\0') {
		    (void) strcpy (BNRNAME, LOGNAME);
		}
	    }
	} else {
	    (void) strcpy (BNRNAME, LOGNAME);
	}
    }


    Setup_parms ();			/* setup parameters */

    /* Check to see if the printer is attached to another one;
     * in this case we submit the job to that queue.
     */

    (void) sprintf (attached_SD, "%s/", SD);
    (void) strcat (attached_SD, Attach_file);
    if (Attach_file && *Attach_file && ((afp = fopen_daemon (attached_SD, "r")))) {
       if (fscanf (afp, "%s", afname) == 1) {
	    if (strsame (afname, Printer)) {
		fatal (XLOG_INFO, "Printer '%s' attached to itself", Printer);
	    }
	    (void) strcpy (attached_Printer, Printer);
	    Printer = strdup (afname);
	    fclose (afp);
	    Setup_parms ();			/* re-setup parameters */
	    printer_is_attached = 1;
       } else {
	    fatal (XLOG_INFO, 
		    "attach file for printer '%s' corrupted -- remove by hand!",
		    Printer);
	} 
       DEBUG2 ("lpr: printer %s reattached to %s",
		attached_Printer, Printer);
       fclose (afp);
    }


    (void) plp_signal (SIGPIPE, (plp_sigfunc_t)SIG_IGN);	/* set signals */
    (void) plp_signal (SIGHUP, cleanup);
    (void) plp_signal (SIGINT, cleanup);
    (void) plp_signal (SIGQUIT, cleanup);
    (void) plp_signal (SIGTERM, cleanup);
    /**************************************************************************
     * The following possibilities exist:
     *                         Files              No Files
     * p format,no 'f' PF      PR ->spool         input->PR -> spool
     * p format, 'f' PF        PR->PF->spool      input->PR->PF->spool
     * ? format, no PF         files->spool files input->spool
     * ? format, PF            PF(files) -> file  input->spool
     *
     * If we have a prefilter (PF) for a file format,  we have to run the files
     * through the prefilter,  and spool the output of the prefilter.
     * The prefilter will create a single output file which is printed
     *
     * If we have 'p' format,  we will create a printer process, and run
     * its output into a temporary file in the spool directory.
     * If there are no names specified, we copy the input to a temporary
     * file in the spool directory.  The Read_stdin file is the name of the
     * output file.  This is also used to hold the name of the output of
     * the PR program.
     *
     * If we have a prefilter,  we place its output in the Read_filter
     * file.
     *************************************************************************/

    pr_used = False;
    if (Format == 'p') {
	/*
	 * run PR and put the output in a spool file
	 */
	    DEBUG3 ("lpr: using PR");
	if (Run_pr () == 0) {
	    Diemsg ("nothing to print");
	}
	Format = 'f';
	pr_used = True;

    } else if (Parmcount == 0) {
        if (Prefilter_name[Format - 'a'] == NULL) {
		DEBUG3 ("lpr: Copying stdin");
	    if (Copy_stdin () == 0) {
		Diemsg ("nothing to print");
	    }
	}

    } else {
	/*
	 * check to see that the input files are printable
	 */
	job_size = 0;
	for (i = 0; i < Parmcount; ++i) {
	    fd = Is_printable (Parms[i].str, &LO_statb);
	    if (fd < 0) {
		Parms[i].str = NULL;
	    } else {
		job_size = 1;
	    }
	    (void) close (fd);
	}
	if (job_size == 0) {
	    Diemsg ("nothing to print");
	}
    }

    /*
     * now we check for prefilter; if we have one, then we have to run all the files
     * through it; we invoke the prefilter with the appropriate parameters. The output of
     * the prefilter (on stdout) is spooled
     */
    if (Prefilter_name[Format - 'a']) {
	/* yes, we have prefilter */
	    DEBUG3 ("lpr: Prefilter %c '%s'", Format,
		 Prefilter_name[Format - 'a']);

        if (Parmcount == 0 || Pr_stdin) {	/* also if pr has been run */
	    if (Do_prefilter (Prefilter_name[Format - 'a'], NULL) <= 0) {
	        job_size = 0;
	    } else {
	        job_size = 1;
	    }

	} else {
	    for (i = 0; i < Parmcount; i++) {
		if ((Parms[i].str == NULL) ||
			(Do_prefilter (Prefilter_name[Format - 'a'], &(Parms[i])) <= 0))
		{
		    Parms[i].str = NULL;
		} else {
		    job_size = 1;
		}
	    }
	}
	if (job_size == 0) {
	    Diemsg ("nothing to print");
	}
    }
    /*
     * we now either have to spool a single file, produced by the PR and/or Prefilters,
     * or we have to spool all the input files
     */
    Make_job ();

    if (Save_copy) {
        Save_copy = False;
	copy_message = True;
	orig_Printer = Printer;
	Printer = SaveCopyArea;     /* get the name of the "save"-area */
	malloc_or_die (src_SD, strlen (SD));  /* get some memory... */
	(void) strcpy (src_SD, SD); /* save the name of the spool dir */ 
	Get_Printer (1);            /* check "save"-area and spool dir */ 
	dest_SD = SD;               /* save the new name */
	s_srcname[0] = '\0';
	s_newname[0] = '\0';
	(void) strcpy (s_srcname, src_SD);
	(void) strcat (s_srcname, "/");
	(void) strcat (s_srcname, save_cfname);
	(void) strcpy (s_newname, dest_SD);
	(void) strcat (s_newname, "/");
	(void) strcat (s_newname, save_cfname);
	malloc_or_die (save_buf, sizeof (save_buf) * sizeof (long));
	if (stat (s_newname, save_buf) < 0 ) { /* this file doesn't exist, 
						* so it can be copied under the 
						* same job-id
						*/
	    user_to_daemon ();
	    if (copy_file (s_srcname, s_newname, -1) < 0) {
	        daemon_to_user ();	/* Switch back before exit */
		logerr_die (XLOG_INFO, 
			    "lpr: copy %s to %s failed", s_srcname, s_newname);
	    }
	    /* 
	     * input was stdin because this pointer has got his value in lpr_job.c 
	     */
	    if (save_dfname) {
	        s_srcname[0] = '\0';
		s_newname[0] = '\0';
		(void) strcpy (s_srcname, src_SD);
		(void) strcat (s_srcname, "/");
		(void) strcat (s_srcname, save_dfname);
		(void) strcpy (s_newname, dest_SD);
		(void) strcat (s_newname, "/");
		(void) strcat (s_newname, save_dfname);
		 if (copy_file (s_srcname, s_newname, -1) < 0) {
		     daemon_to_user ();	/* Switch back before exit */
		     logerr_die (XLOG_INFO, "lpr: copy %s to %s failed", 
				 s_srcname, s_newname);
		 }
	    } else {   
	        for (j = 0; j < Parmcount; j++) {
	             s_srcname[0] = '\0';
		     s_newname[0] = '\0';
		     (void) strcpy (s_srcname, src_SD);
		     (void) strcat (s_srcname, "/");
		     (void) strcat (s_srcname, Parms[j].filename);
		     (void) strcpy (s_newname, dest_SD);
		     (void) strcat (s_newname, "/");
		     (void) strcat (s_newname, Parms[j].filename);
		     if (copy_file (s_srcname, s_newname, -1) < 0) {
		         daemon_to_user ();	/* Switch back before exit */
			 logerr_die (XLOG_INFO, "lpr: copy %s to %s failed", 
				     s_srcname, s_newname);
		     }
		}
	    }
	    daemon_to_user ();
	    /* 
	     * switch back to the original spool dir and printer 
	     */
	    SD = src_SD; 
	    Printer = orig_Printer;
	    Job_number_save = Job_number;
	} else { /* a file with the same job-id exists, so we have to rename 
		  * the file(s) that are to be copied and 
		  * then rewrite the control-file with the correct 
		  * data of the data file
		  */
   	    if (Job_number == 999) {
	        Job_number_save = 1;
	    } else {
	        Job_number_save = Job_number + 1;
	    }
	    (void) strncpy (dup_save_cfname, save_cfname, 3);
	    (void) sprintf (&(dup_save_cfname[3]), 
			    "%03d%s", Job_number_save, &(save_cfname[6]));
	    s_newname[strlen (s_newname) - strlen (save_cfname)] = '\0';
	    (void) strcat (s_newname, dup_save_cfname);
	    while (stat (s_newname, save_buf) > 0)
	      {
		Job_number_save++;
		if (Job_number_save > 999) {
		    Job_number_save = 1;
		}
		(void) sprintf (&(dup_save_cfname[3]), "%03d%s", 
				Job_number_save, &(save_cfname[6]));
		s_newname[strlen (s_newname) - strlen (save_cfname)] = '\0';
		(void) strcat (s_newname, save_cfname);
	      }
	    user_to_daemon ();
	    if (copy_file (s_srcname, s_newname, -1) < 0) {
	        daemon_to_user ();	/* Switch back before exit */
		logerr_die (XLOG_INFO, "lpr: copy %s to %s failed", 
			    s_srcname, s_newname);
	    }
	    /* Now the contents of the cf file must be 
	     * adjusted to the new seq number: 
	     * 1. read the file
	     * 2. check the old sequence number and change it
	     * 3. rewrite the cf file
	     */
	    if ((cfp_save = fopen_daemon (s_newname, "r")) == NULL) {
	         logerr_die (XLOG_INFO, "lpr: cannot open file %s for reading.");
	    }
	    user_to_daemon ();
	    (void) sprintf (Job_number_char, "%03d", Job_number);
	    new_cf_buffer[0] = '\0';
	    while (fgets ((char *) cf_buffer, 128, cfp_save) != NULL)
	      {
		if (strstr (cf_buffer, Job_number_char) != 0 && 
		    strstr (cf_buffer, Host) != 0) {
		    (void) sprintf (Job_number_save_char, "%03d", Job_number_save);
		    cf_buffer[4] = Job_number_save_char[0];
		    cf_buffer[5] = Job_number_save_char[1];
		    cf_buffer[6] = Job_number_save_char[2];
		}
		strcat (new_cf_buffer, cf_buffer);
	      }
	    (void) fclose (cfp_save);
	    if ((cfp_save = fopen_daemon (s_newname, "w")) == NULL) {
	         logerr_die (XLOG_INFO, "lpr: cannot open file %s for reading.");
		 daemon_to_user (); /* switch back */
	    } else {
	         fputs (new_cf_buffer, cfp_save);
		 user_to_daemon ();
		 (void) fclose (cfp_save);
		 daemon_to_user ();
	    }
	    /* now copy the datafile(s) */
	    if (save_dfname) {
                s_srcname[0] = '\0';
                s_newname[0] = '\0';
                (void) strcpy (s_srcname, src_SD);
                (void) strcat (s_srcname, "/");
                (void) strcat (s_srcname, save_dfname);
                (void) strcpy (s_newname, dest_SD);
                (void) strcat (s_newname, "/");
		save_dfname[3] = Job_number_save_char[0]; 
		save_dfname[4] = Job_number_save_char[1]; 
		save_dfname[5] = Job_number_save_char[2]; 
                (void) strcat (s_newname, save_dfname);
		user_to_daemon ();
		if (copy_file (s_srcname, s_newname, -1) < 0) {
		    daemon_to_user (); /* Switch back before exit */
		    logerr_die (XLOG_INFO, "lpr: copy %s to %s failed", 
				s_srcname, s_newname);
		}
            } else {   
                for (j = 0; j < Parmcount; j++) {
                     s_srcname[0] = '\0';
                     s_newname[0] = '\0';
                     (void) strcpy (s_srcname, src_SD);
                     (void) strcat (s_srcname, "/");
                     (void) strcat (s_srcname, Parms[j].filename);
                     (void) strcpy (s_newname, dest_SD);
                     (void) strcat (s_newname, "/");
		     Parms[j].filename[3] = Job_number_save_char[0];
		     Parms[j].filename[4] = Job_number_save_char[1];
		     Parms[j].filename[5] = Job_number_save_char[2];
                     (void) strcat (s_newname, Parms[j].filename);
		     user_to_daemon ();
                     if (copy_file (s_srcname, s_newname, -1) < 0) {
                         daemon_to_user ();     /* Switch back before exit */
                         logerr_die (XLOG_INFO, "lpr: copy %s to %s failed", 
                                     s_srcname, s_newname);
                     }
                }
	    }
	    daemon_to_user ();
	}
    }

    if (!No_start_server) {
        /* start up the server */
        Startserver ();
    }

    (void) sleep (1);
    called_by_lpr = 1;
    Job_number_by_lpr = Job_number;
    Displayq (0, 0);
    if (rank_for_lpr[0] == '\0') {
        (void) sprintf (rank_for_lpr, "%s", "active");
    }
    called_by_lpr = 0;

    /* inform the user, if -q option (be_quiet) is not set */
    if (! be_quiet) {
	if (printer_is_attached) {
	    (void) fprintf (stdout, 
			    "Your print job was sent to '%s' [attached to '%s'], job %d, rank: %s", 
			    attached_Printer, Printer, Job_number, rank_for_lpr);
	} else {
	    (void) fprintf (stdout, 
			    "Your print job has been sent to printer '%s', job %d, rank: %s", 
			    Printer, Job_number, rank_for_lpr);
	}
	if (copy_message) {
	    (void) fprintf (stdout, 
			    "\nand besides copied to the '%s' save area, job %d.\n", 
			    SaveCopyArea, Job_number_save);
	} else {
	    (void) fprintf (stdout, ".\n");
	}
    }
    (void) fflush (stdout);
    return False;
}
