/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1987, 1988, Patrick Powell
 * version 3.3.0 Justin Mason July 1994
 ***************************************************************************
 * MODULE: lpr_job.c
 * Submit job in the spool directory
 ***************************************************************************/

#include "lpr.h"
#include <assert.h>
#include "library/errormsg.h"
#include "lpr_canprint.h"
#include "lpr_filters.h"
#include "lpr_global.h"
#include "lpr_job.h"
#include "lpr_parms.h"
#include "lpr_temp.h"

#ifdef NIS
#include <rpc/rpc.h>
#include <rpcsvc/ypclnt.h>
#include <rpcsvc/yp_prot.h>
#endif

/***************************************************************************
 * Get_fullpath(): get the current directory
 * this fails if the user cannot read the cwd, so don't call it unless
 * it's necessary. SECURITY: Make sure the environment has been sanitised,
 * otherwise this can give rise to a serious security hole!!!
 ***************************************************************************/

static char *Get_fullpath (void)
{
    extern char *getwd (char *);
    static char cwd_dir[MAXPATHLEN] = "";
    /*
     * get current working directory
     */
    if (*cwd_dir) {
	return (cwd_dir);
    }
    if (getwd (cwd_dir) == 0) {
	logerr_die (XLOG_INFO, "Get_fullpath: getwd failed");
    }
    return (cwd_dir);
}

/***************************************************************************
 * return the full path name
 ***************************************************************************/

static char *Full_path (char *file) {
    static char full[MAXPATHLEN];
    static char no_automount[MAXPATHLEN];
    struct stat stbuf1, stbuf2;
    int st_result;

    assert(file!=(char*)0);

    if (file[0] == '/') {
        assert((strlen(file) + 1) <= sizeof(full));
	(void) strcpy (full, file);

    } else {
	char *home;

	home = Get_fullpath ();
        assert((strlen(home) + 2 + strlen(file)) <= sizeof(full));	    
	(void) sprintf (full, "%s/%s", home, file);
    }

    /*
     * if this fails, there's a problem with the code; we
     * should NOT be calling Get_fullpath() after
     * changing the directory. Don't put the stat call
     * in the assert, otherwise it won't happen if we're
     * compiled with -DNDEBUG.
     */
    st_result = stat (file, &stbuf2); assert (st_result >= 0);

    /* try and be a little automounter-conscious -- this will
     * help in case the filesystem is unmounted by the automounter
     * between spool-time and print-time.
     */
    if (strncmp (full, AUTOMOUNT_STR, sizeof(AUTOMOUNT_STR) - 1) == 0) {
      (void) sprintf (no_automount, "%s%s",
#ifdef AUTOMOUNT_SUBST
                     AUTOMOUNT_SUBST,
#else
                     "/",
#endif
                     full + sizeof(AUTOMOUNT_STR) - 1);

        if (stat (no_automount, &stbuf1) != -1) {
            /* file exists on this path -> check inode */
            if ((stbuf1.st_ino == stbuf2.st_ino)
		    && (stbuf1.st_dev == stbuf2.st_dev))
	    {
		/* it's valid, return the stripped string */
		return (no_automount);
	    }
	}
    }
    return (full);
}

/***************************************************************************
 * makes a link to the target, and puts out entry
 ***************************************************************************/

static char *Make_link (char *target, struct stat *statb)
{
#ifdef NO_SYMLINKS
    fatal (XLOG_DEBUG, "no symbolic links allowed for security reasons");

#else				/* NO_SYMLINKS */
    char *lf;			/* link file */
    char buf[MAXPATHLEN];
    int i;

    assert(target!=(char*)0);
    assert(statb!=(struct stat*)0);
    assert(SD!=(char*)0);
    /*
     * get temp file name
     */
    lf = Get_tmp_data ();
    (void) sprintf (buf, "%s/%s", SD, lf);
    /*
     * generate absolute path name
     */
    target = Full_path (target);
	DEBUG3 ("Make_link: from %s to %s", buf, target);

    user_to_daemon (); i = symlink (target, buf); daemon_to_user ();

    if (i < 0) {
	logerr_die (XLOG_INFO, "Make_link: symbolic link from %s to %s failed",
		    buf, target);
    }
    if (stat (target, statb) < 0) {
	logerr_die (XLOG_INFO, "Make_link: stat failed '%s'", target);
    }
    return (lf);
#endif				/* NO_SYMLINKS */
}

/***************************************************************************
 * stat the file in the specified directory
 ***************************************************************************/

static void 
Get_stat (char *dir, char *file, struct stat *statb)
{
    static char buf[BUFSIZ];	/* for holding the pathname */

    assert(file!=(char*)0);
    assert(statb!=(struct stat*)0);
    if (dir && *dir) {
        assert((strlen(dir)+2+strlen(file))<=sizeof(buf));
	(void) sprintf (buf, "%s/%s", dir, file);
    } else {
        assert((strlen(file)+1)<sizeof(buf));
	(void) strcpy (buf, file);
    }
    if (stat (buf, statb) < 0) {
	logerr_die (XLOG_INFO, "Get_stat: cannot stat %s", buf);
    }
    DEBUG5 ("Get_stat: %s is %ld", buf, (long) statb->st_size);
}

/***************************************************************************
 * Prints a line out in the control file
 ***************************************************************************/

static void
Entry (char c, char *cp, int fd) {
    int buflen;
    char buf[MAXPARMLEN + 1];

    assert(cp!=(char*)0);
    DEBUG4 ("Entry: %c '%s'", c, cp);
    if (strchr (cp, '\n')) {		/* check for sneaky newlines */
	fatal (XLOG_CRIT,
	    "Newline in '%c' control field: possible crack attempt by %s!",
			c, LOGNAME);
    }

    buflen = strlen (cp) + 2;
    if ((int) buflen >= MAXPARMLEN) {
	/* make sure we don't go overboard */
	cp[MAXPARMLEN] = '\0';
    }

    sprintf (buf, "%c%s\n", c, cp);
    if (write (fd, buf, buflen) != buflen) {
	logerr_die (XLOG_INFO, "error while writing control file");
    }
    return;
}

void 
Make_symlink_entry (dev_t dev, ino_t ino, int cfd) {
    char s[32];		/* RFC-compliant length */

    (void) sprintf (s, "%lu %lul", (unsigned long)dev, (unsigned long)ino);
    assert(strlen(s)<sizeof(s));
    Entry ('S', s, cfd);
    return;
}

void
fatal_changed (char *fname) {
    fatal (XLOG_CRIT, "file %s changed during printing, %s@%s", fname, Person, Host);
    return;
}

/***************************************************************************
 * 1. Get a control file (actually a temp file)
 * 2. Write the header information into the file
 * 3. If you have a filter output, queue that
 * 4. otherwise if you have a simple file, queue that
 * 5. otherwise copy all of the files
 ***************************************************************************/

static char control_field_order[] = "HPJCNMDTLABEFGIKOQRSUVWXYZ";
/*			             `--+----'`-----------+---'
 * 	    BSD wants these first ------'		  |
 * These are just the remainder of the applicable fields -'
 */

void 
Make_job (void) {
    char *cf;			/* control file name */
    char *tmpf;			/* tmp control file name */
    int cfd, sdfd;		/* control file fp */
    int fd;			/* data file fd */
    int i, c;			/* ACME Integers, Inc. */
    long j, jobsize;
    char *s;			/* ACME Pointers, Inc. */
    struct stat f_stat, c_stat;	/* source and copy files status */
    unsigned long secs, usecs;

    jobsize = 0;
    /*
     * get the control file
     */
    cf = Get_cf ();

    /* create a temporary file in /tmp.
     * This will be owned by (but exclusively opened by) the user,
     * and we'll later read the control file from it.
     */
    tmpf = tmpnam (NULL);
    cfd = open (tmpf, O_RDWR | O_CREAT | O_EXCL, 0600);
    if (cfd < 0) {
	logerr_die (XLOG_INFO, "cannot open %s", tmpf);
    }

    if (!BK) {
	/* place in the current time as accurately as possible. */
	hires_time (&secs, &usecs);
	if (usecs) {
	    (void) sprintf (WHENSP, "%lu.%lu", secs, usecs);
	} else {
	    (void) sprintf (WHENSP, "%lu", (unsigned long) time ((time_t *) 0));
	}
    }

    /* set the class name to the priority, if not specified. */
    if (BK) {
	/* BSD lpds use this for the originating hostname!! >:( */
	(void) strcpy (CLASSNAME, ShortHost);

    } else {
	CLASSNAME[0] = Priority;
	CLASSNAME[1] = '\0';
    }

    /* set the job name to the files if not specified */
    if (Parmcount == 0) {
	if (JOBNAME[0] == '\0') {
	    strncpy (JOBNAME, "(stdin)", MAXPARMLEN);
	    JOBNAME[MAXPARMLEN] = '\0';
	}

    } else {
	char *nl, *el;
	nl = JOBNAME;
	el = JOBNAME + 99;			/* as per RFC */
	for (i = 0; i < Parmcount; ++i) {
	    if (Parms[i].str) {
		if (JOBNAME[0] == '\0') {
		    if (((nl = estrcp (nl, Parms[i].str, el)) == NULL) ||
			((nl = estrcp (nl, " ", el)) == NULL))
		    {
			break;	/* out of space */
		    }
		}
	    }
	    JOBNAME[100] = '\0';
	}
    }

    /*
     * Insert EU entry into JOBNAME (J) entry. This bit is unnecessarily paranoid about
     * commas getting in, they can't get in from the user options.
     */
    if ((EU) && (EUCS_banner_format)) {
	int jlen = 0;
	JOBNAME[0] = '\0';
	for (i = 0; i < num_euent; i++) {
	    char *value = EU_entry[i].value;
	    char *name = EU_entry[i].name;
	    if (value && *value) {
		if ((int) (jlen + strlen (name) + strlen (value) + 2) < MAXPARMLEN)
		{
		    if (jlen)
			JOBNAME[jlen++] = ',';
		    /* no commas allowed in name */
		    for (; *name; name++) {
			if (*name == ',')
			    JOBNAME[jlen++] = ' ';
			else
			    JOBNAME[jlen++] = *name;
		    }
		    JOBNAME[jlen++] = '=';

		    /* no commas allowed in value */
		    for (; *value; value++) {
			if (*value == ',')
			    JOBNAME[jlen++] = ' ';
			else
			    JOBNAME[jlen++] = *value;
		    }
		} else {
		    log (XLOG_INFO,
			"Make_job: EU control file entry too long: ignoring %s=%s",
			name, value);
		}
	    }
	}
	JOBNAME[jlen] = '\0';
    }

    /*
     * Put entries in for each character in the Parms list
     *
     * BSD lpd uses the L record as the trigger for doing the banner.
     * If the L record comes before the J then the banner will show
     * the job string for the previous job.  To work around this we
     * try to be careful in emulating the BSD order of control file
     * entries. -- Paul Haldane (EUCS), jmason.
     */

    for (i = 0; control_field_order[i]; ++i) {
	int field;

	field = control_field_order[i];
	s = CFparm[field - 'A'];
	if (*s == '\0') {
	    continue;
	}

	if (u_option) {
	    /* If the U option is set, the job should not appear under
	     * the bannername "root", but under this name on which
	     * the U option has been set.
	     */
	    if (field == 'P') {
		s = BNRNAME;
	    } else if (field == 'L') {
		s = "U-OPTION";
	    }
	}

	if (field == 'J' && (pr_used) ) {
	   (void) strcpy (s, "(stdin)");
	}

	if (BK) {
	    /* backward compatibility flag. Stick to the RFC as much as
	     * possible, including truncating possibly overlong fields.
	     */
	    switch (field) {
	    case 'D': case 'E': case 'Q': case 'R': case 'Z':
		/* PLP args - forget them */
		break;

	    case 'J':
		s[99] = '\0'; goto bk_makeentry;

	    case 'N':
		s[131] = '\0'; goto bk_makeentry;

	    case 'T':
		s[79] = '\0'; goto bk_makeentry;

	    case 'L': 
		/* BSD lpds use this for the username,
		 * not the GECOS info.
		 */
		(void) strcpy (s, LOGNAME);
		s[31] = '\0'; goto bk_makeentry;

	    case 'C': case 'H': case 'M': case 'P':
		s[31] = '\0';

bk_makeentry:
	    default:
		Entry (field, s, cfd);
		break;
	    }

	} else {
#ifdef SHORTHOSTNAME
	    /*
	     * this is for machines that will always 
	     * have their fqdn attached in the H field
	     * of the control file; shorthostname is ok.
	     */
	    if (field == 'H') {
		s = ShortHost;
	    }
#endif
	    Entry (field, s, cfd);
	}
    }

    /*
     * put out the created files or copy the ones passed; first find if there are any.
     */

    if (Read_stdin || Filter_stdin || Pr_stdin) {
	if (Filter_stdin) {
	    s = Filter_stdin;
	} else if (Pr_stdin) {
	    s = Pr_stdin;
	} else {
	    s = Read_stdin;
	}
	Get_stat (SD, s, &f_stat);
	j = (f_stat.st_size + 1023) / 1024;
	for (c = 0; c == 0 || c < Copies; ++c) {
	    Entry (Format, s, cfd);
	    jobsize += j;
	    if (MX > 0 && jobsize > MX) {
		Diemsg ("job too large. split up");
	    }
	}
	Entry ('U', s, cfd);
	if (Save_copy) {
	   save_dfname = s;
	}
	Entry ('N', "(stdin)", cfd);
	/* write Q entry to control file */
	Entry ('Q', Printer, cfd);
    } else {
	int prefilter_flag;
	
	/*
	 * we create an entry for each file in the list
	 */
	for (i = 0; i < Parmcount; ++i) {
	    if (!Parms[i].str) {	/* skip if non-printable */
		continue;
	    }
	    /*
	     * if the filename field is set, a prefilter has
	     * been run and has already put the files in the
	     * spool directory, so we just have to stat 'em.
	     */
	    if (*Parms[i].filename) {
		char fullpath[MAXPATHLEN];

		DEBUG3 ("prefilter output in spool dir: %s",
			Parms[i].filename);

		prefilter_flag = True;
                assert((strlen(SD)+2+strlen(Parms[i].filename))<=sizeof(fullpath));
		(void) sprintf (fullpath, "%s/%s", SD, Parms[i].filename);

		if (stat (fullpath, &f_stat) < 0) {
		    fatal (XLOG_CRIT, "Make_job: cannot stat %s", fullpath);
		}

	    } else {
		DEBUG3 ("Make_job: copying/linking to spool dir: %s",
			Parms[i].str);

		prefilter_flag = False;
		fd = -1;
		if (Use_links) {
		    /*
		     * try to make a symbolic link to the file
		     * (only use a single stat here; we're going to do
		     * another Is_printable test in a minute).
		     */
		    if (stat (Parms[i].str, &f_stat) < 0)
			fatal_changed (Parms[i].str);
		    s = Make_link (Parms[i].str, &c_stat);

		} else {
		    /*
		     * make a copy in the spool directory
		     */
		    if ((fd = Is_printable (Parms[i].str, &f_stat)) < 1)
			fatal_changed (Parms[i].str);
		    s = Copy_file (fd, Parms[i].str, &c_stat);
		}
		(void) strcpy (Parms[i].filename, s);
		if (fd >= 0) {
		    (void) close (fd);
		}

#ifndef HAVE_FTRUNCATE_KERNEL_BUG
		/* we don't actually need this at all, but it's safer
		 * to keep checking. However, it breaks with the
		 * SunOS 4.1.2 ftruncate() kernel bug.
		 */
		if (f_stat.st_size != c_stat.st_size)
		    fatal_changed (Parms[i].str);
#endif
	    }

	    Parms[i].size = (f_stat.st_size + 1023) / 1024;
		DEBUG4 ("Make_job: %s -> %s, %d",
		     Parms[i].str, Parms[i].filename, Parms[i].size);
	}

	for (i = 0; i < Parmcount; ++i) {
	    if (!Parms[i].str) {	/* skip if non-printable */
		continue;
	    }
	    for (c = 0; c == 0 || c < Copies; ++c) {
		/* are we symlinking these files into the spool dir?
		 * if we are, check their printability again then put
		 * the dev/ino data into the control file.
		 */
		if (Use_links) {
		    if ((fd = Is_printable (Parms[i].str, &f_stat)) < 0)
			fatal_changed (Parms[i].str);
		    Make_symlink_entry (f_stat.st_dev, f_stat.st_ino, cfd);
		}

		Entry (Format, Parms[i].filename, cfd);
		jobsize += Parms[i].size;
		if (Use_links == False && MX > 0 && jobsize > MX) {
		    Diemsg ("The job is too large (over %ldk). split up.", 
			    (long) jobsize);
		}
	    }
	    Entry ('U', Parms[i].filename, cfd);
	    Entry ('N', Parms[i].str, cfd);

	    /* write Q entry to control file */
	    Entry ('Q', Printer, cfd);

	    if (Remove) {
		if (Use_links == False || prefilter_flag) {
		    /* we should do the unlinking ourself */
		    if (unlink (Parms[i].str) < 0) {
			Warnmsg ("could not remove %s - %s",
					Parms[i].str, Errormsg (errno));
		    }

		} else {
		    if (Parms[i].str[0] != '/')
			Parms[i].str = Full_path (Parms[i].str);

		    Entry ('U', Parms[i].str, cfd);
		}
	    }
	}
    }

    /* flush the control file, chown it to daemon, and rename it */

    if (Ctrlfile_logdir && *Ctrlfile_logdir) {
	char buff[128];			/* log it for debugging */
	(void) sprintf (buff, "lpr-out.%s", ShortHost);
	Log_cf (cf, buff, cfd);
    }

    (void) lseek (cfd, 0L, SEEK_SET);

    /* copy the control file into the spooldir. */
    sdfd = Open_SD (cf);
    user_to_daemon ();
    {
	char buf[BUFSIZ];
	int n;
	while ((n = read (cfd, buf, sizeof (buf))) > 0) {
	    if (write (sdfd, buf, n) != n)
		logerr_die (XLOG_INFO, "copy: cannot write to file %s.", cf);
	}
	if (n < 0)
	    logerr_die (XLOG_INFO, "copy: error reading from %s", tmpf);
    }
    (void) close (sdfd); 

    daemon_to_user (); 
    Rename_cf (cf);
    if (Save_copy) {
       cf[0] = 'c';
       save_cfname = cf;
    }
    (void) close (cfd); 
    unlink (tmpf);
    return;
}
