/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1987, 1988, Patrick Powell
 * version 3.3.0 Justin Mason July 1994
 ***************************************************************************
 * MODULE: displayq.c
 * Display the status of a Printer queue
 ***************************************************************************/

#include "lp.h"
#include "common/displayq.h"
#include "library/errormsg.h"
#include "library/utils.h"

static int header;
static int server;

static char *short_time (time_t);
static void showentry (int, struct plp_queue *);
static void showstatus (int);

static int Old_header_format;
char progress_name[BUFSIZ];

static void
log_no_printer (char *pr) {
    (void) printf ("Printer '%s' does not exist!\n", pr);
    (void) fflush (stdout);
    if (ThisIsLPD) {		/* we're the daemon -- log this error */
	log (XLOG_WARNING,
		"displayq: printer '%s' does not have a printcap entry!", pr);
    }
}

/***************************************************************************
 *  1. get the printcap entries
 *  2. get the queue entries
 *  3. get the active server files and see if they are in queue
 *  4. print the queue status to stdout
 ***************************************************************************/

void 
Displayq (int zeige, int use_old_format)
{
    int i;			/* ACME Integers, Inc. */
    struct plp_queue *q;	/* job entries */
    FILE *afp;			/* attach file pointer */
    char afname[MAXPARMLEN];	/* attach file name */
    short att_mark;             /* Flag if a printer is attached to another printer */
    int seq;

    att_mark = False;
    
    if (use_old_format) {
	Old_header_format = True;
    } else {
	Old_header_format = False;
    }

    if (Use_with_cde && Queue_format == REQ_DLONG) {
	Old_header_format = True;
    }

reattach:			/* restart from here if printer is attached to another */
    /*
     * get the printcap entry
     */
    if (Get_pc_entry (Printer, Status_pc_vars, Status_pc_len) == 0) {
	log_no_printer (Printer);
	return;
    }
    

    /* 
     * If there are multiple servers for one queue, you can address 
     * every single one of them
     */
    if (SS == 0 || *SS == 0) { 
	;
    } else {
	(void) strcpy (Printer, SS);
	if (Get_pc_entry (Printer, Status_pc_vars, Status_pc_len) == 0) {
	    log_no_printer (Printer);
	    return;
	}
    }

    if (SD == 0 || *SD == 0) {
	log (XLOG_WARNING, "Printer '%s' does not have a spool directory!",
	     Printer);
	return;
    }

    DEBUG5 ("Displayq: doing %s on host %s", Printer, Host);

    /* construct the progress file name */
    (void) sprintf (progress_name, "%s/progress", SD);

    /*
     * If you have an NFS spool dir, use the remote lpd.
     */
    if (RM && *RM && NW) {
	if (Use_with_cde) {
	    ;
	} else {
	    Remote_status ();
	    return;
	}
    }
    /* chdir to spool directory */
    chdir_SD ();

 
    /*
     * check to see if attached to another printerq, alter 
     * printer if attached
     */
    if (Attach_file && *Attach_file && ((afp = fopen_daemon (Attach_file, "r")))) {
	if (fscanf (afp, "%s", afname) == 1) {
	    if (strsame (afname, Printer)) {
		fatal (XLOG_INFO, "Printer '%s' attached to itself", Printer);
	    }
	    (void) strcpy (unattached_printer, Printer);
	    Printer = strdup (afname);
	    att_mark = True;
	    (void) fclose (afp);
	    
	} else {
	    fatal (XLOG_INFO, 
		   "attach file for printer '%s' corrupted -- remove by hand!",
		   Printer);
	    (void) fclose (afp);
	}
	
	DEBUG2 ("Displayq: printer %s reattached to %s",
		unattached_printer, Printer);
	goto reattach;
    }
    
    /*
     * get the numbers of jobs here
     */
    Jobcount = Getq ();
    /*
     * Checkactive checks for the active jobs
     */
    server = Checkactive ();

    if (!called_by_lpr) {
	if (Old_header_format) {
	    if (SA) {
		(void) printf ("Save area %s:\n", Printer);
	    } else {
		if (LO_statb.st_mode & (DISABLE_PRINT)) {
		    (void) printf ("Warning: %s is down: printing disabled\n", Printer);
		} 
		if (LO_statb.st_mode & (DISABLE_QUEUE)) {
		    (void) printf ("Warning: %s queue is turned off\n", Printer);
		}
		
		if (server) {
		    (void) printf ("%s is ready and printing\n", Printer);
		} else {
		    (void) printf ("Warning: no daemon present\n");
		}
		(void) printf ("\n");
	    }
	} else if (Show_empty_queues || Jobcount > 0 ||
		   (zeige && RM && *RM) ||
		   (LO_statb.st_mode & (DISABLE_QUEUE | DISABLE_PRINT)))
	    {
		Print_queuehead (LO_statb.st_mode,
				 att_mark ? unattached_printer : NULL);
		showstatus (!(LO_statb.st_mode & DISABLE_PRINT));
	    }
	(void) fflush (stdout);
    }
    seq = 1;

    /* go thru list of jobs first to find the active job(s).
     * display it straightaway, irrespective of priority.
     */
    header = False;
    for (i = 0; i < Jobcount; ++i) {
	q = &Queue[i];
	if (q->q_daemon && (Parmcount <= 0 || Match_entry (q) || (called_by_lpr))) {
	    showentry (seq, q);
	    seq++;
	    header = True;
	}
    }

    if (seq > 1) {
	seq = 1;
    }
    /* now, display the non-active jobs. */
    for (i = 0; i < Jobcount; ++i) {
	q = &Queue[i];
	if (q->q_daemon == 0) {
	    if (Parmcount <= 0 || Match_entry (q) || (called_by_lpr)) {
		showentry (seq, q);
		seq++;
		header = True;
	    }
	}
    }

    if (Jobcount == 0 && (Old_header_format)) {
	(void) fprintf (stdout, "no entries\n");
	(void) fflush (stdout);
    }

    if (zeige) {
	if (RM && *RM) {
	    DEBUG4 ("Displayq: getting Remote_status():");
	    Remote_status ();
	}
    }
    return;
}

/**************************************************************************/


static char old_hdr_format[] = "%*s %-10s %-4s %*.*s %s\n";
static char old_info_format[] = "%*s %-10s %-4d %*.*s %d bytes\n";

#define RANK_COLUMN_WIDTH	6
#define SIZE_COLUMN_WIDTH	6
#define FILE_COLUMN_WIDTH	25
#define COPY_COLUMN_WIDTH	13

#define LONG_FILE_COLUMN_WIDTH	50
#define LONG_SIZE_COLUMN_WIDTH  12

#define OLD_FILE_COLUMN_WIDTH	37

#define TWENTY_FOUR_HOURS	(24 * 60 * 60)

/* short_time includes a mod from Ed Santiago to cut down the date length
 * even more so lpq output fits into 80 columns again.
 */
static char *
short_time (time_t tvec) 
{
    static char s[40];

    /*
     * 012345678901234567890123456789
     * Mon May 30 20:53:27 CDT 1988
     */
    if (tvec == (time_t)0) return ("-");

    (void) strcpy (s, ctime (&tvec));
    s[16] = '\0';
    if (time((time_t *) NULL) - tvec < TWENTY_FOUR_HOURS) {
	return (&s[11]);
    } else {
	s[10] = '\0';
	return (&s[4]);
    }
}

/* static char hdr_format[]  = "%*s %-8.8s %2s %-3s %4s %-11.11s %*.*s %*s %s\n"; */
/* static char info_format[] = "%*s %-8.8s %-c  %-3s %4d %-11.11s %*.*s %*s %s\n"; */

static char hdr_format[]  = "%*s %-8.8s %2s %4s %-11.11s %*.*s %3s %4s %*s %s\n";
static char info_format[] = "%*s %-8.8s %c  %4d %-11.11s %*.*s %3s %4s %*s %s\n";

static void 
showentry (int seq, struct plp_queue *q) 
{
    char rank[10];
    char *suffix, *cp;
    char name[BUFSIZ];
    char nbuf[100];
    char sbuf[10];
    char percent[4];
    char job_name[BUFSIZ];
    FILE *fp_progress;


    assert (q!=(struct plp_queue*)0);
    if (!called_by_lpr) {
        if (header == 0 && 
	    ((Queue_format == REQ_DSHORT) || (Queue_format ==  REQ_DSHORT_OLD))){
	    if (Old_header_format) {
		(void) printf (old_hdr_format, -RANK_COLUMN_WIDTH,
			       "Rank", "Owner", "Job",
			       -OLD_FILE_COLUMN_WIDTH, OLD_FILE_COLUMN_WIDTH,
			       "Files", "Total Size");
	    } else {
		if (!called_by_lpr) {
		    int copy_width, file_width;
		
		    copy_width = (Queue_format == REQ_DSHORT) ? 0 : COPY_COLUMN_WIDTH;
		    file_width = FILE_COLUMN_WIDTH - copy_width;
		
		    (void) printf (hdr_format, RANK_COLUMN_WIDTH,
				   "Rank", "Owner", "Pr", "Job", "Host",
				   -file_width+5, file_width-5, "Files", "   %", "Form",
				   SIZE_COLUMN_WIDTH, "Size", "Time");
		}
	    }
	}
    }
    
    (void) sprintf (percent, "   ");

    if (q->q_daemon) {
	if (SV == 0 || *SV == 0) {
	    (void) strcpy (rank, "active");
	    /* get the progress form the file "progress";
	     * of course only for active jobs 
	     */
	    if ((fp_progress = fopen_daemon (progress_name, "r")) == NULL) {
		/* file not available, we write a "   " on output */
		(void) sprintf (percent, "  0");
		percent[3] = '\0';
	    } else {
		/* file exists, we read the information form it */
		fgets (percent, 4, fp_progress);
		(void) fclose (fp_progress);
	    }
	} else {
	    unsigned int namelen = strlen (q->q_server);
	    if (namelen >= RANK_COLUMN_WIDTH) {
		(void) strcpy (rank, q->q_server + (namelen - RANK_COLUMN_WIDTH));
		rank[0] = '/';
	    } else {
		(void) strcpy (rank, q->q_server);
	    }
	}
    } else {
	switch (seq) {
	    case 1: suffix = "st"; break;
	    case 2: suffix = "nd"; break;
	    case 3: suffix = "rd"; break;
	    default: suffix = "th"; break;
	}
	(void) sprintf (rank, "%d%s", seq, suffix);
	    if (called_by_lpr && q->q_num == Job_number_by_lpr) {
	        (void) sprintf (rank_for_lpr, "%s", rank);
		return;
	    }
    }

    /*
     * when called by lpr, only the rank of the actual job is interesting;
     * all others (when available) should be suppressed
     */
    if (called_by_lpr) {
        return;
    }

    if ((Queue_format == REQ_DSHORT) || (Queue_format ==  REQ_DSHORT_OLD)) {
	int left, idx, len;
	char *owner, *from;

	if ((suffix = strchr (q->q_user, '@')) == NULL) {
	    /* print short name */
	    (void) strcpy (name, &q->q_from);
	    if ((suffix = strchr (name, '.')) != NULL)
		*suffix = '\0';
	    owner = q->q_user;
	    from = name;
	} else {
	    (void) strcpy (name, q->q_user);
	    name[suffix - q->q_user] = '\0';
	    owner = name;
	    from = suffix + 1;
	}

	/* fill in the filenames into the Files header string */
	if (Old_header_format)
	    left = OLD_FILE_COLUMN_WIDTH;
	else
	    left = FILE_COLUMN_WIDTH;

	for (idx = 0, cp = nbuf; left > 0 && idx < q->q_entries; idx++) {
	    (void) strcpy (job_name, q->q_data[idx].name);
	    len = strlen (job_name);
	    if (len <= (left - (idx < q->q_entries - 1 ? 4 : 0))) {
		if (idx > 0)
		    *cp++ = ' ';
		if (q->q_data[idx].copies > 1) {
		    if (q->q_data[idx].copies < 10) {
			len += 4;
		    } else if (q->q_data[idx].copies < 100) {
			len += 5;
		    } else {
			len += 6;
		    }
		    (void) sprintf (cp, "%d * ", q->q_data[idx].copies);
		    (void) strcat (cp, job_name);
		} else {
		    (void) strcpy (cp, job_name); 
		}
		left -= (len + (idx > 0 ? 1 : 0));
		cp += len;
	    } else {
		if (left - 4 > 0) { 
		    *cp++ = ' ';
		    (void) strncpy (cp, job_name, left - 4);
		    cp += left - 4;
		}
		break;
	    }
	}
	if (idx < q->q_entries) {
	    if (idx > 0) 
	     *cp++ = ' ';
	    (void) strcpy(cp, "...");
	}

	/* printf the header */
	if (Old_header_format) {
	    (void) printf (old_info_format, -RANK_COLUMN_WIDTH, rank, owner,
			   q->q_num, -OLD_FILE_COLUMN_WIDTH, OLD_FILE_COLUMN_WIDTH,
			   nbuf, q->q_size);
	} else {
	    if (q->q_size < 1024) {
		(void) sprintf (sbuf, "%*ld", SIZE_COLUMN_WIDTH, q->q_size);
	    } else if (q->q_size < (1024 * 1000)) {
		(void) sprintf (sbuf, "%*.1fK", SIZE_COLUMN_WIDTH - 1,
				((double) q->q_size) / 1024.0);
	    } else if (q->q_size < (1024 * 1000 * 1000)) {
		(void) sprintf (sbuf, "%*.1fM", SIZE_COLUMN_WIDTH - 1,
				((double) q->q_size) / (1024.0 * 1000.0));
	    } else {
		(void) printf (sbuf, "%*.1fG", SIZE_COLUMN_WIDTH - 1,
			       ((double) q->q_size) / (1024.0 * 1000000.0));
	    }
	    if (strlen (nbuf) > 21) {
		(void) strncpy (&nbuf[18], "...", 3);
	    }
	    (void) printf (info_format, RANK_COLUMN_WIDTH, 
			   q->q_held ? "hold" : rank, owner, 
			   q->q_held ? '*' : q->q_priority, 
			   q->q_num, from, -FILE_COLUMN_WIDTH+4, FILE_COLUMN_WIDTH-4, 
			   nbuf, percent, q->q_options ? q->q_options : DEF_OPTIONS,
			   SIZE_COLUMN_WIDTH, sbuf, short_time (q->q_time));
	}

    } else {		/* Queue_format == REQ_DLONG */
	int idx, corr = 0;
	
	(void) sprintf (name, "[job %3.3s %s]", q->q_name + 3, q->q_name +6);
	/* (void) sprintf (name, "[job %03d@%-s]", q->q_num, q->q_name + 6);*/
	/* Old header format, for compatibility with the bsd lpq */
	if (Old_header_format) {
	    if (strlen (name) < 18) {
		corr = 17 - strlen (name);
	    }
	    (void) printf("%s: %-*s %s\n", q->q_user,
                      58 - strlen (q->q_user) - strlen (rank) - strlen (name) - corr,
                      rank, name);

	    for (idx = 0; idx < q->q_entries; idx++) {
		int file_width = LONG_FILE_COLUMN_WIDTH;
		(void) strcpy (job_name, q->q_data[idx].name);
		(void) fputs("\t", stdout);
		if (q->q_data[idx].copies > 1) {
		    (void) printf("%2d copies of ", q->q_data[idx].copies);
		    file_width -= sizeof("## copies of ") - 1;
		}
		(void) printf("%-*s %ld bytes\n",
			      file_width + 6 - strlen (q->q_user) - strlen (rank) - strlen (name) -corr,
			      job_name,
			      q->q_data[idx].size);
	     printf ("\t\tSubmitted: %s\n\n", Time_str_cde (q->q_time)); 
	    }
	} else {
	/* we define our own long format */
	    char time_string[26];
	    (void) strncpy (time_string, ctime (&q->q_time), strlen (ctime (&q->q_time)) - 1);
	    time_string[24] = '\0';
	    (void) sprintf (name, "[job %3.3s %s]", q->q_name + 3, q->q_name + 6);
	    if (strcmp (rank, "active") == 0) {
		(void) printf ("%s: %s started at %-*s %s\n", q->q_user, rank, 
			       79 - strlen (q->q_user) - 15 - strlen (rank) - strlen (name),
			      Time_str (), name);
	    } else {
		(void) printf ("%s: %s submitted: %-*s %s\n", q->q_user, rank, 
			       79 - strlen (q->q_user) - 15 - strlen (rank) - strlen (name),
			       time_string, name);
	    }
	    for (idx = 0; idx < q->q_entries; idx++) {
		int file_width = LONG_FILE_COLUMN_WIDTH;
		
		(void) fputs ("\t  ", stdout);
		if (q->q_data[idx].copies > 1) {
		    (void) printf ("%2d copies of ", q->q_data[idx].copies);
		    file_width -= sizeof ("## copies of ") - 1;
		}
		
		(void) printf ("%-*s %*ld bytes\n",
			       file_width,
			       q->q_data[idx].name,
			       LONG_SIZE_COLUMN_WIDTH,
			       q->q_data[idx].size);
	    }
	    if (percent[0] != ' ') {
		(void) printf ("\t  printed: %s %%.\n", percent);
	    }
	}
    }
    (void) fflush (stdout);
    return;
}

/***************************************************************************
 * show the status files for the server devices
 ***************************************************************************/

static void 
showstatus (int k) 
{
    static char server_name[BUFSIZ];	/* Name of server file */
    long active;
    pid_t pid;
    char buf[BUFSIZ];
    char *sp, *ep;		/* ACME Pointer */
    char saved_Printer[PRNAMELEN];

    /*
     * If there are jobs in the queue, but no server running, print a warning message.
     */
    if (SV == 0 || *SV == 0) {
	DEBUG2 ("showstatus: single device: k=%d, server=%d", k, server);

	if (Jobcount > 0 && server == 0 && k) {
	    (void) printf ("Warning: queue is not active (no server present).\n");
#if 0
	    /*
	     * Suggest why nothing is moving...
	     */
	    if (LO_statb.st_mode & DISABLE_PRINT == False) {
		(void) printf ("Possible reasons:\n");
		(void) printf ("\t* the server hasn't started yet.\n");
		(void) printf ("\t* the remaining jobs are being held.\n");
		if (current_limit() > 0) {
		    (void) printf
			("\t* the jobs in the queue are over the size limit.\n");
		}
		if (FL && *FL) {
		    (void) printf
			("\t* the jobs in the queue need a different form type.\n");
		}
	    }
#endif
	}
	printstatus (0);
	return;
    }

    DEBUG2 ("showstatus: multiple devices: k=%d, server=%d", k, server);

    /*
     * check for each of the servers
     */
    (void) strcpy (server_name, SV);
    (void) strcpy (saved_Printer, Printer);

    for (sp = server_name; sp; sp = ep) {
	if ((ep = strchr (sp, ','))) {
	    *ep = '\0';
	    ++ep;
	}
	strcpy (Printer, sp);
	
	/*
	 * get the lock file and the status from the server
	 */
	if (Get_multi_pc_entry (sp, NULL, Server_pc_vars, Server_pc_len) == 0) {
	    fatal (XLOG_INFO, "no server entry for %s", sp);
	}
	buf[0] = '\0';
	active = Checklockfile (sp, &pid, buf, sizeof (buf), &LO_statb);

	(void) printf (" Server %s", sp);

	/* display the comment field */
	if (CM && *CM) {
	    (void) fputs (" (", stdout);
	    Print_comment (CM);
	    (void) fputs (")", stdout);
	}

	(void) fputs (" is ", stdout);

	if (active) {
	    (void) printf ("active (pid %ld)", (long) pid);
	} else if (LO_statb.st_mode & DISABLE_PRINT) {
	    (void) fputs ("down", stdout);
	} else {
	    (void) fputs ("idle", stdout);
	}

	if (LO_statb.st_mode & DISABLE_PRINT) {
	    (void) fputs (" (printing disabled)", stdout);
	}

	(void) fputs (".\n", stdout);
	(void) fflush (stdout);
	printstatus (0);
    }
    (void) strcpy (Printer, saved_Printer);
    return;
}
