/***************************************************************************
* Module: move.c
* Purpose: moves the job(s) from one printer queue to another printer queue
*          and informs the user about the moved jobs via E-Mail
****************************************************************************/

#include <pwd.h>

#include "lp.h"
#include "library/errormsg.h"
#include "library/utils.h"
#include "common/hold.h"
#include "common/move.h"

#define MOVE_MESSAGE \
"Message from PLP printer spooling system.\n\nDear user %s,\n\nuser %s has moved your print job %d [%s] \n\
from printer '%s' to printer '%s', job %d."

extern char Mail_log_message[];

static char *new_SD;		/* destination spool directory */
static int New_Job_number;      /* New job number, needed in move.c */
extern int create_history_entry (int, struct plp_queue *);

/**************************************************************************
 * move_file() 
 * rename "from" to "to", or copy if they're on different
 * filesystems.
 **************************************************************************/
static void
move_file (char *from, char *to) {
    user_to_daemon ();
    if (rename (from, to) != 0) {
	if (errno == EXDEV) {
	    if (copy_file (from, to, -1) < 0) {
		daemon_to_user ();	/* Switch back before exit */
		logerr_die (XLOG_INFO, "move_file: copy %s to %s failed", from, to);
	    }
	    if (unlink (from) < 0) {
		daemon_to_user ();	/* Switch back before exit */
		logerr (XLOG_INFO, "move_file: unlink %s failed", from);
	    }
	} else {
	    daemon_to_user ();		/* Switch back before exit */
	    logerr_die (XLOG_INFO, "move_file: rename %s to %s failed", from, to);
	}
    }
    daemon_to_user ();
    return;
}

/***************************************************************************
 * Move a job
 * ACTIONS:
 * 1. Move job's files to other spool directory
 ***************************************************************************/

static int
move_job (FILE *cfp, struct plp_queue *q, int Job_number) {
    struct stat stbuf;		/* for statting files */
    char l[BUFSIZ];		/* for reading the control file */
    char fullname[BUFSIZ];	/* for the full pathname */
    char newname[BUFSIZ];	/* full pathname of new name  */
    char *filepart;		/* end of the pathname */
    char *newfilepart;		/* end of the pathname */
    struct ch_newname           /* Struct for the new cf filename */
    {
	char anf[3];
	char anumber[3];
	char temp_shorthost[64];
    };
    struct ch_newname tmp_new_filepart; 
    char a_Job_number[4];       /* the new Jobnumber */
    char cf_buffer[BUFSIZ];     /* the new cf-file */
    int i;                      /* flag for removing junk from cf_buffer */

    assert(q!=(struct plp_queue*)0);
    assert(cfp!=(FILE*)0);

    tmp_new_filepart.anf[0] = '\0';
    a_Job_number[0] = '\0';
    i = 0;
    if (new_SD == 0 || *new_SD == 0) {
	fatal (XLOG_INFO, "no directory entry");
    }
    /* get the directory name and copy to buffer */
    if (strlen (new_SD) + sizeof (q->q_name) + 2 > sizeof (newname)) {
	logerr_die (XLOG_NOTICE, "INTERNAL: move_job, file name too big");
    }
    (void) strcpy (newname, new_SD);
    newfilepart = &newname[strlen (newname)];
    *newfilepart++ = '/';

    (void) strcpy (fullname, SD);
    filepart = &fullname[strlen (fullname)];
    *filepart++ = '/';
    (void) strcpy (filepart, q->q_name);

    /* check that the control file is there */
    if (stat (fullname, &stbuf) < 0) {
	return True;		/* Doesn't exist */
    }
    while (fgets (l, sizeof (l), cfp)) {
	/* clobber the \n */
	l[strlen (l) - 1] = '\0';
	if (islower (l[0]) || ( l[0] =='U') ) {
	    if (islower (l[0])) {
		(void) strcpy (filepart, &l[1]);
		if (stat (fullname, &stbuf) < 0) {
		    (void) fclose (cfp);
		    DEBUG5 ("move_job: data file %s does not exist", l);
		} else {
		    /* Move file */
		    (void) strcpy ((char *) &(tmp_new_filepart.anf[0]), &(l[1]));
		    sprintf (a_Job_number , "%03d", Job_number);
		    tmp_new_filepart.anumber [0] = a_Job_number [0];
		    tmp_new_filepart.anumber [1] = a_Job_number [1];
		    tmp_new_filepart.anumber [2] = a_Job_number [2];
		    (void) strcpy (newfilepart, (char *) &(tmp_new_filepart.anf[0]));
		    
		    strncat (cf_buffer, &l[0], 1);
		    strcat (cf_buffer, (char *) &tmp_new_filepart);
		    strcat (cf_buffer, "\n");
		    
		    DEBUG4 ("move_job: renaming %s to %s ", fullname, newname);
		    move_file (fullname, newname);
		    Setuid_debug ("move_job");
		}
	    } else {
		/* hopefully 'd' preceedes 'U'  */
		strncat(cf_buffer, &l[0], 1);
		if (tmp_new_filepart.anf[0] == '\0')
		    fatal (XLOG_INFO, "tmp_new_filepart unset!");
		strcat (cf_buffer, (char *) &tmp_new_filepart);
		strcat (cf_buffer, "\n");
	    }
	} else {
	    if ( i == 0 ) /* First run; erase all the junk in cf_buffer */
		{
		    (void) strcpy (cf_buffer, &l[0]); 
		    strcat (cf_buffer, "\n");
		    i++; 
		} else {
		    if (l[0] == 'Q') { /* write the new printer name */
		        strcat (cf_buffer, "Q");
			strcat (cf_buffer, &Parms[1].str[0]);
			strcat (cf_buffer, "\n"); 
		    } else {
		        strcat (cf_buffer, &l[0]);
			strcat (cf_buffer, "\n"); 
		    }
		}
	}
    }

    DEBUG5 ("move_job: control file %s is ok", q->q_name);
    /* close the control file */
    (void) fclose (cfp);
    /* 
     * now the control file will be completly new written 
     * to avoid having any junk in the filedescriptor
     */
    if ((cfp = fopen_daemon (q->q_name, "w")) == NULL) {
        logerr (XLOG_INFO, "move_job: control file %s not writeable.", q->q_name);
    } else {
        fputs (cf_buffer , cfp);
	(void) fclose (cfp);
    }
    (void) strcpy (&(tmp_new_filepart.anf[0]), q->q_name);
    if (!*a_Job_number)
	fatal (XLOG_INFO, "move_job: new jobnumber could not be set.");

    tmp_new_filepart.anumber [0] = a_Job_number [0] ;
    tmp_new_filepart.anumber [1] = a_Job_number [1] ;
    tmp_new_filepart.anumber [2] = a_Job_number [2] ;
    (void) strcpy (newfilepart, &(tmp_new_filepart.anf[0]));
    (void) strcpy (filepart, q->q_name);
    
    move_file (fullname, newname);
    New_Job_number = Job_number;
    return False;
}

/***************************************************************************
 * domove (struct plp_queue *q, int Job_number)
 * move the job
 * 1. Lock the control file.
 * 2. If unsuccessful, find the server PID and kill it off.
 * 3. move the job.
 ***************************************************************************/

static void
domove (struct plp_queue *q, int Job_number) {
    FILE *cfp;
    int r;
    char username[64];            /* holds the username */
    struct passwd *pwent;         /* to get the real username */
    struct plp_queue *q_for_hist; /* for the history */

    assert(q!=(struct plp_queue*)0);
    q_for_hist = q;
    if ((cfp = Lockcf (q->q_name)) == NULL) {
	/* hmmm... looks like an active server */
	if ((cfp = fopen_daemon (q->q_name, "rw")) == NULL) {
	    /* nope, the file has really gone */
	    logerr (XLOG_INFO, "control file %s not read-writable", q->q_name);
	    return;
	}
	/* well, we will just have to kill off the server */
	if (q->q_daemon == 0) {
	    /*
	     * Hmmm... we have this fellow running the file, and it is locked.  That
	     * means that it just started running this guy. Better check again.
	     */
	    (void) Checkactive ();
	}
	if (q->q_daemon) {
	    (void) fprintf (stdout, "Killing off %s server %d.",
			    q->q_server, q->q_daemon);
	    user_to_root ();
	    r = killpg (q->q_daemon, SIGINT);
	    root_to_user ();
	    if (r < 0 ) {
		if (errno != ESRCH)
		    perror("killpg");
		DEBUG2 ("Server %s (%d) was not alive.",
			q->q_server, q->q_daemon);
	    }
	}
    }
    DEBUG3 ("domove: moving files for job %s", q->q_name);
    if (move_job (cfp, q, Job_number) == 0) {
      /* clearing the array "username" */
        int i = 0;
	for (i = 0; i < 64; i++) {
	  username[i] = '\0';
	}
	(void) strncpy (username, q->q_user, strlen (q->q_user));
	pwent = getpwnam (username);
	if (pwent != NULL) {
	    (void) sprintf (username, pwent->pw_gecos);
	}
        sprintf (Mail_log_message, MOVE_MESSAGE, 
		 username, Person, q->q_num, q->q_data->name, Printer, 
		 Parms[1].str, New_Job_number);
	job_moved = 1;
	sendmail (q, 1);
	job_moved = 0;
	/* update history file */
	user_to_daemon ();
	if (create_history_entry (6, q_for_hist) ) {
	    log (XLOG_WARNING, "domove: errors during create_history_entry");
	}
	daemon_to_user ();
    }
    return;
}

void
move (void) {
    int i;
    struct plp_queue *q;	/* job entry */
    int perms;			/* hold perms values */
    int perms_move;             /* for move permissions */
    int control_perms;		/* has control perms */
    int control_perms_move;     /* for checking if the user can move his job */
    short job_is_there;         /* flag, if the job was found in the queue */
    char *src_Printer;		/* source queue name */
    char buf[MAXPATHLEN];	/* holds the pathname */
    char tmp_buf[MAXPATHLEN];	/* holds the pathname */
    int Job_number;             /* jobnumber of the job to be moved */
    char *mem_q_user;           /* pointer on memuser */
    char memuser[4];            /* holds the user name from the controlfile*/
    struct stat st_cfbuf;       /* for statting files */

    mem_q_user = memuser;
    control_perms_move = 0;
    job_is_there = 0;

    /*
     * If explicitly asked for debug then echo messages to stdout otherwise (db flag in
     * printcap) just log messages in file.
     */
    Echo_on_stdout = Debug;
    src_Printer = Printer;
    /*
     * get the printcap entry
     */
    if (Get_pc_entry (Printer, Status_pc_vars, Status_pc_len) == 0) {
	(void) fprintf (stdout, "Printer or Queue %s does not exist!\n", Parms[1].str);
	(void) fflush (stdout);
	return;
    }
    if (SD == 0 || *SD == 0) {
	DEBUG2 ("not a Printer");
	return;
    }

    /* First check if the jobs to be moved are really in that queue */
    Jobcount = Getq ();
    (void) Checkactive ();
    for (i = 0; i < Jobcount; ++i) {
	 q = &Queue[i];  
	     /* searching through the queue for matching 
	      * jobnumber or
	      * username or
	      * hostname 
	      */
	 if (Match_entry (&Queue[i])) { 
	     mem_q_user = q->q_user;
	     job_is_there = 1;
	 } 
    }
    /* Flush out error message in case that there are no jobs to be moved */
    if (!job_is_there) {
	if (Parms[0].num < 0 ) {
	    if (strcmp (Parms[0].str, "*") == 0 ) {
		(void) fprintf (stdout,
				"Warning: no jobs exist on printer '%s'.\n", Printer);
	    } else {
		(void) fprintf (stdout,
			   "Warning: no jobs from host '%s' exist on printer '%s'.\n", 
			   Parms[0].str, Printer);
	    } 
	} else {
	    (void) fprintf (stdout, 
			    "Warning: job %d does not exist on printer '%s'.\n", 
			    Parms[0].num, Printer);
	}
	(void) fflush (stdout);
	return;
    }

    /* get the printcap entry of destPrinter*/
    Printer = Parms[1].str;
    if (Get_pc_entry (Printer, Status_pc_vars, Status_pc_len) == 0) {
        (void) fprintf (stdout, "Printer or Queue %s does not exist!\n", Parms[1].str);
        (void) fflush (stdout);
        return;
    }

    new_SD = (char *) malloc (strlen (SD) + 1);
    (void) strcpy (new_SD, SD);

    /*
     * set the flags needed
     */
    Is_local = hostcmp (FQDN, Host) == 0;
    Is_root = strsame (Person, "root");
    
    perms = 'C';                
    perms_move = 'M';           /* check for control or move perms 
				 * on destination queue 
				 */
    DEBUG4 ("move: checking 'C' or 'M' permissions");
    if (!Is_root) {
	if (Checkperm (FQDN, Person, First_name, &perms, (int *) 0, 0) < 1) {
	    if (Checkperm (FQDN, Person, First_name, &perms_move, (int *) 0, 0) < 1) {
		(void) fprintf (stdout, "Sorry user %s, but you do not have move permission to printer '%s'!\n", Person, Printer);
		(void) fflush (stdout);
		return;
	    }
	}
    }
    DEBUG4 ("move: permission (M) set on printer %s", Printer);

    control_perms_move = 1;
    Printer = src_Printer;

    if (Get_pc_entry (Printer, Status_pc_vars, Status_pc_len) == 0) {
	(void) fprintf (stdout, "Printer or Queue %s does not exist!\n", Printer);
	(void) fflush (stdout);
	return;
    }
    if (SD == 0 || *SD == 0) {
	DEBUG2 ("not a Printer");
	return;
    }

    /* chdir to source spool directory */
    chdir_SD ();

    /*
     * check to see that the user has RMJOB privs on this machine
     */

    DEBUG4 ("move: checking 'R' permission");
    perms = 'R';		/* must be able to at least use the Printer */
    if (!Is_root) {
	  if (Checkperm (FQDN, Person, First_name, &perms, (int *) 0, 0) < 1) {
	      (void) fprintf (stdout,
			      "Sorry user %s, but you do not have move permission on printer '%s'!\n",
			      Person, First_name);
	      control_perms = 0;
	      return;
	  }
    }
    DEBUG4 ("move: permission (R) set on printer %s", Printer);


    perms = 'C';		/* check for control perms */
    control_perms = 1;

    if (!Is_root) {
	if (Checkperm (FQDN, Person, First_name, &perms, (int *) 0, 0) < 1) {
		control_perms = 0;
	}
    }
    DEBUG4 ("move: permission (C) %s set on printer %s", 
	    control_perms ? "" : "not", Printer);

    DEBUG4 ("move: job belongs to %s, user is %s", mem_q_user, Person);
    /* check if the jobowner and the copier are the same - root has 
       permission too */
    if (!control_perms) {
	if (strcmp (mem_q_user, Person) != 0) {
	    (void) fprintf (stdout, "Sorry user %s,\nbut you cannot move job %d, owner %s, to printer '%s' (not owner)!\n", Person, Parms[0].num, mem_q_user, Printer);
	    (void) fflush (stdout);
	    return;
	}
    }

    DEBUG3 ("move: Is_root: %s, Is_local: %s, control_perms: %s",
	    Is_root ? "yes" : "no", 
	    Is_local ? "yes" : "no", 
	    control_perms ? "yes" : "no");

    /*
     * get the job queue (second pass)
     */
    Jobcount = Getq ();
    (void) Checkactive ();
    /*
     * run down list
     */
   if ((RM != NULL) && (*RM != '\0')) { /* It's a local queue */
       (void) fprintf (stdout, "Local  queue  '%s'", Printer);
   } else { /* local printer */
       (void) fprintf (stdout, "Local printer '%s'", Printer);
   }
    (void) fprintf (stdout, " (%s):\n", Host);
    (void) fflush (stdout);

    DEBUG5 ("move: Jobcount = %d", Jobcount);
    /*
     * get the sequence file name; this is needed when the job will be
     * moved in a spool directory where alread a job with this number
     * exists. Better increase the number of the moved job up by 1.
     */
    (void) sprintf (buf, "%s/.seq.%s", new_SD, ShortHost);
    DEBUG3 ("move: sequence file name '%s'", buf);

    for (i = 0; i < Jobcount; ++i) {
	q = &Queue[i];
	if (q->q_daemon == 0) {		/* only non-active jobs may be moved */
	    if (Match_entry (&Queue[i])) {
		Job_number = Get_seqnumber (buf);
		(void) sprintf (tmp_buf, "%s/cfX%03d%s", 
				new_SD, Job_number, ShortHost);
		DEBUG3 ("move: checking if cf-file %s exists.", tmp_buf);
		while (stat (tmp_buf, &st_cfbuf) == 0 ) {
		       Job_number = Get_seqnumber (buf);
		       tmp_buf[0] = '\0';
		       (void) sprintf (tmp_buf, "%s/cfX%03d%s", 
				       new_SD, Job_number, ShortHost);
		       DEBUG3 ("move: found existing cf-file %s, increasing Jobnumber.", 
			       tmp_buf);
		}
		DEBUG3 ("move: cf-file %s does not exist, continueing.", tmp_buf);
		DEBUG3 ("move: new number %d; checking %s, %d, %s",
				Job_number, q->q_name, q->q_num, q->q_user);
		if (shouldhold (q, control_perms)) {
		    domove (q, Job_number);
		    (void) fprintf (stdout,
			    "Moving job %d, owner %s, from printer '%s' to '%s', job %d.\n",
			    q->q_num, q->q_user, Printer, Parms[1].str, New_Job_number);
		    (void) fflush (stdout);
		}
	    }
	} else {
	     if ((Parms[0].num == q->q_num) || (!strcmp (Parms[0].str, q->q_user)) 
		 || (!strcmp (Parms[0].str, "*")) ) {
	           (void) fprintf (stdout,
				   "Warning: job %d active, it cannot be moved. Please use requeue and then move.\n", 
				   q->q_num);
		   (void) fflush (stdout);
	     }
	}
    }
    Startserver (); 

    /* and kick the server that we moved the job to */
    Printer = Parms[1].str;
    Startserver ();

    Printer = src_Printer;
    (void) free (new_SD);
}

