/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1987, 1988, Patrick Powell
 * version 3.3.0 Justin Mason July 1994
 ***************************************************************************
 * MODULE: rmjob.c
 * Remove jobs from a printer queue
 **************************************************************************/

#include "lp.h"
#include "library/errormsg.h"
#include "library/utils.h"

static int remove_all;		/* remove all jobs */

extern void Remote_remove (void);
extern int create_history_entry (int, struct plp_queue *);

/***********************************************************************
 *  a simple application of the removal decision table
 *
 *  user   local  job   from  remove_all    perms inlist    remove?
 *  ===============================================================
 *  root   yes    *     *     yes           *     *         yes
 *  root   yes    *     *     no            *     yes       yes
 *  root   no     *     mach  yes           *     *         yes
 *  root   no     *     mach  no            *     yes       yes
 *  user   -      user  mach  -             R     yes       yes
 *  user   -      *     *     -             C     yes       yes
 *
 * Returns: 1 if removal is indicated, 0 otherwise
 ***********************************************************************/
static int
shouldremove (struct plp_queue *q, int control_perms) {
    int i, same_host;

    assert(q!=(struct plp_queue*)0);
    DEBUG5 ("shouldremove: Person = %s, From = %s, job %s@%s",
	     Person, FQDN, q->q_user, &q->q_from);
    same_host = (hostcmp (FQDN, &q->q_from) == 0);
    i = Match_entry (q);
    DEBUG5 ("shouldremove: Match_entry = %d, permissions = %d", 
	    i, control_perms);

    if (Is_root && Is_local && remove_all)
	return True;
    if (Is_root && Is_local && !remove_all && i)
	return True;
    if (Is_root && !Is_local && same_host && remove_all)
	return True;
    if (Is_root && !Is_local && same_host && !remove_all && i)
	return True;
    if (strsame (Person, q->q_user) && remove_all)
	return True;
    if (strsame (Person, q->q_user) && i)
	return True;
    if (i && control_perms)
	return True;
    return False;
}

/***************************************************************************
 * remove the job
 * 1. Lock the control file.
 * 2. If unsuccessful, find the server PID and kill it off.
 * 3. Use brute force and remove the files.
 ***************************************************************************/

static void
doremove (struct plp_queue *q) {
    FILE *cfp;
    int r;
    struct plp_queue *q_for_hist;

    assert(q!=(struct plp_queue*)0);
    /* is the controlfile there...? */
    if ((cfp = fopen_daemon (q->q_name, "r")) == NULL) {
	/* nope, the file has really gone */
	logerr (XLOG_INFO, "control file %s not readable", q->q_name);
	return;
    }
    /* well, we will just have to kill off this server */
    if (q->q_daemon == 0) {
	/*
	 * Hmmm... we have this fellow running the file, and it is locked.  That
	 * means that it just started running this guy. Better check again.
	 */
	(void) Checkactive ();
    }
    if (q->q_daemon) {
	(void) fprintf (stdout, "Killing off %s server %d.\n",
			q->q_server, q->q_daemon);
	
	user_to_root();
	r = killpg (q->q_daemon, SIGINT);
	root_to_user();
	if (r < 0) {
	    if (errno != ESRCH)
		logerr (XLOG_CRIT, "killpg");
	    else
		DEBUG2 ("server %s (%d) was not alive", q->q_server, q->q_daemon);
	}

    } else {
	/* if a server was present and it has been killed, then the 
         * control- and datafiles have been removed; here we have no 
         * server present, so we use brute force; we simply remove files 
         */
	DEBUG3 ("removing files for job %s", q->q_name);
    }

    q_for_hist = q;
    Remove_job (cfp, q);

    /* remove the dead files of bannering also */
    if (RB) {
	(void) unlink_daemon ("banner_pg");
	(void) unlink_daemon ("banner_epg");
    }
    
    
    /* update history file */
    user_to_daemon ();
    if (create_history_entry (5, q_for_hist) ) {
	log (XLOG_WARNING, "Local_do_job: errors during create_history_entry");
    }
    daemon_to_user ();
    r = fclose (cfp);
    assert (r==0);
}

/***************************************************************************
 * rmjob()
 *  1. get the printcap entries
 *  2. get the queue entries
 *  3. get the active server files and see if they are in queue
 *  4. scan the queue, checking each job for removal
 *
 *  user   local  job   from  remove_all    perms inlist    remove?
 *  ===============================================================
 *  root   yes    *     *     yes           *     *         yes
 *  root   yes    *     *     no            *     yes       yes
 *  root   no     *     mach  yes           *     *         yes
 *  root   no     *     mach  no            *     yes       yes
 *  user   -      user  mach  -             R     yes       yes
 *  user   -      *     *     -             C     yes       yes
 *
 *  5. remove the job;  this may necessitate stopping or killing a deamon
 ***************************************************************************/

void
rmjob (void) {
    int i;			/* ACME Integers, Inc. */
    struct plp_queue *q;	/* job entry */
    int perms;			/* hold perms values */
    int control_perms;		/* has control perms */
    FILE *afp;			/* attach file pointer */
    char afname[MAXPARMLEN];	/* attach file name */
    short att_mark = 0;         /* Flag if a printer is attached to another printer */

    /*
     * If explicitly asked for debug then echo messages to stdout otherwise (db flag in
     * printcap) just log messages in file.
     */
    Echo_on_stdout = Debug;

    /*
     * get the printcap entry
     */
reattach:			/* restart here if printer attached */
    if (Get_pc_entry (Printer, Status_pc_vars, Status_pc_len) == 0) {
	(void) fprintf (stdout, "Printer or Queue %s does not exist!\n", Printer);
	(void) fflush (stdout);
	return;
    }
    if (SD == 0 || *SD == 0) {
	DEBUG2 ("not a Printer");
	return;
    }
    chdir_SD ();
    /*
     * check to see if attached to another printerq, alter printer if attached
     */
    if (*Attach_file && ((afp = fopen_daemon (Attach_file, "r")))) {
	if (fscanf (afp, "%s", afname) == 1) {
	    if (strsame (afname, Printer)) {
		fatal (XLOG_INFO, "Printer %s attached to itself", Printer);
	    }
	    (void)strcpy (unattached_printer, Printer);
	    Printer = afname;
	    att_mark = 1;
	}
	DEBUG2 ("printer reattached to %s", Printer);
	fclose (afp);
	goto reattach;
    }
    /*
     * set the flags needed
     */
    Is_local = (hostcmp (FQDN, Host) == 0);
    Is_root = strsame (Person, "root");
    remove_all = (Parmcount > 0 && strsame (Parms[0].str, "-all"));
    /*
     * check to see that the user has RMJOB privs on this machine
     */
    perms = 'R';		/* must be able to at least use the Printer */
    if (!Is_root &&
	Checkperm (FQDN, Person, First_name, &perms, (int *) 0, 0) <= 0) {
	(void) fprintf (stdout, 
		"Sorry user %s, but you do not have remove permission on '%s'!",
			Person, First_name);
	log_request ("%s@%s: remove job '%s' from queue '%s' failed: no permission",
		     Person, From, (Parmcount > 0) ? Parms[0].str : "unspecified",
		     First_name);
	DEBUG4 ("rmjob: root %d, local %d, rm_all %d (failed)",
		 Is_root, Is_local, remove_all);
	return;
    }

    perms = 'C';		/* check for control perms */
    if (!Is_root && Checkperm (FQDN, Person, First_name, &perms, (int *) 0, 0) <= 0)
    {
	control_perms = 0;
    } else {
	control_perms = 1;
    }

    DEBUG4 ("rmjob: Is_root %d, Is_local %d, remove_all %d, control_perms %d",
	     Is_root, Is_local, remove_all, control_perms);
    /*
     * check for remote machine and networked file system
     */
    if (RM && NW) {
	Remote_remove ();
	return;
    }
    /*
     * get the job queue
     */
    Jobcount = Getq ();
    /*
     * Checkactive checks for the active jobs
     */
    (void) Checkactive ();

    (void) Print_queuehead (LO_statb.st_mode,
			    att_mark ? unattached_printer : NULL);

    /*
     * run down list
     */
    for (i = 0; i < Jobcount; ++i) {
	 q = &Queue[i];
	 DEBUG5 ("rmjob: checking %s, %d, %s", q->q_name, q->q_num, q->q_user);

	 if (shouldremove (q, control_perms)) {
	     printf ("%s: Removing job %d [owner: %s, host: %s, jobname: %s",
		     Printer, q->q_num, q->q_user, &q->q_from, q->q_data->name);
	     if (q->q_entries > 1) {
  	         printf (" ... ].\n");
	     } else {
	         printf ("].\n");
	     }
	     fflush (stdout);

	     if (Log_LPRMs) {
		 log_request ("lprm job %d from queue %s for %s@%s",
			      q->q_num, Printer, Person, FQDN);
	     }
	     doremove (q);
	 } else {
	     if (Match_entry (q)) {
		 (void) fprintf (stdout, "Sorry user %s, but you do not have permission to remove job %d.\n", Person, q->q_num);
	     }
	 }
    }
    /*
     * check for remote machine
     */
    if (RM) {
	Remote_remove ();
    }
    /*
     * give the server a kick
     */
    (void) Startserver ();
}
