/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1987, 1988, Patrick Powell
 * version 3.3.0 Justin Mason July 1994
 ***************************************************************************
 * MODULE: Setup_filter.c
 *************************************************************************** 
 * create the command line to invoke a filter
 *  format: the format type of the data the filter will be invoked with
 *           (a single (char) letter identifies the type).
 *  filtername: an string containing the filter Name and arguments
 *
 * Actions:
 *  creates a string which contains the command to invoke a filter.
 *
 *  If the filtername contains %-escapes (eg. %J, %h, etc.) these will
 *  be filled in with the appropriate values. Otherwise, the command
 *  will be of the form:
 *
 *  filtername arguments \   <- from filtername
 *      -P'Printer' -wwidth -llength -xwidth -ylength [-c] [-iindent] \
 *	[-Z'options'] [-C'class'] [-J'job'] [-R'accntname'] \
 *      [-N'filename'] [-S'"printer descr"'] [-Y'permsline'] \
 *      -n'login' -h'Host' -Fformat [affile]
 *
 *  The 'o' (of filter) only gets the -w, -l, -x, -y and -F options
 *************************************************************************/

#include "lp.h"
#include "common/setup_filter.h"
#include "library/errormsg.h"
  
static char *add_stropt (char *s, char *e, char *s1, char *s2);
static char *add_numopt (char *s, char *e, char *s1, int n);
static char *add_chropt (char *s, char *e, char *s1, int n);
static char *filter_cmd;
static char job_format;
static char SEQNUMBER[4];

static char *
do_percent (char *s, char *e, char type) {
    switch (type) {
	/* two printers: one the user invoked, the second is
	 * that printer's canonical name for accounting.
	 */
	case 'p': s = add_stropt (s, e, NULL, Printer); break;
	case 'P': s = add_stropt (s, e, NULL, First_name); break;
	case 'a': s = add_stropt (s, e, NULL, AF == NULL ? "" : AF); break;
	case 's': s = add_stropt (s, e, NULL, PS == NULL ? "" : PS); break;
        case 'u': s = add_stropt (s, e, NULL, BNRNAME); break;
	case 'l': s = add_numopt (s, e, NULL, PL); break;
	case 'x': s = add_numopt (s, e, NULL, PX); break;
	case 'y': s = add_numopt (s, e, NULL, PY); break;
	case 'm': s = add_numopt (s, e, NULL, CO); break;	/* cost */
	case 'q': s = add_stropt (s, e, NULL, SEQNUMBER); break;
	case 'w': s = add_numopt (s, e, NULL, atoi (PWIDTH)); break;
	case 'n': s = add_stropt (s, e, NULL, LOGNAME); break;
	case 'h': s = add_stropt (s, e, NULL, FROMHOST); break;
	case 'i': s = add_stropt (s, e, NULL, INDENT); break;
	case 'R': s = add_stropt (s, e, NULL, ACCNTNAME); break;
	case 'J': s = add_stropt (s, e, NULL, JOBNAME); break;
	case 'C': s = add_stropt (s, e, NULL, CLASSNAME); break;
	case 'Z': s = add_stropt (s, e, NULL, ZOPTS); break;
	case 'N': s = add_stropt (s, e, NULL, FILENAME); break;

	/* comment from printcap entry */
	case 'S': s = add_stropt (s, e, NULL, get_pr_commnt()); break;

	/* line of perms file where match took place */
	case 'Y': s = add_numopt (s, e, NULL, prmsline); break;

	case 'c': s = add_stropt (s, e, NULL, job_format == 'l' ? " -c" : ""); break;
	case 'F': s = add_chropt (s, e, NULL, job_format); break;

	default:
	    log (XLOG_WARNING, "%%%c unknown in filter string", type);
    }
    return s;
}

char *
Setup_filter (int format, char *filtername) {
    static char cmd[BUFSIZ];	/* the command buffer */
    char *bp, *ep;		/* buffer pointer and general purpose pointer */
    char *s;			/* pointer into printcap string */
    int found_percent;		/* flag */
    char *time_buf;             /* temporary time buffer string */

    assert (filtername && *filtername);

    bp = cmd;		/* set up end of buffer pointer for error detection */
    ep = cmd + sizeof (cmd);

    filter_cmd = filtername;	/* for do_percent() */
    job_format = format;	/* ditto! */

    /* defaults */
    if (PWIDTH[0] == '\0') {
	(void) sprintf (PWIDTH, "%d", PW);
    }
    if (Queue_tmp != NULL) {
        ;
    } else {
        Queue_tmp = Queue;
      /*	SEQNUMBER[0] = '\0'; */
    }
   (void) sprintf (SEQNUMBER, "%-3d", Queue_tmp->q_num);

    found_percent = False;
    for (s = filtername; *s && (bp < ep); s++) {
	if (*s == '%') {
	    found_percent = True; 
	    s++;
	    bp = do_percent (bp, ep, *s);
	} else {
	    *bp++ = *s;
	}
    }
    
    /* @@@ Fix from Dave Atkin */
    *bp = 0;     /* ensure cmd is properly terminated */
    /* end fix */

    if (found_percent) {		/* we don't append any more options */
	return (strdup (cmd));
    }

    /* otherwise, stick on the old PLP set.
     * DO NOT ADD NEW OPTIONS TO THIS SET -- THEY ARE ONLY PROVIDED FOR
     * COMPATIBILITY!!
     */

    bp = estrcp (cmd, filtername, ep);
    bp = add_stropt (bp, ep, " -P", First_name);
    bp = add_numopt (bp, ep, " -w", atoi(PWIDTH));
    bp = add_numopt (bp, ep, " -l", PL);
    bp = add_numopt (bp, ep, " -x", PX);
    bp = add_numopt (bp, ep, " -y", PY);
    bp = add_stropt (bp, ep, " -N", FILENAME);
    bp = add_stropt (bp, ep, " -S", get_pr_commnt());
    bp = add_numopt (bp, ep, " -Y", prmsline);
#ifdef ENH_FILTER_CMDS
    if (format != 'o') { /* these parameters are NOT for the OF filter! */
	malloc_or_die (time_buf, strlen (ctime (&Queue_tmp->q_time) + 1));
	time_buf = ctime (&Queue_tmp->q_time);
	time_buf[24] = '\0';   /* strip the \n character */
	bp = add_stropt (bp, ep, " -D", time_buf); /*spooled time */
	spooled_time = Queue_tmp->q_time;
	/* we need this informations from the job BEFORE the actual one */
	if (BB && HL && BP && *BP && !no_bb_banner) {
	    bp = add_stropt (bp, ep, " -q", seqnumber_bb);
	    bp = add_stropt (bp, ep, " -n", username_bb);
	    bp = add_stropt (bp, ep, " -h", hostname_bb);
	    bp = add_stropt (bp, ep, " -J", jobname_bb);
	    bp = add_stropt (bp, ep, " -L", bnrname_bb);
	} else {
	    bp = add_stropt (bp, ep, " -n", LOGNAME);
	    bp = add_stropt (bp, ep, " -h", FROMHOST);
	    bp = add_stropt (bp, ep, " -J", JOBNAME);
	    bp = add_stropt (bp, ep, " -L", BNRNAME);
	    bp = add_chropt (bp, ep, " -F", format);
	    bp = add_stropt (bp, ep, " -C", CLASSNAME);
	    bp = add_stropt (bp, ep, " -q", SEQNUMBER);
	}
    }
#endif /* ENH_FILTER_CMDS */

    if (format == 'o') {	/* of format */
	bp = add_chropt (bp, ep, " -F", format);
	if (bp == 0) {
	    log (XLOG_INFO, "filter command too long '%s'", cmd);
	    return False;
	}
	return (strdup(cmd));
    }
    /*
     * -PPrinter -wwidth -llength -xwidth -ylength [-c] [-iindent] [-Zoptions]
     * [-Cclass] [-Jjob] -nlogin -hHost -Fformat [affile]
     */
    if (format == 'l') {	/* add literal flag */
	bp = add_stropt (bp, ep, " -c", "");
    }
    if (INDENT[0]) {
	bp = add_stropt (bp, ep, " -i", INDENT);
    }
    if (ZOPTS[0]) {
	bp = add_stropt (bp, ep, " -Z", ZOPTS);
    }
    if (ACCNTNAME[0]) {
	bp = add_stropt (bp, ep, " -R", ACCNTNAME);
    }
    if (JOBNAME[0]) {
	bp = add_stropt (bp, ep, " -J", JOBNAME);
    }
    if (CLASSNAME[0]) {
	bp = add_stropt (bp, ep, " -C", CLASSNAME);
    }
    bp = add_stropt (bp, ep, " -n", LOGNAME);
    bp = add_stropt (bp, ep, " -h", FROMHOST);
    /*
     * accounting_filename
     */
    bp = add_chropt (bp, ep, " -F", format);
    if (AF && *AF) {
	bp = add_stropt (bp, ep, " ", AF);
    }
    if (bp == NULL) {
	log (XLOG_INFO, "filter command too long '%s'", cmd);
    }
    return (strdup (cmd));
}

/* enhanced for security by jmason. This now wraps the supplied argument
 * in ticks (ie. -J'(stdin)') before passing it to sh for the exec; it
 * also checks the argument to see if it, in turn, contains ticks.
 * If it does, they are prepended with a backslash character (\).
 * PLP's safe-exec routines grok this correctly.
 *
 * This should stop users from being able to get lpd to exec arbitrary
 * command-lines as daemon, while still allowing a large degree of freedom
 * to filter writers. (however, filter writers still need to exercise
 * care, otherwise the "daemon" userid is vulnerable).
 */
static char *
add_stropt (char *s, char *e, char *s1, char *s2) {
    if (s1) {
	s = estrcp (s, s1, e);
    }
    *(s++) = '\'';		/* wrap the argument in ticks (ie. -J'(stdin)'). */
    while (*s2 && (s < e)) {
	if (*s2 == '\'') {
	    *(s++) = '\\';
	    if (s >= e)
		return NULL;
	}
	*(s++) = *s2++;
	if (s >= e)
	    return NULL;
    }
    if (s >= e)
	return NULL;

    *(s++) = '\''; *s = '\0';
    return (s);
}

static char *
add_numopt (char *s, char *e, char *s1, int n) {
    char b[64];			/* a 64-decimal-place int? I hope not! ;) */
    (void) sprintf (b, "%d", n);
    assert(strlen(b)<sizeof(b));
    if (s1)  /* Avoid crash */
       s = estrcp (s, s1, e);
    s = estrcp (s, b, e);
    return (s);
}

static char *
add_chropt (char *s, char *e, char *s1, int n) {
    char b[2];
    b[0] = n;
    b[1] = '\0';
    if (s1)  /* Avoid crash */
      s = estrcp (s, s1, e);
    s = estrcp (s, b, e);
    return (s);
}
