/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1994, Bjarne Steinsbo
 ***************************************************************************
 * MODULE: Filestack.c
 * functions to manipulate a stack of open files
 ***************************************************************************/

#include "lp.h"
#include "library/errormsg.h"
#include "library/filestack.h"
#include "library/readconf.h"

#define	INCLUDE_KEYWORD		"include"
#define	INCLUDE_KEYWORD_LEN	7

#define MAX_DEPTH		10	/* Max recursion depth on includes */

/*
 * NOTE!  Only one stack maintained.  This can be shared for different
 * purposes, but not at the same time.
 */

static FILE *file_stack[MAX_DEPTH];	/* A stack of files */
static int file_tos = 0;		/* Top of this stack */

/*
 * File_push (&file, new_filename, path)
 *
 * Push file on stack, fopen new filename, if successfull, set file to 
 * recently fopened file.
 *
 * Returns: 1 = fopen was successful, 0 = fopen failed
 */

int
File_push (FILE **file, char *fname, char *path) {
    FILE *new = NULL;
    char *step;
    static char full_name[MAXPATHLEN];

    if (fname == 0 || *fname == 0)
	return False;

	DBGPCAP8 ("File_push: file name '%s'", fname);

    if (file_tos >= MAX_DEPTH)
	fatal (XLOG_INFO, "File_push: stack is full. Recursion?");

    fname = reallocstrcpy (0, fname, 0);
    path = reallocstrcpy (0, path, 0);
    if (fname[0] == '/') {
	/* No substitutions on absolute path names */
	new = fopen (fname, "r");
    } else {
	fname = Expand_str (fname);
	step = (char *) strtok (path, ":");
	do {
	    strcpy (full_name, step);
	    strcat (full_name, "/");
	    strcat (full_name, fname);
		DBGPCAP8 ("File_push: try filename '%s'", full_name);
	    new = fopen (full_name, "r");
	    if (new) {
		break;
	    }
	} while ((step = strtok (NULL, ":")) != NULL);
    }
    if (new == NULL) {
	log (XLOG_INFO, "include: could not open '%s'", fname);
    } else {
	file_stack[file_tos++] = *file;
	*file = new;
    }
    free (fname); free (path);
    return (new != NULL);
}

/*
 * File_pop ()
 *
 * Pop file from stack.
 *
 * Returns: The last pushed file, or NULL if stack is empty.
 */

FILE *
File_pop (void) {
	DBGPCAP8 ("File_pop TOS = %d", file_tos);
    if (file_tos <= 0)
	return NULL;
    return file_stack[--file_tos];
}

/* Gets_strlen is for external use -- it should help avoid using strlen(). */
unsigned int Gets_strlen;

/*
 * Special version of fgets where "include"ing other files is invisibly
 * taken care of, and leading and trailing whitespace are stripped.
 * Also drops the newline.
 */
char *
Gets (char *buf, int n, FILE **file, char *path) {
    char *s;
    FILE *f;

    assert (*file != NULL);

    while ((s = fgets (buf, n, *file)) != NULL) {
	while (*s && (*s == ' ' || *s == '\t'))
	    s++;				/* Skip leading whitespace */

	if (s != buf) {
            char *to = buf;
            /* we need to shorten the string by shuffling the
             * non-whitespace chars to the left. */

            while (*s != '\0')
		*to++ = *s++;

	    *to = '\0';
	    s = buf;
	}

	while (*s != '\0') {
	    if ((*s == '\n') ||			    /* get the newline here */
		(*s == '#' && (s == buf ||	    /* comment at start of line */
		s[-1] == ' ' || s[-1] == '\t')))    /* preceded by whitespace */
	    {
		*s = '\0';
		break;
	    }
	    s++;
	}

	if (s > buf) {
	    int len = (s - buf) - 1;

	    s = buf;
	    while ((len >= 0) && ((s[len] == '\t') || (s[len] == ' ')))
		s[len--] = '\0';	    /* Strip whitespace at end of line */
 
	    assert (len >= -1);
	    Gets_strlen = (unsigned int) len + 1; /* for external use */
	    assert (Gets_strlen == strlen (s));

	} else {
	    s = buf;
	    Gets_strlen = 0;
	}

	/* Check for "include" at beginning of line */
	if ((*s == INCLUDE_KEYWORD[0]) &&
		!strncmp (s, INCLUDE_KEYWORD, INCLUDE_KEYWORD_LEN))
	{
	    s += INCLUDE_KEYWORD_LEN;
	    while (*s && (*s == ' ' || *s == '\t'))
		s++;
	    File_push (file, s, path);

	} else if (*s) {			/* don't end on blank lines */
	    break;
	}
    }

    /* Check for files on stack at EOF */
    while (!s) {
	f = File_pop ();
	if (!f) {
	    break;
	} else {
	    fclose (*file);
	    *file = f;

	    return (Gets (buf, n, file, path));		/* redo from start */
	}
    }
    return s;
}
