/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1987, 1988, Patrick Powell
 ***************************************************************************
 * MODULE: printcap.c
 * Super slick printcap information extraction routines.
 * Note: these routines were designed after studying the CURSES routines,
 * which are HORRIBLE...  These may not be any better, but at least
 * they are a lot faster.
 *
 * The design is based on the assumption that we will know all the printcap
 * entries that we want to know ahead of time,  and get them en masse.
 * Furthermore,  these can be placed in sorted order.
 *
 * Thus,  given a set of these,  we scan the printcap entry, extract the
 * field name,  determine if in the list,  and then update the information
 * only if in list.  Result:
 * 1.  1 scan of printcap file per lookup of printer name, finds K entries
 * 2.  Log N order search of list,  for a K log N search of printcap
 *
 * dimakop@cti.gr enhancement:
 * Added Pr_comment variable to keep the printer description and pass it
 * to the banner filter later on when that is invoked.
 ***************************************************************************/

#include "lp.h"
#include "library/errormsg.h"
#include "library/printcap.h"

#define NUM_PCAPS	16	/* basic # of printcaps (dynamic) */
#define SIZE_PCAP	256	/* basic size of All_printers list (dynamic) */

/*
 * char **All_printers() returns a pointer to an array of strings which 
 * contains the names of the printers in the printcap file.  The list 
 * is terminated with a 0 (NIL) entry.
 * 
 * char *First_printer() returns a pointer to the first printer in 
 * the printcap file.
 * 
 * initpcvars( PC_ENTRY pc[]; int len) set the default values for variables
 * listed in pc[]
 * 
 * The following routines return 1 if printcap entry found, 0
 * otherwise. of the printcap entry.
 * 
 * int Set_pc_entry(char *printer; PC_ENTRY pc[]; int len) set the
 * values of the variables listed in pc from the printcap 
 * 1. scan the printcap file and extract the printcap
 *    entry for printer.  Make a copy of the entry. 
 * 2. find the name of the printer (first entry) and set the First_name 
 *    variable to point to it. 
 * 3. Initialize the printcap variables according to the defaults in 
 *    the "pc" array. 
 * 4. Scan the printcap entry, and 
 *    for each tagged field in the entry: 
 *    a. search the "pc" array for the tag using a binary search. 
 *    b. if the entry is found, determine the type; 
 *    c. if an integer, convert 
 *       from string to integer and update variable. 
 *    d. if a flag, set or clear it 
 *    e. if a string, copy into a buffer and set the string variable 
 *       to point to it.
 *
 * int
 * Get_pc_entry(char *printer; PC_ENTRY pc[]; int len) call initpcvars() 
 * and then Set_pc_entry.
 */

#ifdef HESIOD
static int hes_flag = False;
#endif
#ifdef NIS
static int nis_flag = False;
#endif

/*
 * The strbuf holds strings
 */
static char strbuf[BUFSIZ];	/* buffer to hold strings */
static char *strloc;		/* next string location */

static char *pr_names;		/* buffer to hold them */
static int max_names = 0;	/* max number of names in printcap */
static int max_list;		/* size of list */
static char *el, *nl;		/* pointers into the buffers */

char **pr_list = NULL;		/* printer names in printcap */
int num_pcs;			/* number of printers in printcap */

static char Pr_comment[255];    /* keeps the printer description from the
				 * printcap file in order to be passed to
				 * the banner filter. */

static char **last_continuation = NULL;
static int last_cont_index, last_cont_size = 4;

static int Set_pc_entry (char *, PC_ENTRY[], int);

#if (defined(NIS) || defined(HESIOD)) && defined(CROSS_DOMAIN)
char *out_domain = NULL;	/* lookup was in the local domain */
int change_domain = True;	/* allow changing of out_domain */

/* Check if match is in local domain and set out_domain if not. */
void
set_domain (char *where, char *domain) {
    if (change_domain) {
	if (out_domain)
	    free (out_domain);

	if (strcasecmp (where, domain) == 0) {
	    out_domain = NULL;
	} else {
	    out_domain = (char *) malloc ((unsigned) strlen (where) + 1);
	    strcpy (out_domain, where);
	}
    }
    return;
}

#endif

/* makes a copy of str in strbuf and returns the start.
 * Side Effect: modifies strloc.
 */
static char *
fix_str (char *str) {
    char *cp;			/* ACME Pointers, Inc. */
    int l;

    l = strlen (str);
    cp = strloc;
    if (!((strloc + l) < (strbuf + sizeof (strbuf)))) {
	fatal (XLOG_INFO, "fix_str: string buffer overflow");
    }
    (void) strcpy (strloc, str);
    strloc += (l + 1);
    return (cp);
}

/*
 * initpcvars (PC_ENTRY pc_vars[]; int pc_len) initialize the variables in pc_vars
 * with their default values. sets the strloc to start of string buffer.
 */
void
initpcvars (PC_ENTRY pc_vars[], int pc_len) {
    int i;			/* ACME Integer, Inc. */
    /*
     * intialize with the defaults
     */
    First_name = 0;
    All_names = 0;
    Pr_comment[0] = '\0';
    for (i = 0; i < pc_len; ++i) {
	switch (pc_vars[i].kind) {
	case PC_NUM:
	    *(int *) pc_vars[i].var = pc_vars[i].idefault;
	    break;
	case PC_LONG:
	    *(long *) pc_vars[i].var = pc_vars[i].idefault;
	    break;
	case PC_FLAG:
	    *(int *) pc_vars[i].var = pc_vars[i].idefault;
	    break;
	case PC_STRING:
	    *(char **) pc_vars[i].var = pc_vars[i].sdefault;
	    break;
	default:
	    fatal (XLOG_INFO, "getpcvars; bad kind %d in table entry %d",
		   pc_vars[i].kind, i);
	}
	pc_vars[i].set = False;	/* only been set to the default value */
    }
    strloc = strbuf;
    DBGPCAP7 ("initpcvars: done");
    return;
}

static void
Fix_pc_entry (char *name, PC_ENTRY pc_vars[], int pc_len, char *capbuf, int i) {
    char *end_of_names;
    char ep, *bp;

    end_of_names = strchr (capbuf, ':');
    if (end_of_names == NULL) {
	fatal (XLOG_CRIT, "bad printcap: '%s'", capbuf);
    }
    *end_of_names = '\0';

    if (First_name == NULL) {
	bp = strchr (capbuf, '|');
	if (bp) {
	    ep = *bp;
	    *bp = '\0'; First_name = fix_str (capbuf);
	    *bp = ep;
	} else {
	    First_name = fix_str (capbuf);
	}
    }
    if (All_names == NULL) {
	All_names = fix_str (capbuf);
    }
    *end_of_names = ':';
    
    /*
     * get the printcap variable values
     */
    getpcvars (pc_vars, pc_len, capbuf, DO_CONT);

#if (defined(NIS) || defined(HESIOD)) && defined(CROSS_DOMAIN)
    /*
     * Cross-domain printing support for NIS and Hesiod.
     */
    if (out_domain) {
	int i;
	change_domain = False;	/* printrelay doesn't affect out_domain */
	DBGPCAP5 ("Pix_pc_entry: printer %s in remote domain %s", name, out_domain);

	for (i = 0; i < num_confs; i++) {
	    if (
#ifdef HESIOD
		(hes_flag && Set_pc_entry_hes (conf[i].domain, "printrelay", capbuf)) ||
#endif
#ifdef NIS
		(nis_flag && Set_pc_entry_nis (conf[i].domain, "printrelay", capbuf)) ||
#endif
		False)
	    {	
		/* override printcap variables */
		getpcvars (pc_vars, pc_len, capbuf, DO_CONT);
		if (RP && !strchr (RP, '@') &&
		    Printer && strchr (Printer, '@'))
		{
		    char buf[BUFSIZ];	/* add domain to RP */
		    strcpy (buf, RP);
		    strcat (buf, strchr (Printer, '@'));
		    RP = fix_str (buf);
		}
	    }
	    change_domain = True;
	}
    }
#endif
    return;
}

/*
 * Get_pc_entry (char *name; PC_ENTRY pc_vars[]; int pc_len) 1. calls initpcvars() to
 * initialize variables 2. calls Set_pc_entry() to set them
 */
int
Get_pc_entry (char *name, PC_ENTRY pc_vars[], int pc_len) {
    int retval;

    DBGPCAP2 ("Get_pc_entry: get printcap entry for '%s'", name);
    initpcvars (pc_vars, pc_len);
    last_cont_index = 0;
    retval = Set_pc_entry (name, pc_vars, pc_len);
    endpcvars (pc_vars, pc_len);
    if (NU)
	return False;
    return retval;
}

/*
 * version of Get_pc_entry used by multiple-device queues.
 * If fromname is NULL, use the SS value from the mainname entry.
 */

int
Get_multi_pc_entry (char *mainname, char *fromname, PC_ENTRY pc_vars[], int pc_len) {
    int retval;

    initpcvars (pc_vars, pc_len);
    last_cont_index = 0;
    if (fromname != NULL) {
	DBGPCAP2 ("Get_multi_pc_entry: get printcap entry for '%s' (based on '%s')",
		mainname, fromname);
	retval = Set_pc_entry (mainname, pc_vars, pc_len);
	retval &= Set_pc_entry (fromname, pc_vars, pc_len);
	NU = False;

    } else {
	DBGPCAP2 ("Get_multi_pc_entry: get printcap entry for '%s' (based on SS)", 
		  mainname);
	retval = Set_pc_entry (mainname, pc_vars, pc_len);
	retval &= Set_pc_entry (SS, pc_vars, pc_len);
	NU = False;
    }
    endpcvars (pc_vars, pc_len);
    return retval;
}

static void
fill_out_comment (char *start, char *end) {
    int right, left;
    char *lastpipe;

    lastpipe = strrchr (start, '|');		/* last | in list */
    left = lastpipe ? ((int) (lastpipe - start)) : -1; /* +1 in strncpy() */
    right = end ? ((int) (end - start)) : 0;

    if (left < right) {
	(void) strncpy (Pr_comment, &start[left+1], right-left-1);
	Pr_comment [right-left-1] = '\0';

    } else {
	Pr_comment[0] = '\0';
    }
    return;
}

/* take a peek into the printcap parameters looking for an
 * "oh" (only on host) parameter.
 */
static int
AllowedByPrintcap (char *printer, char *names, char *pcapEntry) {
    char *s, *next_sep;
    if ((s = strstr (pcapEntry, "oh=")) != NULL) {
	if (s == pcapEntry || s[-1] == ':') {
	    next_sep = strchr (s, ':');
	    if (next_sep) *next_sep = '\0';
	    if (hostcmp (s + 3, ShortHost)) {
		DBGPCAP5 ("AllowedByPrintcap: can't use this entry: '%s'", s);
		return False;		/* not for use on this host */
	    }
	    if (next_sep) *next_sep = ':';
	}
    }
    return True;
}

/*
 * Chk_name deals with name matching.  The first line of the printcap 
 * entry is a sequence of names separated by |'s, so we compare against 
 * each such name.  
 * The normal : terminator after the last name (before the first field) 
 * stops us.
 *
 * This function also fills variable Pr_comment in with the printer 
 * description from the printcap file, ie. the last alias for the printer; 
 * by convention this is a brief description, although PLP also provides 
 * the CM field for this. I'm handling this as follows: 
 * if CM is set, use that instead.
 */
int
Chk_name (char *buf, char *name) {
    char *end_of_names, *start_of_names;
    unsigned int l;

    l = strlen (name);

    start_of_names = buf;
    end_of_names = strchr (buf, ':');

    if (end_of_names == NULL) {
	/* nothing on this line but crap -- ignore it */
	return False;
    }
    *end_of_names = '\0';

    while (buf && *buf) {
	if (strncmp (name, buf, l) == 0) {
	    char *chk;			/* check for trailing whitespace */
	    int failed_check;

	    failed_check = False;
	    for (chk = &(buf[l]); !failed_check; chk++) {
		if (ISSPACE(*chk)) {
		    continue;		/* trailing whitespace, ignore */

		} else if (*chk == '\0' || *chk == '|') {
		    if (!AllowedByPrintcap (name, start_of_names, end_of_names + 1))
		    {	/* oops -- we must have an "oh" or something in there */
			failed_check = True;
			continue;
		    }
		    *end_of_names = ':';
		    fill_out_comment (start_of_names, end_of_names);
		    return True;		/* match! */

		} else {
		    failed_check = True;	/* not a match, non-white char */
		}
	    }
	}
	if ((buf = strchr (buf, '|')) != NULL) {
	    buf++;
	}
    }

    *end_of_names = ':';
    return False;			/* no match */
}

/*
 * Return the (numeric) option id. 
 * Numeric options look like li#80 i.e. the option string
 * is separated from the numeric value by a # character.  If the 
 * option is not found we return -1. Note that we handle octal numbers 
 * beginning with 0.
 */
static inline int
getnum (char *bp) {
    int i, base;

    if (*bp != '#') {
	log (XLOG_INFO, "warning: param not in numeric format: '%s'", bp - 2);
    }
    bp = bp + 1;
    i = 0;
    base = 10;
    if (*bp == '0') {
	base = 8;
    }
    while (isdigit (*bp)) {
	i = i * base + (*bp - '0');
	++bp;
    }
    return (i);
}

/*
 * same as getnum but for unsigned longs.
 */
static inline long
getulong (char *bp) {
    unsigned int i, base;

    if (*bp != '#') {
	log (XLOG_INFO, "warning: param not in long format: '%s'", bp - 2);
    }
    bp = bp + 1;
    i = 0;
    base = 10;
    if (*bp == '0') {
	base = 8;
    }
    while (isdigit (*bp)) {
	i = i * base + (*bp - '0');
	++bp;
    }
    return (i);
}

/*
 * Handle a flag option. 
 * Flag options are given "naked", i.e. followed by a :, @ or end
 * of the buffer.  Return 1 if we find the option, or 0 if it is not given.
 */
static inline int
getflag (char *bp) {
    if (*bp == 0 || *bp == ':') {
	return True;
    } else if (*bp != '@') {
	log (XLOG_INFO, "warning: param not in flag format: '%s'", bp - 2);
    }
    return False;
}

/*
 * Get a string valued option. 
 * These are given as xx=string. There is a limited amount of
 * decoding done. \n -> '\n', \nnn -> '\nnn' ^<CHAR> -> CTRL-CHAR
 */
static char *
getstr (char *str) {
    char buf[BUFSIZ];
    char *bp, *cp, *op;		/* ACME Chain and Pointers, Inc */
    int c, i;			/* ACME Integers, Inc */
    static char norm[] = "E^\\:nrtbf";
    static char esc[] = "\033^\\:\n\r\t\b\f";

    bp = str;
    op = buf;			/* output area */
    if ((*bp != '=') && (*bp != '#')) {
	/* we don't mind people using #: some numeric parameters allow 
	 * filenames for the value to be read from, so a string 
	 * format is really used.
	 */
	log (XLOG_INFO, "warning: param not in string format: '%s'", bp - 2);
    }
    bp++;

    while (((c = *bp++) != '\0') && c != ':') {
	switch (c) {
	case '^':
	    c = *bp++;
	    if (c == 0) {
		fatal (XLOG_INFO, "getstr: bad escape string in printcap");
	    }
	    c = c & 037;
	    break;
	case '\\':
	    c = *bp++;
	    if (c == 0) {
		fatal (XLOG_INFO, "getstr: bad escape string in printcap");
	    }
	    cp = strchr (norm, c);
	    if (cp) {
		c = esc[cp - norm];
	    } else if (isdigit (c)) {
		c = c - '0';
		for (i = 0; i < 3 && isdigit (*bp); ++i) {
		    c = 8 * c + *bp - '0';
		    ++bp;
		}
	    }
	    break;
	default:
	    break;
	}
	*op++ = c;
    }
    *op++ = '\0';
    if ((int) strlen (buf) > 0) {
	return (fix_str (buf));
    } else {
	return ((char *) NULL);
    }
}

/*
 * fixentry: fix up the value for the entry in the printcap
 */
static void
fixentry (PC_ENTRY *pc_vars, char *cp) {
    DBGPCAP7 ("fixentry: found %c%c", cp[-2], cp[-1]);
    switch (pc_vars->kind) {
    case PC_NUM:
	if (!pc_vars->set) {
	    *(int *) pc_vars->var = getnum (cp);
	    pc_vars->set = True;
	    DBGPCAP7 ("fixentry: value # %d", *(int *) pc_vars->var);
	}
	break;
    case PC_LONG:
	if (!pc_vars->set) {
	    *(long *) pc_vars->var = getulong (cp);
	    pc_vars->set = True;
	    DBGPCAP7 ("fixentry: value # %d", *(long *) pc_vars->var);
	}
	break;
    case PC_FLAG:
	if (!pc_vars->set) {
	    *(int *) pc_vars->var = getflag (cp);
	    pc_vars->set = True;
	    DBGPCAP7 ("fixentry: value @ %d", *(int *) pc_vars->var);
	}
	break;
    case PC_STRING:
	if (!pc_vars->set) {
	    *(char **) pc_vars->var = getstr (cp);
	    pc_vars->set = True;
	    DBGPCAP7 ("fixentry: value = %s", *(char **) pc_vars->var ?
		    *(char **) pc_vars->var : "(NULL)");
	}
	break;
    default:
	fatal (XLOG_INFO, "fixentry: bad kind in the pc_vars");
	break;
    }
    return;
}

void
endpcvars (PC_ENTRY pc_vars[], int pc_len) {
    if (DefaultPrintcap != NULL) {
	/* append the default printcap */
	if (*DefaultPrintcap != ':') {
	    log (XLOG_INFO, "default-printcap must start with ':'");
	}
	getpcvars (pc_vars, pc_len, DefaultPrintcap, NO_CONT);
    }

    if (DbgPcap > 7)
	show_pc (pc_vars, pc_len);
    return;
}

/*
 * Set_pc_entry( char *name; PC_ENTRY pc_vars[]; int pc_len) 
 * 1. gets the "name" printcap entry; capbuf[] holds it; strbuf is 
 *    used to hold string constants. 
 * 2. calls getpcvar routine to find fields in the printcap entry 
 *    and set variables. 
 *
 * Returns: 1 if successful, and printcap entry found; 
 *          0 otherwise. 
 * Side Effect: sets the First_name to the first name in the printcap 
 * entry and All_names to the list of printcap entry names.
 */

static int
Set_pc_entry (char *name, PC_ENTRY pc_vars[], int pc_len) {
    int retval, i;
    char capbuf[BUFSIZ];
#ifdef HESIOD
    hes_flag = False;
#endif
#ifdef NIS
    nis_flag = False;
#endif

    if (name == NULL || *name == '\0') {
	return False;
    }
    retval = 0;
    for (i = 0; i < num_confs; i++) {
	DBGPCAP2 ("Set_pc_entry: printcap-order: lookup (%d/%d): type %d",
		i+1, num_confs, conf[i].type);

	switch (conf[i].type) {
#ifdef LOCAL
	case LOCAL_TYPE:
	    if (!retval) {
		retval = Set_pc_entry_file (name, capbuf);
		DBGPCAP5 ("Set_pc_entry: Set_pc_entry_wrapper: found=%d", retval);
	    }
	    break;
#endif
#ifdef NIS
	case NIS_TYPE:
	    if (!retval) {
		DBGPCAP2 ("Set_pc_entry: Looking in NIS, domain %s",
			     conf[i].domain ? conf[i].domain : "(null)");
		nis_flag = retval = (int) Set_pc_entry_nis (conf[i].domain, name, capbuf);
	    }
	    break;
#endif
#ifdef HESIOD
	case HESIOD_TYPE:
	    if (!retval) {
		DBGPCAP2 ("Set_pc_entry: Looking in hesiod, domain %s",
			     conf[i].domain ? conf[i].domain : "(null)");
		hes_flag = retval = (int) Set_pc_entry_hes (conf[i].domain, name, capbuf);
	    }
	    break;
#endif
	default:
	    log (XLOG_INFO,
		     "Cannot do a lookup in printcap of type %d for domain %s.\n",
		     conf[i].type, (conf[i].domain ? conf[i].domain : "local domain"));
	}
    }

    if (retval) {
	Fix_pc_entry (name, pc_vars, pc_len, capbuf, i);
    }
    return (retval);
}

/* make a contains a unique continuation identifier, of the form
 * "fromname#toname" where fromname is the printcap entry we came from,
 * and toname is the one we intend to include.
 */
static char *
make_cont_id (char *capbuf, char *name) {
    char *id, *colon;
    unsigned int len;

    if ((colon = strchr (capbuf, ':')) == NULL) {
	colon = strchr (capbuf, '\0');	/* impossible, but just be sure */
    }

    assert (colon > capbuf);
    len = colon - capbuf;	/* length of printer-name section of capbuf */

    malloc_or_die (id, len + strlen (name) + 2);
    strncpy (id, capbuf, len);
    id[len] = '#';
    strcpy (id + len + 1, name);
    return (id);
}

/* avoid infinite looping -- looking up a 'tc' continuation entry
 * sometimes causes the printcap-reading interface (file, nis, whatever)
 * to read the same entries again. This means that we can't simply check
 * the last continuation we read; if there's more than one :tc:
 * entry in the entire printcap _database_, it'll recurse!! We have to
 * keep track of the lot. :(
 */
static void
continuation (char *capbuf, char *name, PC_ENTRY pc_vars[], int pc_len) {
    char *cont_id;
    int i;

    if ((name == NULL) || (*name == '\0')) {
	return;		/* can't :tc: to a null string! */
    }
    cont_id = make_cont_id (capbuf, name);

    for (i = 0; i < last_cont_index; i++) {
	if (strsame (last_continuation[i], cont_id)) {
	    
	    free (cont_id);
	    return;
	}
    }

    chkandrealloc_list (last_continuation, last_cont_index, last_cont_size, 4);
    last_continuation[last_cont_index++] = cont_id;
    DBGPCAP3 ("continuation: jamais vu, including '%s' (%d)",
   	 name, last_cont_index);
    (void) Set_pc_entry (name, pc_vars, pc_len);
    NU = False;		/* we're using it as part of a valid printer entry. */
    endpcvars (pc_vars, pc_len);
}

/*
 * getpcvars- passed a table of entries, tries to find them all
 */
void
getpcvars (PC_ENTRY *pc_vars, int pc_len, char capbuf[], int nocont) {
    int h, l, m;		/* high, low, middle */
    char *cp, *cp_start;	/* string value */
    int dir;			/* direction */
    char *cont;			/* continuation */

    /*
     * now scan the printcap entry, getting the variables
     */
    cp_start = cp = capbuf;
    while (cp && ((cp = strchr (cp, ':')) != NULL)) {
	/*
	 * we are now positioned at the ":"
	 */
	cp = cp + 1;
	if ((cp - cp_start > 1) && (cp[-2] == '\\')) {
	    continue;
	}
	/* note: islower is a macro! do not combine the above statement! */
	if (!islower (cp[0]) || !islower (cp[1])) {
	    continue;
	}
	DBGPCAP9 ("getpcvars: entry %c%c", cp[0], cp[1]);
	/*
	 * check for continuations and recurse if there is one
	 */
	if ((nocont) && (cp[0] == 't') && (cp[1] == 'c')) {
	    cont = getstr (cp + 2);
	    continuation (capbuf, cont, pc_vars, pc_len);

	} else {
	    /*
	     * binary search the table
	     * (set m here to shut up gcc -Wall)
	     */
	    m = l = 0;
	    h = pc_len - 1;
	    dir = 1;
	    while (dir && l <= h) {
		m = (l + h) / 2;
		dir = strncmp (cp, pc_vars[m].pc_name, 2);
		if (dir < 0) {
		    h = m - 1;	/* too high */
		} else {
		    l = m + 1;	/* too low */
		}
	    }
	    if (dir == 0) {
		/* bingo! we found it */
		DBGPCAP9 ("getpcvars: found %c%c%c%c%c",
			 cp[0], cp[1], cp[2], cp[3], cp[4]);
		(void) fixentry (&pc_vars[m], cp + 2);
	    }
	}
	/*
	 * we finished, on to next
	 */
    }
    return;
}

/*
 * Debugging and display information
 */
void
show_pc (PC_ENTRY *pc_vars, int pc_len) {
    int i;

    (void) log (XLOG_DEBUG, "printcap entry %s:", First_name ?
		    First_name : "(no first name)");

    for (i = 0; i < pc_len; ++i) {
	switch (pc_vars[i].kind) {
	case PC_NUM:
	    log (XLOG_DEBUG, "     %s: PC_NUM %d", pc_vars[i].pc_name,
			    *(int *) pc_vars[i].var);
	    break;
	case PC_LONG:
	    log (XLOG_DEBUG, "     %s: PC_LONG %ld", pc_vars[i].pc_name,
			    *(long *) pc_vars[i].var);
	    break;
	case PC_FLAG:
	    log (XLOG_DEBUG, "     %s: PC_FLAG %d", pc_vars[i].pc_name,
	   		    *(int *) pc_vars[i].var);
	    break;
	case PC_STRING:
	    log (XLOG_DEBUG, "     %s: PC_STRING %s", pc_vars[i].pc_name,
			    *(char **) pc_vars[i].var ?
			    *(char **) pc_vars[i].var :
			    "(NULL)");
	    break;
	default:
	    log (XLOG_DEBUG, "     %s: unknown", pc_vars[i].pc_name);
	    break;
	}
    }
    return;
}

/****************************************************************************
 * init_printer_list initialises the list of all printers
 ****************************************************************************/
void
init_printer_list (void) {
    DBGPCAP5 ("init_printer_list: max_names %d, max_list %d", max_names, max_list);
    if (max_names == 0) {
	max_list = SIZE_PCAP;
	max_names = NUM_PCAPS;
	malloc_or_die (pr_names, max_list);
	pr_list = (char **) malloc ((unsigned) (max_names * sizeof (char *)));
	if (pr_list == NULL) {
	    logerr_die (XLOG_INFO, Malloc_failed_msg);
	}
	el = pr_names + max_list;
    }

    nl = pr_names;
    num_pcs = 0;
    pr_list[0] = NULL;
    Pr_comment[0] = '\0';
    last_cont_index = 0;
    return;
}

/****************************************************************************
 * add_printer_list adds a printer name to the printer list
 ****************************************************************************/
void
add_printer_list (char *n) {
    int offset, i;
    char *name, *sepind;

    name = strdup (n);
    if ((sepind = strchr (name, PNAME_SEPCH)) != NULL) {
	*sepind = '\0';
    }

    DBGPCAP4 ("add_printer_list: adding %s", name);
    for (i = 0; i < num_pcs; i++) {	/* check if in list already */
	if (strsame (name, pr_list[i])) {
	    free(name);
	    return;
	}
    }

tryagain:
    pr_list[num_pcs] = nl;
    if ((nl = estrcp (nl, name, el)) != NULL) {
	++nl;
	++num_pcs;
    }

    if (num_pcs >= max_names) {
	DBGPCAP1 ("add_printer_list: need more names");
	max_names += NUM_PCAPS;
	pr_list = (char **)
		realloc ((char *) pr_list, (unsigned) max_names * sizeof (char *));
	if (pr_list == NULL) {
	    logerr_die (XLOG_INFO, Malloc_failed_msg);
	}
    }

    if (nl == NULL || nl >= el) {
	DBGPCAP1 ("add_printer_list: need more space");
	max_list += SIZE_PCAP;

	pr_names = (char *) realloc (pr_names, max_list);
	if (pr_names == NULL) {
	    logerr_die (XLOG_INFO, Malloc_failed_msg);
	}

	offset = (int) pr_names - (int) pr_list[0];
	for (i = 0; i <= num_pcs; i++) {
	    pr_list[i] += offset;
	}
	el = pr_names + max_list;

	if (nl == NULL) {
	    nl = pr_list[num_pcs];
	    goto tryagain;
	} else {
	    nl += offset;
	}
    }
    pr_list[num_pcs] = NULL;
    free(name);
    return;
}

/****************************************************************************
 * extract_pc_entry takes the string held in capbuf and extracts the pc vars
 * and puts the names of the printers into a list
 ****************************************************************************/
void
extract_pc_entry (char capbuf[], int cont) {
    char *cp;			/* ACME Pointer, Inc. */

    initpcvars (Status_pc_vars, Status_pc_len);
    getpcvars (Status_pc_vars, Status_pc_len, capbuf, cont);
    endpcvars (Status_pc_vars, Status_pc_len);

    if (SD != NULL && *SD != '\0') {
	char *end_of_names;

	end_of_names = strchr (capbuf, ':');
	if (end_of_names == NULL) {
	    fatal (XLOG_CRIT, "bad printcap: '%s'", capbuf);
	}
	*end_of_names = '\0';

	if ((cp = strchr (capbuf, '|')) != NULL) {
	    *cp = '\0';
	}
	add_printer_list (capbuf);
    } else {
	DBGPCAP7 ("no sd -- ignoring this printcap");
    }
    return;
}

char **
All_printers (void) {
    int i;

    init_printer_list ();
    for (i = 0; i < num_confs; i++) {
	DBGPCAP2 ("printcap-order: lookup (%d/%d): type %d",
	    i+1, num_confs, conf[i].type);

	switch (conf[i].type) {
#ifdef LOCAL
	case LOCAL_TYPE:
	    All_printers_file ();
	    break;
#endif
#ifdef NIS
	case NIS_TYPE:
	    (void) All_printers_nis (conf[i].domain);
	    break;
#endif
#ifdef HESIOD
	case HESIOD_TYPE:
	    (void) All_printers_hes (conf[i].domain);
	    break;
#endif
	default:
	    log (XLOG_INFO, "Cannot do a lookup in printcap of type %d for domain %s.",
		     conf[i].type, (conf[i].domain ? conf[i].domain : "local domain"));
	}
    }
    if (DbgPcap > 3) {
	int j;
	log (XLOG_DEBUG, "All_printers: %d,", num_pcs);
	for (j = 0; j < num_pcs; ++j) {
	    log (XLOG_DEBUG, " %s", pr_list[j]);
	}
    }
    return (pr_list);
}

char *
First_printer (void) {
    char *retval = NULL;
    int i;

    for (i = 0; i < num_confs; i++) {
	DBGPCAP2 ("First_printer: printcap-order: lookup (%d/%d): type %d",
		i+1, num_confs, conf[i].type);

	switch (conf[i].type) {
#ifdef LOCAL
	case LOCAL_TYPE:
	    if (!retval)
		retval = (char *) First_printer_file ();
	    break;
#endif
#ifdef NIS
	case NIS_TYPE:
	    if (!retval)
		retval = (char *) First_printer_nis (conf[i].domain);
	    break;
#endif
#ifdef HESIOD
	case HESIOD_TYPE:
	    if (!retval)
		retval = (char *) First_printer_hes (conf[i].domain);
	    break;
#endif
	default:
	    log (XLOG_INFO,
		     "Cannot do a lookup in printcap of type %d for domain %s.",
		     conf[i].type, (conf[i].domain ? conf[i].domain : "local domain"));
	}
    }
    if (retval == NULL) {
	fatal (XLOG_INFO, "First printer: bad printcap");
	return (NULL);
    } else {
	DBGPCAP3 ("First_printer: %s", retval);
	return (retval);
    }
}

char *
get_pr_commnt (void) {
    if (CM && *CM) {
	return (CM);		/* "official" comment */
    } else {
	return (Pr_comment);	/* last alias in list */
    }
}
