#include "lp.h"
#include "library/errormsg.h"
#include "library/filestack.h"
#include "library/filecache.h"
#include "library/printcap.h"
#include "library/utils.h"

#ifdef LOCAL
static int getpcent_file (char *tb, int size);
static int openprent (void);
static int nextprent (void);
static void closeprent (void);
static int pc_entry_file (char *buf, int size);

static FILE *pfp = NULL;	/* printcap data base file pointer */

static struct file_cache pc_cache;

/**************************************************************************
 *     finds the next entry in the printcap file;
 *     it places it in the tb buffer, and returns
 *     1 if entry found, 0 otherwise
 * Also associated functions: {open,next,close}prent().
 *************************************************************************/

static char *current_pcfile = NULL;
static char *next_pcfile = NULL;
static char *tokpath = NULL;
static char read_from_cache;

#if (defined(NIS) || defined(HESIOD)) && defined(CROSS_DOMAIN)
extern int change_domain;
#endif

static int
openprent (void) {

    if (Use_printcap_cache) {
	rewind_cache (&pc_cache);

	if (pc_cache.entries[0]) {
	    /* we have a cache, use it. */
	    read_from_cache = True;
	    DBGPCAP5 ("openprent: reading printcaps from cache");
	    return False;
	}
	/* otherwise, we are being called by build_pc_cache (). */
	read_from_cache = False;
	DBGPCAP5 ("openprent: reading printcaps from file");
    }

    if (tokpath)
	free (tokpath);

    if (Printcap_path == NULL || !*Printcap_path) {
	logerr_die (XLOG_INFO, "printcap-path is unset!");
    }

    tokpath = (char *) strdup (Printcap_path);		/* just copy the path. */

    

    /* init the strtok(). */
    next_pcfile = (char *) strtok (tokpath, ":");

    if (next_pcfile == NULL)
	fatal (XLOG_NOTICE, "Empty printcap path");
    return (nextprent ());
}

static int
nextprent (void) {
    do {
	if (next_pcfile == NULL) {
	    DBGPCAP6 ("nextprent: end of path");
	    return True;
	}
	if (pfp != NULL) {
	    (void) fclose (pfp);
	    pfp = NULL;
	}
	if ((pfp = fopen (next_pcfile, "r")) == NULL) {
	    DBGPCAP6 ("nextprent: failed to open %s, skipping", next_pcfile);
	    next_pcfile = (char *) strtok (NULL, ":");

	} else {
	    current_pcfile = next_pcfile;
	    DBGPCAP6 ("nextprent: opened %s", current_pcfile);

	    next_pcfile = (char *) strtok (NULL, ":");
	    return False;
	}
    } while (pfp == NULL);
    return True;	/* the last entr{y,ies} were invalid, end of path */
}

static void
closeprent (void) {
	DBGPCAP5 ("closeprent");

    if (Use_printcap_cache) {
	rewind_cache (&pc_cache);
	if (read_from_cache) {
	    return;
	}
    }

    while (File_pop () != NULL);	/* close included files */

    if (pfp != NULL) {
	(void) fclose (pfp);
	pfp = NULL;
    }
    if (tokpath != NULL) {
	free (tokpath);
	tokpath = NULL;
    }
    next_pcfile = current_pcfile = NULL;
    return;
}

static int
getpcent_file (char *tb, int size) {
    char buf[BUFSIZ];		/* read buffer */
    char *bp, *cp, *ep;		/* next read position */
    int more_pc_left;		/* flag: more of this printcap to be read? */
    int last_eof;		/* run out of files */

    if (current_pcfile == NULL) {
	openprent ();
    }
    last_eof = False;

    /* We read a single printcap entry into the bp buffer. */
    while (!last_eof) {
	if (pfp == NULL) {
	    fatal (XLOG_DEBUG, "cannot open any printcap files!");
	}
	bp = Gets (buf, sizeof (buf), &pfp, Printcap_include_path);
	if (bp == 0) {
	    last_eof = nextprent ();
	    continue;
	}
	break;
    }
    if (last_eof) {
	return False;		/* ran out of files */
    }
    /*
     * read the file in, stripping off the \<newlines>
     */
    ep = tb + size;
    bp = tb;
    while (!last_eof) {

	if (Gets_strlen) {
	    cp = buf + Gets_strlen;
	    more_pc_left = (cp[-1] == '\\');
	    if (more_pc_left) {
		cp[-1] = '\0';
	    }

	} else {
	    more_pc_left = False;
	}

	if ((bp = estrcp (bp, buf, ep)) == NULL) {
	    fatal (XLOG_INFO, "printcap entry too long in file %s: '%s'",
		   current_pcfile, bp);
	}
	if (!more_pc_left) {
	    DBGPCAP9 ("getpcent_file: %s", tb);
	    return True;
	}
	if (pfp == NULL) {
	    fatal (XLOG_WARNING, "cannot read any printcap files!");
	}
	while ((Gets (buf, sizeof (buf), &pfp, Printcap_include_path) == 0)
		&& !last_eof)
	{
	    last_eof = nextprent ();
	}
    }

    log (XLOG_INFO, "bad printcap entry: %s", tb);
    return False;
    /* let the calling code do closeprent(). */
}

/***************************************************************************
 * All_printers: 
 * reads the printcap file, and forms a list of all the printer names in
 * the file.
 ***************************************************************************/
void
All_printers_file (void) {
    char capbuf[BUFSIZ];

    openprent ();
    while (pc_entry_file (capbuf, BUFSIZ))
	extract_pc_entry (capbuf, DO_CONT);
    closeprent ();
    return;
}

/***************************************************************************
 * First_printer: 
 * reads the printcap file and finds the first printer name
 ***************************************************************************/
char *
First_printer_file (void) {
    char capbuf[BUFSIZ];
    char *bp;
    static char first[PRNAMELEN + 1];

    openprent ();
    while (pc_entry_file (capbuf, BUFSIZ)) {
	initpcvars (Status_pc_vars, Status_pc_len);
	getpcvars (Status_pc_vars, Status_pc_len, capbuf, DO_CONT);
	endpcvars (Status_pc_vars, Status_pc_len);

	if (SD != NULL && *SD != '\0') {
	    if ((bp = strchr (capbuf, '|')) == NULL) {
		bp = strchr (capbuf, ':');
	    }
	    if (bp != NULL)
		*bp = '\0';
	    (void) strcpy (first, capbuf);
	    closeprent ();
	    return (first);
	}
    }
    closeprent ();
    return (NULL);
}

/***************************************************************************
 * order of lookup: 
 * printername/hostname printername 
 ***************************************************************************/

int
Set_pc_entry_file (char *name, char *capbuf) {
    int found;
    char tcapbuf[BUFSIZ];

    if (strchr (name, PNAME_SEPCH)) {
        fatal (XLOG_INFO, "printer name '%s' contains '%c'",
                    name, PNAME_SEPCH);
    }
    if (strchr (name, '@')) {
        fatal (XLOG_INFO, "printer name '%s' contains '@'", name);
    }

#if (defined(NIS) || defined(HESIOD)) && defined(CROSS_DOMAIN)
    if (change_domain)
	out_domain = NULL;
#endif

    DBGPCAP5 ("Set_pc_etry_file: doing printcap lookup (file) for printer %s", name);
    openprent ();
    found = False;
    while (pc_entry_file (tcapbuf, BUFSIZ) > 0) {
	DBGPCAP7 ("Set_pc_entry_file: pc_entry %s", tcapbuf);
	if ((*tcapbuf == 'i') && (!strncmp (tcapbuf, "include ", 8))) {
	    log (XLOG_INFO, "missing newline before '%s'", tcapbuf);

	} else if (Chk_name (tcapbuf, name)) {
	    strcpy(capbuf, tcapbuf);
	    found = True;
	    break;
	}
    }
    closeprent ();
    DBGPCAP5 ("Set_pc_entry_file: found printcap entry: %s", found ? "yes" : "no");
    return (found);
}

void
build_pc_cache (void) {
    char buf[BUFSIZ];
    struct file_cache *pc_cache_ref;

    if (!Use_printcap_cache) {
	DBGPCAP6 ("build_pc_cache: printcap caching disabled");
        return;
    }

    pc_cache_ref = &pc_cache; chkandrealloc_cache (&pc_cache_ref);
    flush_cache (&pc_cache);		/* force openprent() to reread files */

    openprent ();
    while (getpcent_file (buf, BUFSIZ)) {
	DBGPCAP7 ("build_pc_cache: caching pc entry %d [%s]", pc_cache.ind, buf);
	pc_cache_ref = &pc_cache; chkandrealloc_cache (&pc_cache_ref);
	cache_addline (&pc_cache, buf);
    }

    cache_addeof (&pc_cache);
    closeprent ();
    return;
}

void
createPcapCache (void) {
    if (!Use_printcap_cache) {
	DBGPCAP6 ("createPcapCache: printcap caching disabled");
        return;
    }
    pc_cache.size = 0;
    build_pc_cache ();
    return;
}

static int
pc_entry_file (char *buf, int size) {
    if (!Use_printcap_cache) {
        return (getpcent_file (buf, size));
    }

    if (pc_cache.size == 0) {
	fatal (XLOG_NOTICE, "printcap not cached!");
    }
    if (pc_cache.entries[pc_cache.ind] == NULL) {
	/* there's nothing more to read */
	rewind_cache (&pc_cache);
	return False;
    }

    (void) strncpy (buf, pc_cache.entries[pc_cache.ind], size - 1);
    
    pc_cache.ind++;

    return True;
}

void
dump_printcap_cache (void)		/* debugging. */
{
    int i;
    for (i = 0; pc_cache.entries[i]; i++) {
	printf ("CACHED (pcap): [%s]\n", &(pc_cache.entries[i][1]));
    }
    return;
}

void
flush_printcap_cache (void) {
    if (!Use_printcap_cache) {
	return;
    }
    DBGPCAP2 ("flush_printcap_cache: flushing printcap cache");
    build_pc_cache ();
    return;
}
#endif				/* LOCAL */
