/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1987, 1988, Patrick Powell
 ***************************************************************************
 * MODULE: startserver.c
 * send a "start" message to the LPD server
 ***************************************************************************
 * Startserver()
 * This originally used to use the UNIX domain sockets,  but now uses the
 * AF_INET domain sockets.  The actual sending is done by calling
 * Remote_start,  after setting the RP (remote printer) and RM fields
 * appropriately.
 * 1. Save the RP and RM values.
 * 2. Set RP to Printer and RM to Host
 * 3. If you are running NW, start the remote one
 * 4. Send the start message.
 * 5. Restore RP and RM
 * 6. Return success indication
 ***************************************************************************/

#include "lp.h"
#include "library/errormsg.h"
#include "library/utils.h"

void
Startserver (void)
{
    char *old_RP, *old_RM;	/* original RP and RM values */

    old_RP = old_RM = NULL;
    if ((NW == False) || (RP == NULL) || (RM == NULL)) {
	old_RP = RP;
	old_RM = RM;
	RP = Printer;
	RM = Host;
    }

    if (Remote_start () == False) {
	/* a friendly message for the user's benefit */
	log (XLOG_INFO,
	    "%s@%s: job spooled, but daemon is dead; tell operator to start daemon.",
	    RP, RM);
    }

    if (old_RP) {
	RP = old_RP;
	RM = old_RM;
    }
}

/*
 * Really really portable implementation of waitpid() -- based on an
 * original implementation nicked from the tcl distribution. This
 * should work on every system that implements wait3() or waitpid()
 * and supports SIGCHLD or SIGCLD with any of the semantics.
 *
 * It needs to handle the SIGCHLD/SIGCLD signal itself.
 *
 * -- Justin Mason <jmason@iona.ie> 94.
 *
 * Here's the comment from the original version:
 * 
 *	This procedure emulates the POSIX waitpid kernel call on
 *	BSD systems that don't have waitpid but do have wait3.
 *	This code is based on a prototype version written by
 *	Mark Diekhans and Karl Lehenbauer.
 *
 * Copyright (c) 1993 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

/*
 * A linked list of the following structures is used to keep track
 * of processes for which we received notification from the kernel,
 * but the application hasn't waited for them yet (this can happen
 * because wait may not return the process we really want).  We
 * save the information here until the application finally does
 * wait for the process.
 */

typedef struct WaitInfo {
    pid_t pid;				/* Pid of process that exited. */
    plp_status_t status;		/* Status returned when child exited
					 * or suspended. */
    struct WaitInfo *nextPtr;		/* Next in list of exited processes. */
} WaitInfo;

static WaitInfo *deadList = NULL;	/* First in list of all dead
					 * processes. */
#ifndef NDEBUG
static int catching_SIGCHLD = False;	/* flag for an assert() */
#endif

/* If a child process has changed status, but we're waiting for some
 * particular process, we should store it for later use.
 * Tricky point: first check for an existing entry for the
 * process and overwrite it if it exists (e.g. a previously
 * stopped process might now be dead).
 */

static void
waitpid_add_to_list (plp_status_t status, pid_t result) {
    register WaitInfo *w;

    for (w = deadList; w != NULL; w = w->nextPtr) {
	if (w->pid == result) {
	    w->status = status;
	    DEBUG4 ("plp_waitpid: overwriting %d (%s)",
			result, Decode_status (&status));
	    return;
	}
    }
    w = (WaitInfo *) malloc (sizeof(WaitInfo));
    w->pid = result;
    w->status = status;
    w->nextPtr = deadList;
    DEBUG4 ("plp_waitpid: saving %d (%s)", result, Decode_status (&status));
    deadList = w;
}

/*
 * See if there's a suitable process that has already stopped or
 * exited. If so, remove it from the list of exited processes and
 * return its information.
 *
 * Also, store the location of the start of the list before each
 * checking run; if this changes, more entries have been added,
 * and we should re-check.
 */

static WaitInfo *lastCheckedPtr = NULL;

static pid_t
waitpid_check_list (pid_t pid, plp_status_t *statusPtr, int options) {
    WaitInfo *waitPtr, *prevPtr;
    pid_t result;

    do {
	DEBUG4 ("plp_waitpid: checking savelist for %d", pid);
	lastCheckedPtr = deadList;
	for (waitPtr = deadList, prevPtr = NULL; waitPtr != NULL;
		prevPtr = waitPtr, waitPtr = waitPtr->nextPtr)
	{
	    if ((pid != waitPtr->pid) && (pid != -1)) {
		continue;
	    }
	    if (!(options & WUNTRACED) && (WIFSTOPPED (waitPtr->status))) {
		continue;
	    }
	    result = waitPtr->pid;
	    *statusPtr = *((int *) &waitPtr->status);
	    if (prevPtr == NULL) {
		deadList = waitPtr->nextPtr;
	    } else {
		prevPtr->nextPtr = waitPtr->nextPtr;
	    }
	    free ((char *) waitPtr); 
	    return result;
	}
    } while (lastCheckedPtr != deadList);
    /* we check lastCheckPtr against deadList there in case a SIGCHLD
     * has arrived since our last check and has caused a new child to
     * be inserted into the list from the signal handler.
     */
    return -1;
}

/*
 *----------------------------------------------------------------------
 *
 * waitpid --
 *
 *	This procedure emulates the functionality of the POSIX
 *	waitpid kernel call, using the BSD wait3 kernel call.
 *	Note:  it doesn't emulate absolutely all of the waitpid
 *	functionality, in that it doesn't support pid's of 0
 *	or < -1.
 *
 * Results:
 *	-1 is returned if there is an error in the wait kernel call.
 *	Otherwise the pid of an exited or suspended process is
 *	returned and *statusPtr is set to the status value of the
 *	process.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

/* Check to see if any processes have stopped or exited. */

static pid_t
waitpid_wrapper (pid_t pid, plp_status_t *statusPtr, int options) {
    pid_t result;
    plp_status_t status;

retry_waitpid:

#ifdef HAVE_WAITPID
    DEBUG6 ("plp_waitpid: calling waitpid");
    result = waitpid (pid, &status, options);
#else
#ifdef HAVE_WAIT3
    DEBUG6 ("plp_waitpid: calling wait3");
    result = wait3 (&status, options, 0);
#else
    XXX wait not supported!!;
#endif /* HAVE_WAIT3 */
#endif /* HAVE_WAITPID */

    /* interpret the result for debugging purposes */
    if (result == 0) {
	DEBUG4 ("plp_waitpid: kid is still alive");

    } else if (result == -1 && errno == ECHILD) {
	DEBUG4 ("plp_waitpid: no kids left alive");

    } else if (result < 0) {
	DEBUG4 ("plp_waitpid: returning error");

    } else if ((pid != result) && (pid != -1)) {
	/* this is not the correct process */
	waitpid_add_to_list (status, result);
	goto retry_waitpid;

    } else if (!(options & WUNTRACED) && (WIFSTOPPED(status))) {
	/* this is not the correct process */
	waitpid_add_to_list (status, result);
	goto retry_waitpid;

    } else {
	*statusPtr = *((plp_status_t *) &status);
	DEBUG4 ("plp_waitpid: returning %d (%s)",
		    result, Decode_status (statusPtr));
    }
    return result;
}

#ifndef HAVE_SIGPROCMASK
static jmp_buf env_wait;
#endif

static pid_t waiting_for;
static int waiting_opts;

plp_signal_t
sigchld_handler (int signo) {
    pid_t pid;
    int status;
    int tmp_errno;

    tmp_errno = errno;
/* on SVR4, we need to reestablish the signal handler. */
#ifdef SVR4
#define re_sigchld()	plp_signal(SIGCHLD, sigchld_handler);
#else
#define re_sigchld()	;
#endif

    assert (signo == SIGCHLD);
    DEBUG3 ("sigchld_handler: caught SIGCHLD");

    status = 0;
    while ((pid = waitpid_wrapper (-1, &status, waiting_opts|WNOHANG)) > 0) {
	waitpid_add_to_list (status, pid);
	if ((waiting_for == pid) || (waiting_for == -1)) {
	    if (!(waiting_opts & WUNTRACED) && PLP_WIFSTOPPED(status))
		continue;	/* don't report stopped kids */

	    waiting_for = 0;	/* this is the one we're waiting for */
#ifndef HAVE_SIGPROCMASK
	    re_sigchld();	/* only works on SVR4 */
	    longjmp (env_wait);
#endif
	}
    }
    re_sigchld();		/* only works on SVR4 */
    errno = tmp_errno;
}

#ifdef HAVE_SIGPROCMASK
    sigset_t zeromask, newmask, oldmask;
#endif

static void
block_sigchld (void) {
    int tmp_errno;

    tmp_errno = errno;
#ifdef HAVE_SIGPROCMASK
    sigemptyset (&zeromask);
    sigemptyset (&newmask); sigaddset (&newmask, SIGCHLD);
    if (sigprocmask (SIG_BLOCK, &newmask, &oldmask) < 0)
	logerr_die (XLOG_INFO, "sigprocmask");
#else
    /* argh -- setting SIGCHLD with signal() can be really unportable.
     * We just assume the BSD behaviour, as later systems have
     * sigprocmask implementations (which are much more portable).
     */
    plp_signal (SIGCHLD, SIG_IGN);
#endif
    errno = tmp_errno;
}

static void
unblock_sigchld (void) {
    int tmp_errno;

    tmp_errno = errno;
#ifdef HAVE_SIGPROCMASK
    sigprocmask (SIG_SETMASK, &oldmask, NULL);
#else
    plp_signal (SIGCHLD, sigchld_handler);
#endif
    errno = tmp_errno;
}

static pid_t
block_for_kid (pid_t pid, plp_status_t *statusPtr, int options) {
    DEBUG4 ("blocking for pid %d", pid);
    assert (!(options & WNOHANG));

#ifdef HAVE_SIGPROCMASK
    /* we can now tell sigchld_handler that we're waiting for
     * this pid.
     */
    waiting_for = pid; waiting_opts = options;
    while (waiting_for)
	sigsuspend (&zeromask);
#else

    /* we can't use sigsuspend() -- use setjmp() instead. */
    if (setjmp (env_wait) == 0) {
	waiting_for = pid; waiting_opts = options;
	while (waiting_for)
	    pause ();
    }
#endif

/* @@@ Fix from Dave Atkin */
    /*    unblock_sigchld (); */
    /* end fix */
    DEBUG4 ("block_for_kid: pid %d has changed status, checking list", pid);

    /* once we've got to this point, check_list has the
     * value -- pass over to it.
     */
    
    /* @@@ Fix from Dave Atkin */
    { 
	pid_t result;  /* Remove race conditions */
	result = waitpid_check_list (pid, statusPtr, options);
	unblock_sigchld ();
	return (result);
    }
    /* end fix */
    /*     return (waitpid_check_list (pid, statusPtr, options)); */
}

pid_t
plp_waitpid (pid_t pid, plp_status_t *statusPtr, int options) {
    pid_t result;
    int no_kids_left;

    assert (catching_SIGCHLD == True);	/* sanity check */
    no_kids_left = False;
    do {
	block_sigchld ();	/* race condition otherwise */

	result = waitpid_check_list (pid, statusPtr, options);
	if (result != -1) {
	    /* already got that kid */
	    unblock_sigchld (); return result;
	}
	
	/* check for "no kids" error -- use WNOHANG */
	result = waitpid_wrapper (pid, statusPtr, options | WNOHANG);

	/* race condition ends after waitpid_wrapper -- we can unblock.
	 * If the child has changed status while we were checking the list and
	 * running waitpid_wrapper, then SIGCHLD will be delivered
	 * immediately, the deadList will be updated, and we'll go
	 * through this loop again.
	 */
	unblock_sigchld ();

	if (result > 0) {
	    return result;
	}
	if (result == -1 && errno == ECHILD) {
	    no_kids_left = True;
	}

    } while (lastCheckedPtr != deadList);

    /* another race starts here. We block _before_ the two simple
     * tests below, just to keep on the safe side.
     */
    block_sigchld ();

    if (options & WNOHANG) {
	unblock_sigchld (); return 0;
    }
    if (no_kids_left) {
	errno = ECHILD;
	unblock_sigchld (); return -1;
    }

    /* second race ends inside this function. */
    return (block_for_kid (pid, statusPtr, options));
}

void
Setup_waitpid (void) {
    /* the portable waitpid() emulation function needs this SIGCHLD handler. */
    (void) plp_signal(SIGCHLD, sigchld_handler);

#ifndef NDEBUG
    catching_SIGCHLD = True;
#endif
}
