/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1987, 1988, Patrick Powell
 * version 3.3.0 Justin Mason July 1994
 ***************************************************************************
 * MODULE: utils.c
 * Utility routines used by several programs
 ***************************************************************************/

/* to get the real name */
#include <pwd.h>

#include "lp.h"
#include "library/errormsg.h"
#include "library/utils.h"

static int compar (const void *v1, const void *v2);

#define REMOVE_MESSAGE \
"Message from PLP printer spooling system.\n\nDear user %s,\n\nuser %s has deleted your print job %d [%s] \n\
from printer or queue '%s'."

#define SELF_REMOVE_MESSAGE \
"Message from PLP printer spooling system.\n\nDear user %s,\n\nyou have deleted your print job %d [%s] \n\
from printer or queue '%s'."

extern char Mail_log_message[];

/*
 * Time_str_long: returns the full ctime() string...
 */
char 
*Time_str_long (void) {
    time_t tvec;
    static char s[32];

    (void) time (&tvec);
    (void) strcpy (s, (char *) ctime (&tvec));
    s[24] = '\0';
    return (s);
}

/*
 * Time_str: return "cleaned up" ctime() string...
 * 
 * Thu Aug 4 12:34:17 BST 1994 -> Thu Aug  4 12:34:17
 */
char 
*Time_str (void) {
    time_t tvec;
    static char s[32];

    (void) time (&tvec);
    (void) strcpy (s, (char *) ctime (&tvec));
    s[19] = '\0';
    return (&(s[4]));
}

/*
 * Time_str_hist: return "cleaned up" ctime() string...for displaying the time in history functions
 * 
 * Thu Aug 4 12:34:17 BST 1994 -> Thu Aug  4 12:34:17
 */
char 
*Time_str_hist (time_t tvec) {
    static char s[32];

    (void) strcpy (s, (char *) ctime (&tvec));
    s[19] = '\0';
    return (&(s[4]));
}

/*
 * Time_str_cde: return "cleaned up" ctime() string...for displaying 
 * the time in displayq.c for cde dtprintinfo
 * 
 * Thu Aug 4 12:34:17 BST 1994 -> 04/08/97 12:34:17
 */
char 
*Time_str_cde (time_t tvec) {
    static char cde_s[20];
    struct tm *zeit;
    
    zeit = localtime (&tvec);
    (void) sprintf (cde_s, "%02d/%02d/%02d %02d:%02d:%02d", 
		    zeit->tm_mon + 1,
		    zeit->tm_mday,
		    zeit->tm_year,
		    zeit->tm_hour,
		    zeit->tm_min,
		    zeit->tm_sec);
    cde_s[18] = '\0';
    return (&(cde_s[0]));
}

/*
 * Open a file as daemon.
 */
inline int 
open_daemon (char *name, int flag, int perms) {
    int fd;
    user_to_daemon ();
    fd = open (name, flag, perms);
    daemon_to_user ();
    return (fd);
}

/*
 * fopen a file; the file has been created, and we want to do this as daemon for
 * permissions reasons.
 */
inline FILE 
*fopen_daemon (char *name, char *how) {
    FILE *fp;
    user_to_daemon ();
    fp = fopen (name, how);
    daemon_to_user ();
    return (fp);
}

/*
 * UNLINK a file as daemon
 */
inline int 
unlink_daemon (char *name) {
    int f;
    user_to_daemon ();
    f = unlink (name);
    daemon_to_user ();
    return (f);
}


/*
 * CHMOD a file as daemon
 */
inline int 
chmod_daemon (char *name, int perms) {
    int f;
    user_to_daemon ();
    f = chmod (name, perms);
    daemon_to_user ();
    return (f);
}

/*
 * Make a dir as daemon, owned by daemon.
 */
inline int 
mkdir_daemon (char *name, int mode) {
    int f;

    user_to_daemon ();
    f = mkdir (name, mode);
    daemon_to_user ();
    return (f);
}

/***************************************************************************/

void 
fix_SD (void) {
    struct stat sdstat;
    int f, last, retries;

    if (!SD || !*SD || strsame (SD, "/dev/null"))
	return;

    last = 0;
    for (retries = 0; retries < 3; retries++) {
	f = stat (SD, &sdstat);

	if (f < 0) {
	    if (last == errno) {/* didn't fix the last error */
		logerr_die (XLOG_NOTICE, "fix_SD: can't fix \"%s\"!!", SD);

	    } else if (errno == ENOENT) {	/* dir doesn't exist */
		logerr (XLOG_NOTICE, "creating the spool dir \"%s\"", SD);
		if (mkdir_daemon (SD, 0777)) {
		    logerr_die (XLOG_NOTICE, "fix_SD: couldn't mkdir \"%s\"", SD);
		}
		(void) chmod_daemon (SD, 042755);

	    } else {
		logerr_die (XLOG_NOTICE, "fix_SD: can't fix \"%s\"!!", SD);
	    }
	    last = errno;

	} else {		/* it exists, anyway */
	    if ((sdstat.st_uid != DaemonUID) || (sdstat.st_gid != DaemonGID)) {
		log (XLOG_NOTICE, "owner/group of \"%s\" is incorrect: %d/%d != %d/%d",
			SD, sdstat.st_uid,sdstat.st_gid,DaemonUID,DaemonGID);
		(void) chown (SD, DaemonUID, DaemonGID);	/* try anyway */

	    } else if (!sdstat.st_mode & S_IFDIR) {
		log (XLOG_NOTICE, "\"%s\": not a directory, fixing", SD);
		if (!unlink (SD)) {
		    logerr_die (XLOG_NOTICE, "fix_SD: couldn't unlink \"%s\"", SD);
		}
		if (mkdir (SD, 0777)) {
		    logerr_die (XLOG_NOTICE, "fix_SD: couldn't mkdir \"%s\"", SD);
		}
		(void) chmod_daemon (SD, 042755);

	    } else if ((sdstat.st_mode & 0777) != 0755) {
		log (XLOG_NOTICE, "fixing permissions on \"%s\": (was mode %o)",
		     SD, sdstat.st_mode & 0777);
		(void) chmod_daemon (SD, 042755);

	    } else {
		break;				/* wahey! it's fixed */
	    }
	}
    }
    if (retries >= 3) {
	fatal (XLOG_NOTICE, "couldn't fix \"%s\".", SD);
    }
}

inline void 
chdir_SD (void) {
    if (chdir (SD) == 0) {
	return;			/* c'est bon */
    }
    fix_SD ();
    if (chdir (SD) < 0) {
	logerr_die (XLOG_NOTICE, "fix_SD: can't fix \"%s\"!!", SD);
    }
}

inline DIR 
*opendir_SD (void) {
    DIR *dir;

    if ((dir = opendir (SD)) != NULL) {
	return (dir);		/* c'est bon */
    }
    fix_SD ();
    if ((dir = opendir (SD)) == NULL) {
	logerr_die (XLOG_NOTICE, "fix_SD: can't fix \"%s\"!!", SD);
    }
    return (dir);
}

/***************************************************************************
 * Remove_job (FILE *fp; struct plp_queue *q)
 * Remove a job and all associated files; if success == JSUCC, remove
 * U links as well
 * ACTIONS:
 * 1. Read the control file
 * 2. get data file names and unlink them if they match the job
 * 3. Unlink the control file
 * 4. Send a mail to the user
 ***************************************************************************/

void
Remove_job (FILE *fp, struct plp_queue *q) {
    char parm[BUFSIZ];		/* line buffer */
    char *arg;			/* pointer to the argument */
    int c, i = 0;		/* ACME Integers, Inc. */
    char username[64];          /* holds the username */
    struct passwd *pwent;       /* to get the real username */

    DEBUG5 ("Remove_job: removing job %s", q->q_name);
    clearerr (fp);
    /* get all the values in case printer is remote */
    if (RM && RP) {
	(void) Get_pc_entry (Printer, All_pc_vars, All_pc_len);
    }
    if (fseek (fp, 0L, 0) < 0) {
	logerr_die (XLOG_INFO,
		    "Remove_job: fseek failed (%s)", q->q_name);
    } else {
	/*
	 * read through the file, looking for things to remove
	 */
	while (fgets (parm, sizeof (parm), fp)) {
	    if ((arg = (char *) strchr (parm, '\n')) == 0) {
		log (XLOG_INFO, "%s: bad control file, no endline",
		     q->q_name);
		continue;
	    }
	    *arg = '\0';
	    arg = parm + 1;
	    c = parm[0];
	    /*
	     * is it a format line?
	     */
	    if (!isascii (c) || !islower (c)) {
		continue;
	    }
	    if (Job_match (q->q_name, arg) == 0) {
		DEBUG4 ("Remove_job: bad file name '%s'", arg);
		continue;
	    }
	    (void) unlink_daemon (arg);
	    /* if there are some old tf...-files laying around, delete them */
	    if (IF && *IF) {
		arg[0] = 't';
		(void) unlink_daemon (arg);
	    }
	}
    }
    (void) unlink_daemon (q->q_name);
    (void) unlink_daemon ("progress");

    /* only send a message if the job is removed by user request */
    DEBUG3 ("Remove_job: Value of called_from_lprm = %d, mail_to_user_for_remote_remove = %d.", called_from_lprm, mail_to_user_for_remote_remove);
    if ((called_from_lprm) || (mail_to_user_for_remote_remove)) {
        DEBUG3 ("Remove_job: sending mail to user cause of removal of job(s)");

	/* clearing the array "username" */
	for (i = 0; i < 64; i++) {
	  username[i] = '\0';
	}
	(void) strncpy (username, q->q_user, strlen (q->q_user));
	pwent = getpwnam (username);
	if (pwent != NULL) {
	    (void) sprintf (username, pwent->pw_gecos);
	}

        if (strcmp (Person, q->q_user) == 0) {
	    /* the user has deleted his own job */
	    sprintf (Mail_log_message, SELF_REMOVE_MESSAGE, 
		     username, q->q_num, q->q_data->name, Printer);
	} else { /* one of the administrators has deleted this job */
	    sprintf (Mail_log_message, REMOVE_MESSAGE, 
		     username, Person, q->q_num, q->q_data->name, Printer);
	}
	/* send a mail to the user (status = 1, successful) */
	sendmail (q, 1);
	called_from_lprm = 0;
	mail_to_user_for_remote_remove = 0;
    }
}

/***************************************************************************
 * Compare queue entries by 
 * 1. Priority 
 * 2. size, if SZ or jobs_sorted_on_size is set
 * 3. Time 
 * 4. millisecond timing (if there)
 ****************************************************************************/

static int 
compar (const void *v1, const void *v2) {
    struct plp_queue *c1, *c2;
    int x;

    c1 = (struct plp_queue *) v1;
    c2 = (struct plp_queue *) v2;

    /* priority goes first */
    if ((x = ((c1)->q_priority - (c2)->q_priority)))
	return (x);

    /* if the SZ flag or the variable jobs_sorted_on_size is set, the jobs 
     * will be sorted according to their size (smallest will be printed first)
     */
    if ((SZ) || (jobs_sorted_on_size)) {
	if ((c1)->q_data->size < (c2)->q_data->size) {
	    return (x = -1);
	} else if ((c1)->q_data->size > (c2)->q_data->size) {
	    return (x = 1);
	}
    }

    /* we can't simply do the same for these two -- their values may be
     * unsigned longs, so we may get the order wrong for that reason!
     */

    if ((c1)->q_time < (c2)->q_time) {			/* time spooled by lpr */
	return (x = -1);
    } else if ((c1)->q_time > (c2)->q_time) {
	return (x = 1);
    }

    if ((c1)->q_time_hires < (c2)->q_time_hires) {	/* millisecs of q_time */
	return (x = -1);
    } else if ((c1)->q_time_hires > (c2)->q_time_hires) {
	return (x = 1);
    }

    return False;
}

/***************************************************************************
 * int Getq();
 * Scan the spool directory and make a list of server files sorted by
 * priority, creation time, etc.
 * Returns: the number of entries
 * Side Effects: sets struct plp_queue *Queue to point to an array of pointers
 *   to queue entries
 ***************************************************************************/

#include "LPD/banner.h"

unsigned int
Getq (void) {
    static int arraysz = 0;	/* number of entries in array */
    plp_dir_t *d;		/* directory entry */
    unsigned int nitems;	/* ACME Integers, Inc. */
    int idx;			/* index of current data filename  */
    struct stat stbuf;		/* for statting files */
    DIR *dirp;			/* used by directory routines */
    char l[BUFSIZ];		/* for reading the control file */
    char fullname[BUFSIZ];	/* for the full pathname */
    char *filepart;		/* end of the pathname */
    FILE *cfp;			/* control file FP */
    struct plp_queue *q;	/* pointer to queue entry */
    char j_files[FILES_LENGTH + 1];		/* these are EUCS things */
    char j_gecos[USER_INFO_LENGTH + 1];
    char j_options[OPTIONS_LENGTH + 1];
    char j_delivery[DELIVERY_LENGTH + 1];
    int j_fields;

    if (SD == 0 || *SD == 0) {
	fatal (XLOG_INFO, "spool directory is not set!");
    }
    if ((dirp = opendir_SD ()) == NULL) {
	logerr_die (XLOG_INFO, "cannot open spool directory %s", SD);
    }
    /*
     * get the directory name and copy to buffer
     */
    if (strlen (SD) + sizeof (q->q_name) + 2 > sizeof (fullname)) {
	logerr_die (XLOG_NOTICE, "Getq: file name too big");
    }
    (void) strcpy (fullname, SD);
    filepart = &fullname[strlen (fullname)];
    *filepart++ = '/';
    /*
     * assume you have 100 entries to start with
     */
    if (arraysz == 0) {
	arraysz = 100;
	Queue = (struct plp_queue *)
		malloc ((unsigned) (arraysz * sizeof (struct plp_queue)));
	if (Queue == NULL) {
	    logerr_die (XLOG_INFO, Malloc_failed_msg);
	}
    } else {
	for (q = Queue, nitems = 0; nitems < arraysz; nitems++, q++)
	    if (q->q_data)
		(void) free ((void *) q->q_data);
    }

    (void) memset ((void *) Queue, 0,
		   (unsigned) (arraysz * sizeof (struct plp_queue)));

    /*
     * set the item count to 0 and scan the directory
     */
    nitems = 0;
    user_to_daemon ();
    while ((d = readdir (dirp)) != NULL) {

	if (!(((d->d_name[0] == 'h') ||
	       (d->d_name[0] == 'c')) && (d->d_name[1] == 'f')))
	{
	    continue;		/* server control files only */
	}
	/*
	 * Check to make sure the array has space left and realloc the maximum size.
	 */
	if (nitems >= arraysz) {
	    /* try to get another buffer */
	    Queue = (struct plp_queue *) realloc ((char *) Queue,
		    (unsigned) ((arraysz * 2) * sizeof (struct plp_queue)));
	    if (Queue == NULL) {
		logerr_die (XLOG_NOTICE, Malloc_failed_msg);
	    }
	    (void) memset ((void *) (Queue + arraysz), 0,
			   arraysz * sizeof (struct plp_queue));
	    arraysz *= 2;
	}
	/*
	 * get next queue entry and clear it
	 */
	q = &Queue[nitems];

	*j_options = '\0';
	*j_gecos = '\0';
	*j_files = '\0';
	*j_delivery = '\0';

	/* copy the name of the file */
	if (strlen (d->d_name) >= CFNAMELEN) {
	    logerr_die (XLOG_NOTICE, "control file name %s too long", d->d_name);
	}

	(void) strcpy (q->q_name, d->d_name);
	(void) strcpy (filepart, d->d_name);

	if (d->d_name[0] == 'h') 
	    q->q_held = True;
	else
	    q->q_held = False;

	/*
	 * open the control file and read the information
	 * (we are daemon throughout this loop!)
	 */
	if ((cfp = fopen (fullname, "r")) == NULL) {
	    DEBUG4 ("%s cannot be opened - skipping", fullname);
	    continue;		/* it disappeared - skip it */
	}

	/* set a priority */
	q->q_priority = d->d_name[2];
	/* get the job number */
	q->q_num = atoi (&d->d_name[3]);

	while (fgets (l, sizeof (l), cfp)) {
	    /* @@@ Fix from Dave Atkin */
	    if (islower(l[0]))  /* file to print */
		/* end fix */
	    /* if (l[0] == 'N')	*/ 
		q->q_entries++;
	}

	DEBUG5 ("Getq: found %d file names", q->q_entries);

	q->q_data = (struct q_file *)
	    malloc((unsigned) ((q->q_entries + 1) * sizeof(struct q_file)));
	(void) memset((void *) q->q_data, 0,
		      (unsigned) (q->q_entries + 1) * sizeof(struct q_file));

	(void) rewind(cfp);

	idx = 0;

	while (fgets (l, sizeof (l), cfp)) {
	    l[strlen (l) - 1] = 0;		/* clobber the \n */

	    if (islower (l[0])) {
		if (idx >= 0)
		    q->q_data[idx].copies++;
		if (idx < 0 || q->q_data[idx].size == 0) {
		    /* get size of file to be printed */
		    (void) strcpy (filepart, &l[1]);
		    if (stat (fullname, &stbuf) < 0) {
			(void) fclose (cfp);
			DEBUG5 ("Getq: data file %s does not exist", l);
		    } else {
			q->q_size += stbuf.st_size;
			if (idx >= 0)
			    q->q_data[idx].size = stbuf.st_size;
		    }
		} else {
		    q->q_size += q->q_data[idx].size;
		}
	    } else {
		switch (l[0]) {
		case 'N':	/* name of file */
		    (void) strncpy(q->q_data[idx].name, l + 1,
				   sizeof(q->q_data[idx].name));
		    if (idx++ >= q->q_entries) {
			idx--;
			log(XLOG_INFO, "Getq: controlfile %s has changed!",
			     q->q_name);
		    }
		    break;

		case 'P':	/* name of user */
		    (void) strncpy (q->q_user, &l[1], sizeof (q->q_user) - 1);
		    q->q_user[sizeof (q->q_user) - 1] = '\0';
		    break;

		case 'D':	/* when spooled (non-RFC) */
		    {
			char *dot;
			if ((dot = strchr (&l[1], '.')) != NULL) {
			    *dot = '\0'; dot++;
			    q->q_time_hires = (unsigned long) atol (dot);
			} else {
			    q->q_time_hires = 0;
			}
			q->q_time = (time_t) atol (&l[1]);
		    }
		    break;

		case 'J':	/* job name with extra info */
		    if (EU) {
			j_fields = interpret_job (l + 1, j_files, j_gecos,
						  NULL, j_options, NULL, NULL);
			if (*j_options != '\0' &&
			    (!RM || *RM == '\0') &&
			    strcmp (j_options, DEF_OPTIONS) != 0)
			{
			    q->q_options = strdup (j_options);
			} else {
			    q->q_options = NULL;
			}
		    }
		    break;

		default:
		    break;
		}
	    }
	}
	DEBUG4 ("Getq: '%s' %d,%s,%c", q->q_name,
		q->q_num, q->q_user, q->q_priority);

	(void) fclose (cfp);

	/* BSD compatibility: if spool-time isn't set, get it from the
	 * spoolfile's modification date.
	 */
        if (q->q_time == 0) {
	    if (stat (q->q_name, &stbuf) < 0) {
		q->q_time = time ((time_t *) 0);
	    } else {
		q->q_time = stbuf.st_mtime;
	    }
	}
	++nitems;
    }
    (void) closedir (dirp);
    daemon_to_user ();

    *((Queue[nitems]).q_name) = '\0';

    /*
     * sort the queue
     */
    if (nitems) {
	qsort ((char *) Queue, nitems, sizeof (struct plp_queue), compar);
    }
    return (nitems);
}

/***************************************************************************
 * Match_entry (struct plp_queue *q)
 * check the user name and job number against the parmlist starting at start;
 * Return: 1 if match found, 0 otherwise
 ***************************************************************************/
int 
Match_entry (struct plp_queue *q) {
    int i;			/* ACME Integer, Inc. */
    char *k;

    k = &q->q_name[6];          /* get the hostname from the controlfilename */
    DEBUG6 ("Match_entry: Match_entry on %s, %d", q->q_user, q->q_num);
    for (i = 0; i < Parmcount; ++i) {
	if ((Parms[i].num >= 0 && Parms[i].num == q->q_num) /* jobnumber  */
	    || strsame (Parms[i].str, q->q_user)            /* user name  */
	    || (strcmp (Parms[i].str, k) == 0)              /* hostname   */
	    || (strcmp (Parms[i].str, "*") == 0) )	    /* everything */
	{
	    DEBUG4 ("Match_entry: Match_entry on %s, %d succeeded", q->q_user, q->q_num);
	    return True;
	}
    }
    return False;
}

/***************************************************************************
 * bpread (int fd ; char *buf; int size)
 * Almost bombproof way to read command line from socket
 * Returns: -1 if unable to read a line terminated with \n
 *     0 if 0 length line or first character is null (i.e.- no input)
 *     otherwise: length of line
 ***************************************************************************/
int 
bpread (int fd, char *buf, int size) {
    int n;			/* ACME Integer, Inc. */
    int i;

    /*
     * Bombproof line reading
     */
    for (i = 0; i < size; ++i) {
	n = read (fd, &buf[i], 1);
	if (n == 0) {
	    if (i == 0) {
		DEBUG3 ("bpread: EOF");
	    } else {
		DEBUG3 ("bpread: bad format '%d' %s len = %d",
			 buf[0], buf + 1, i);
	    }
	    return False;
	}

	if (n < 0) {
	    if (errno == ECONNRESET) {
		/* sometimes get this when one end exits before the
		 * other gets to notice the connection closing.
		 */
		DEBUG3 ("bpread: EOF (connection reset)");
		return False;
	    }
	    logerr (XLOG_INFO, "bpread: lost connection");
	    return (-1);
	}
	if (buf[i] == '\n') {
	    buf[i] = '\0';
	    DEBUG3 ("bpread: '%d' %s", buf[0], &buf[1]);
	    return (i);
	} else if (buf[i] == 0 && i == 0) {
	    DEBUG3 ("bpread: zero byte first");
	    return (i);
	}
    }
    log (XLOG_INFO, "bpread: bad line %s", buf);
    return (-1);
}

/***************************************************************************
 * estrcp(char *s, *t, *e)
 * copies t to s, and returns the end position
 * Returns: end of string if the result is shorter than e, NULL otherwise
 * Note: this is a very useful routine.  Think about adding it to string
 * functions.
 ***************************************************************************/
inline char 
*estrcp (char *s, char *t, char *e) {
    if (s) {
	while ((*s++ = *t++) != '\0') {
	    if (s >= e)
		break;
	}
	if (s < e) {
	    return (s - 1);
	} else {
	    return ((char *) NULL);
	}
    }
    return (s);
}

/* some Parms dynamic-allocation routines:
 *
 * allocParmsIfNeeded  -- allocate the Parms array, if it isn't already;
 *                        also, shrink the Parms array if it's too large
 *                        (garbage collection)
 *
 * growParmsIfNeeded   -- increase size of the Parms array if necessary
 */

#define PARMS_BUFSIZ	20		/* arbitrary */

void 
allocParmsIfNeeded (void) {
    if (ParmSz == 0) {
	ParmSz = PARMS_BUFSIZ;
	DEBUG6 ("allocating Parms to %d", ParmSz);
	Parms = (struct parm *)
		malloc ((unsigned) (ParmSz * sizeof (struct parm)));

    } else if (Parmcount + PARMS_BUFSIZ < ParmSz) {
	/* shrink the array to slightly larger than it needs to be */
	ParmSz = Parmcount + PARMS_BUFSIZ;
	DEBUG6 ("shrinking Parms to %d", ParmSz);
	Parms = (struct parm *) realloc ((char *) Parms,
				(unsigned) (ParmSz * sizeof (struct parm)));
    }
    if (Parms == NULL) {
	logerr_die (XLOG_NOTICE, Malloc_failed_msg);
    }
    return;
}

void 
growParmsIfNeeded (void) {
    if (Parmcount >= ParmSz) {
	/* try to get another buffer */
	ParmSz = ParmSz + PARMS_BUFSIZ;

        DEBUG6 ("growing Parms to %d", ParmSz);
	Parms = (struct parm *) realloc ((char *) Parms,
				(unsigned) (ParmSz * sizeof (struct parm)));
	if (Parms == NULL) {
	    logerr_die (XLOG_NOTICE, Malloc_failed_msg);
	}
    }
    return;
}

/***************************************************************************
 * splitline (char *line)
 *    splits a line into tokens, and places them in the Parms[] array.
 *    if firstarg is present then it is set to the first argument
 * Side Effects: modifies Parms, Parmcount
 ***************************************************************************/

void 
splitline (char *bp) {
    allocParmsIfNeeded ();
    /* find the start */
    while (*bp) {
	while (*bp && isspace (*bp))
	    bp++;
	if (*bp) {
	    growParmsIfNeeded ();
	    Parms[Parmcount].str = bp;
	    if (isdigit (*bp)) {
		Parms[Parmcount].num = atoi (bp);
	    } else {
		Parms[Parmcount].num = -1;
	    }
	    Parmcount++;
	}
	while (*bp && !isspace (*bp))
	    bp++;
	if (*bp)
	    *bp++ = '\0';
    }
    return;
}

/**********************************************************************
 * Shift_parms (int n)
 * left shift the entries in the Parms[] array by n
 * Side Effects: modifies Parms and Parmcount
 **********************************************************************/

void
Shift_parms (int n) {
    int i;			/* ACME Chain and Integer, Inc. */
    while (Parmcount > 0 && n > 0) {
	for (i = 1; i < Parmcount; ++i) {
	    Parms[i - 1] = Parms[i];
	}
	--Parmcount;
	--n;
    }
    allocParmsIfNeeded ();
    return;
}

/**********************************************************************
 * char possible_options ()
 * returns the name of the Forms_file
 **********************************************************************/
char 
*possible_options (void) {
    if ((!Forms_file) || (!FL || *FL == '\0')) {
	return NULL;
    }
    return (strdup (FL));
}

/**********************************************************************
 * char current_options ()
 * checks the available options in the Forms_file  
 **********************************************************************/
char 
*current_options (void) {
    FILE *opt_fp;
    char *ip, *ret_str;
    char opt_str[80];
    char opt_file_name[MAXPATHLEN];

    if (!Forms_file) {
	return DEF_OPTIONS;
    }
    if (!FL || *FL == '\0')
	return NULL;

    (void) strcpy (opt_file_name, Forms_file);
    if ((ip = strchr (Printer, '.')))
	strcat (opt_file_name, ip);

    if ((opt_fp = fopen_daemon (opt_file_name, "r")) == NULL) {
	log (XLOG_INFO, "Can't open options file for %s", Printer);
	return DEF_OPTIONS;
    } else {
	if (fscanf (opt_fp, "%s", opt_str) != 1) {
	    log (XLOG_INFO, "Couldn't get line from options file for %s", Printer
		    );
	    fclose (opt_fp);
	    return DEF_OPTIONS;
	}
	fclose (opt_fp);
	if (strsame ("-", opt_str)) {
	    strcpy (opt_str, "none");
	}
	ret_str = strdup (opt_str);

	return (ret_str);
    }
}

/**********************************************************************
 * int valid_forms (char *, char *)
 * forms option to try
 * comma delimited list of possible forms
 **********************************************************************/
int 
valid_forms (char *forms, char *choices) {
    char buffer[256];
    char *ptr = buffer;
    char *this;

    if (!Forms_file) {
	return True;
    }
    strcpy (buffer, choices);

    while ((this = strtok (ptr, ",")) != NULL) {
	if (strsame (this, forms)) {
	    return True;
	}
	ptr = NULL;
    }
    return False;
}

/*
 * current_limit() returns the current limit on the size of print jobs (in 1k blocks).
 * If there is no limit then it returns -1. Note that this is different from the mx value
 * in the printcap entry which limits the size of jobs which can be queued.
 * 
 * We need to be able to have different limits on printers attached to one queue.  This
 * allows us to have a generic queue but still give good service for small jobs.  In
 * order to do this we distinguish between limits files by using the "li" printcap
 * parameter, which can either accept a numeric value (the actual limit for that queue)
 * or a filename (a file to be opened to get the limit).
 */

long 
current_limit (void) {
    static long limit = -1L;
    static struct stat last_buf, buf;
    static char cached_Printer[PRNAMELEN] = "";
    static char cached_filename[MAXPATHLEN] = "";

    if (!LI || !*LI) {
	/* no limit on this printer */
	return (-1L);
    }

    if (*LI >= '0' && *LI <= '9') {
	/* limits value is numeric */
	return (atol (LI));
    }

    /* otherwise, limits value is a filename to open.
     * We cache the last limits-file lookup to save
     * repeated opens and closes, and will get a new
     * value if the Printer is different, the filename
     * changes or the modtime of the file changes.
     */
    if ((!strsame (Printer, cached_Printer)) || (!strsame (LI, cached_filename)))
    {
	/* it's changed; force a reload */
	last_buf.st_mtime = 0;
	(void) strcpy (cached_Printer, Printer);
	(void) strcpy (cached_filename, LI);
    }

    if (stat (LI, &buf) != 0) {
	/* the limits file isn't there */
	limit = -1L;
	
	return (-1L);
    }

    if (buf.st_mtime > last_buf.st_mtime) {
	/* file has changed - need to read it */
	int ret;
	FILE *fp;

	if ((fp = fopen (LI, "r")) == NULL) {
	    /* the file exists (the stat worked), but we can't read it */
	    log (XLOG_INFO, "current_limit: can't read limits file");
	    return (-1L);
	}
	ret = fscanf (fp, "%ld", &limit);
	(void) fclose (fp);
	if (ret != 1) {
	    log (XLOG_INFO,
		 "current_limit: can't read value from limits file");
	    return (-1L);
	}
    }

    /* else the limit value from the last lookup is still cached;
     * return it.
     */
    return (limit);
}

/***********************************************************************
 * pr_stat_file_lpstat (statusfile, flag)
 * prints the status of the printer according to the information 
 * stored in the status file (PS or standard). If flag is 1 the
 * availability will also be displayed
 **********************************************************************/
static void 
pr_stat_file_lpstat (char *file, int lpstat_p_option) {
    int i;
    FILE *fp;
    char buf[BUFSIZ];

    if ((fp = fopen_daemon (file, "r")) != NULL) {
	buf[0] = '\0';
	while (fgets (buf, sizeof (buf), fp) != NULL) {
	    /* delete the "." at the end of the string */
	    for (i = 0; i < strlen (buf); i++) {
		if (buf[i] == '.') {
		    buf[i] = '\0';
		}
		if  (buf[i] =='\n') {
		    buf[i] = '\0';
		}
	    }
	    if (strstr (buf, "since") != NULL) {
		(void) printf ("%s", strstr (buf, "since"));
	    } else {
		char t_buf[26];
		time_t zeit;
		zeit = time ((time_t) 0);
		(void) strcpy (t_buf, ctime (&zeit));
		t_buf[24] = '\0';
		(void) printf ("since %s", t_buf);
	    }
	    /* get the actual information from the queue; we use 
	     * the "Checkactive" function here
	     */
	    (void) Checkactive ();
	    if (lpstat_p_option) {
		if (LO_statb.st_mode & (DISABLE_QUEUE))  {
		    (void) printf (". not available.\n");
		} else {
		    (void) printf (". available.\n");
		}
	    } else {
		if ( (LO_statb.st_mode & (DISABLE_PRINT)) && 
		     (LO_statb.st_mode & (DISABLE_QUEUE)) ) {
		    (void) printf (" - \n\t%s\n", buf);
		} else {
		    (void) printf ("\n");
		}
	    }
	    (void) fflush (stdout);
	}
    }
    (void) fflush (stdout);
    (void) fclose (fp);
    return;
}

static void 
pr_stat_file (char *file) {
    FILE *fp;
    char buf[BUFSIZ];

    if ((fp = fopen_daemon (file, "r")) != NULL) {
	buf[0] = '\0';
	while (fgets (buf, sizeof (buf), fp) != NULL) {
	    (void) printf ("  %s", buf);
	}
	(void) fflush (stdout);
	(void) fclose (fp);
    }

    /* have forms info already been printed ? */ 
    if (forms_info_printed)
	return;

    /* Only print forms and limit information for local printers */
    if ((RM == NULL) || (*RM == '\0')) {
	long limit;
	int output = False;

	if (FL && *FL) {
	    (void) printf ("  Current forms = %s (possible: %s)",
		     strsame (current_options (), "none") ? "none" : current_options (), 
			   possible_options ());
	    output = True;
	}
	if ((limit = current_limit ()) != -1L) {
	    (void) printf ("  Current size limit = %ldk", (long) limit);
	    output = True;
	}
	if (output) {
	    forms_info_printed = 1;
	    (void) printf ("\n");
	}
    }
    return;
}

inline void 
printstatus (int lpstat_p_option) {
    if (ST && *ST) {
	if (sysv_lp) {
	    pr_stat_file_lpstat (ST, lpstat_p_option);
	} else {
	    pr_stat_file (ST);
	    forms_info_printed = 1;
	}
    }
    if (PS && *PS) {
	if (sysv_lp) {
	    pr_stat_file_lpstat (PS, lpstat_p_option);
	} else {
	    pr_stat_file (PS);
	    forms_info_printed = 0;
	}
    }
    return;
}

/***************************************************************************
 * const char *Sigstr(n)
 * Return a printable form the the signal
 ***************************************************************************/

#if !defined(HAVE_SYS_SIGLIST) && !defined(HAVE__SYS_SIGLIST)

static struct signame {
    char *str;
    int value;
} signals[] = {

    {
	"SIGHUP", SIGHUP
    }, {
	"SIGINT", SIGINT
    },
    {
	"SIGQUIT", SIGQUIT
    }, {
	"SIGILL", SIGILL
    },
    {
	"SIGTRAP", SIGTRAP
    }, {
	"SIGFPE", SIGFPE
    },
    {
	"SIGKILL", SIGKILL
    }, {
	"SIGSEGV", SIGSEGV
    },
    {
	"SIGPIPE", SIGPIPE
    }, {
	"SIGALRM", SIGALRM
    },
    {
	"SIGTERM", SIGTERM
    }, {
	"SIGSTOP", SIGSTOP
    },
    {
	"SIGTSTP", SIGTSTP
    }, {
	"SIGCONT", SIGCONT
    },
    {
	"SIGCHLD", SIGCHLD
    },
#ifdef SIGIOT
    { "SIGIOT", SIGIOT },
#endif
#ifdef SIGEMT
    { "SIGEMT", SIGEMT },
#endif
#ifdef SIGBUS
    { "SIGBUS", SIGBUS },
#endif
#ifdef SIGSYS
    { "SIGSYS", SIGSYS },
#endif
#ifdef SIGIO
    { "SIGIO", SIGIO },
#endif
#ifdef SIGXCPU
    { "SIGXCPU", SIGXCPU },
#endif
#ifdef SIGXFSZ
    { "SIGXFSZ", SIGXFSZ },
#endif
    { "SIGPROF", SIGPROF }
    /* that's all */
};

static int nsignals = sizeof (signals) / sizeof (struct signame);

const char 
*Sigstr (int n) {
    int i;
    static char buf[40];

    for (i = 0; i < nsignals; ++i) {
	if (signals[i].value == n) {
	    return (signals[i].str);
	}
    }
    (void) sprintf (buf, "signal %d", n);
    return (buf);
}

#else /* HAVE_SYS_SIGLIST || HAVE__SYS_SIGLIST */
  
#ifdef HAVE__SYS_SIGLIST
# ifdef NEED__SYS_SIGLIST
extern char *_sys_siglist[];
# endif
#else
# undef _sys_siglist
# define _sys_siglist sys_siglist
# ifdef NEED_SYS_SIGLIST
extern char *sys_siglist[];
# endif
#endif

const char 
*Sigstr (int n) {
    static char buf[40];
#ifndef HAVE__SYS_NSIG
#ifdef NSIG
    int _sys_nsig = NSIG;
#else
    int _sys_nsig = 32;               /* we still get the interesting ones */
#endif
#endif /* HAVE__SYS_NSIG */

    if (n > _sys_nsig) {
        (void) sprintf (buf, "signal %d", n);
        return (buf);
    } else {
	return (_sys_siglist[n]);
    }
}

#endif /* HAVE_SYS_SIGLIST */

/***************************************************************************
 * Decode_status (plp_status_t *status)
 * returns a printable string encoding return status
 ***************************************************************************/

char 
*Decode_status (plp_status_t *status) {
    static char msg[BUFSIZ];

    *msg = '\0';		/* just in case! */
    if (PLP_WIFEXITED (*status)) {
	(void) sprintf (msg, "exited with status %d", PLP_WEXITSTATUS (*status));

    } else if (PLP_WIFSTOPPED (*status)) {
	(void) sprintf (msg, "stopped");

    } else {
	(void) sprintf (msg, "died%s",
			PLP_WCOREDUMP (*status) ? " and dumped core" : "");

	if (PLP_WTERMSIG (*status)) {
	    (void) sprintf (msg + strlen (msg), ", %s",
			    Sigstr ((int) PLP_WTERMSIG (*status)));
	}
    }
    return (msg);
}

/***************************************************************************
 * pid_t Checkactive() 
 * find the currently active files in the spool queue. 
 * 1.  check for the single server possibility first. 
 * 2.  find the Names of the servers 
 * 3.  find the active files for each of the servers
 ***************************************************************************/
  
pid_t 
Checkactive (void) {
    int i;			/* ACME Integers, Inc. */
    char buf[BUFSIZ];		/* Name of active file */
    static char server[BUFSIZ];	/* Name of server file */
    char *sp, *ep;		/* ACME Pointer */
    pid_t pid;

    pid = 0;
    buf[0] = '\0';
    if (Checklockfile (LO, &pid, buf, sizeof (buf), &LO_statb) && buf[0]) {
	for (i = 0; i < Jobcount; ++i) {
	    if (strsame (buf, Queue[i].q_name)) {
		Queue[i].q_daemon = pid;
		Queue[i].q_server = Printer;
	    } else {
		Queue[i].q_daemon = 0;
		Queue[i].q_server = NULL;
	    }
	}
    }
    /*
     * check for each of the servers
     */
    if (SV && *SV) {
	for (i = 0; i < Jobcount; ++i) {
	    Queue[i].q_daemon = 0;
	    Queue[i].q_server = NULL;
	}
	(void) strcpy (server, SV);

	for (sp = server; sp; sp = ep) {
	    if ((ep = (char *) strchr (sp, ','))) {
		*ep = '\0';
		++ep;
	    }
	    /*
	     * get the lock file and the status from the server
	     */
	    pid = 0;
	    buf[0] = '\0';
	    if (Checklockfile (sp, &pid, buf, sizeof (buf),
	   	 (struct stat *) NULL) && buf[0])
	    {
		for (i = 0; i < Jobcount; ++i) {
		    if (strsame (buf, Queue[i].q_name)) {
			Queue[i].q_daemon = pid;
			Queue[i].q_server = sp;
			break;
		    }
		}
	    }
	}
    }
    return (pid);
}

/***************************************************************************
 * Print either a file or the comment field itself.
 ***************************************************************************/
void 
Print_comment (const char *str) {
    FILE *f;			/* a file to print */
    int pc, c;

    assert(str != (char*)0);
    if (!strchr (str, ' ')) {
	if ((f = fopen (str, "r")) != NULL) {
	    putc (' ', stdout);
	    if ((pc = getc (f)) != EOF) {
		while ((c = getc (f)) != EOF) {
		    putc (pc, stdout);
		    pc = c;
		}
		if (pc != '\n') {
		    putc (pc, stdout);
		}
	    }
	}
    } else {
	(void) fputs (str, stdout);
    }
    return;
}

/***************************************************************************
 * Print the head for the lpq command
 ***************************************************************************/

void 
Print_queuehead (long mode, const char *attached) {
  if (SA) {
     (void) fputs ("Save area ", stdout);
  } else {
     if (FQDN != Host) {
	(void) fputs ("Remote ", stdout);
     } else {
        (void) fputs ("Local  ", stdout);
     }
     if ((SV == 0 || *SV == 0) && (RM == 0 || *RM == 0)) {
	 (void) fputs ("printer ", stdout);
     } else { 
	 (void) fputs (" queue  ", stdout);
     }
  }
  if (attached) {
     (void) fprintf (stdout, "'%s'", attached);
     (void) printf (" [attached to '%s']", Printer);
  } else {
     (void) fprintf (stdout, "'%s'", Printer);
  }

  (void) printf (" on %s", Host);

  if (CM && *CM) {
     (void) fputs (" (", stdout);
     Print_comment (CM);
     (void) putc (')', stdout);
  }
  (void) fputs (": ", stdout);

  if (mode & (DISABLE_QUEUE | DISABLE_PRINT)) {
     
    
     if (mode & DISABLE_QUEUE) {
      (void) fputs("no spooling", stdout);
     }
    
     if (mode & DISABLE_PRINT) {
        if (mode & DISABLE_QUEUE)
	  (void) fputs(", ", stdout);
      
	(void) fputs("no printing", stdout);
     }
  }
  
  (void) putc ('\n', stdout);
  (void) fflush (stdout);
  return;
}

/***************************************************************************
 * Print the head for the lpq -h command
 ***************************************************************************/

void 
Print_queuehead_hist (void) {
  (void) fprintf (stdout, "History of jobs for ");
  if (SA) {
     (void) fputs ("Save area ", stdout);
  } else {
     if (FQDN != Host) {
	(void) fputs ("remote ", stdout);
     } else {
        (void) fputs ("local  ", stdout);
     }
     if ((SV == 0 || *SV == 0) && (RM == 0 || *RM == 0)) {
        (void) fputs ("printer ", stdout);
     } else { 
       (void) fputs (" queue  ", stdout);
     }
  }

  (void) fprintf (stdout, "'%s'", Printer);
  (void) printf (" on %s", Host);

  if (CM && *CM) {
     (void) fputs (" (", stdout);
     Print_comment (CM);
     (void) putc (')', stdout);
  }
  (void) putc (':', stdout);
  (void) putc ('\n', stdout);
  (void) fprintf (stdout, "Time            User      Host             Job Pr Files                 Status\n");
(void) fflush (stdout);
  return;
}

/***************************************************************************
 * Get_Daemon()
 * Get the daemon uid and gid
 * Assume password entry has both values set for the user daemon, but
 * get gid from groups file if a gid with the uid's name exists.
 * btw: it doesn't really matter what gid it has, but it looks nice. ;)
 ***************************************************************************/

void 
Get_Daemon (void) {
    struct passwd *passwd;
    struct group *grp;

    if (Daemon_user == NULL) {
	logerr_die (XLOG_INFO, "haven't read config file yet");
    }
    if ((DaemonUID == 0) || (DaemonGID == 0)) {
	if ((passwd = getpwnam (Daemon_user)) == NULL) {
	    logerr_die (XLOG_INFO, "Get_Daemon: getpwnam(%s) failed", Daemon_user);
	}
	if (DaemonUID == 0) {
	    DaemonUID = passwd->pw_uid;
	}
	if (DaemonGID == 0) {
	    if ((grp = getgrnam (Daemon_user))) {
		DaemonGID = grp->gr_gid;
	    } else {
		DaemonGID = passwd->pw_gid;
	    }
	}
    }
    if (Run_as_daemon) {
	/* LPD does this; set user := daemon. We were originally run
	 * as root, but we should run as daemon for safety, except
	 * for the spots where root perms are strictly necessary.
	 */
	ThisIsLPD = True;
	UserUID = DaemonUID;
    } else {
	ThisIsLPD = False;
    }
    return;
}

/***************************************************************************
 * Job_match (char *control, *data)
 * Check to see if the control and data file names are matching
 * 1. data file must start with "df"
 * 2. next must be letter
 * 3. next must be 3 digit sequence number
 * 4. control and data file sequence number must match
 * return 1 if OK, 0 if not
 ***************************************************************************/

int 
Job_match (char *control, char *data) {
    int i, j;		/* ACME Chain and Integers, Inc. */
    char c;

    i = strlen (data);
    for (j = 0; j < i; ++j) {
	c = data[j];

	if (!isascii (c) || !isprint (c)) {
	    log (XLOG_INFO, "Job_match: bad char in '%s'", data);
	    return (False);
	}
	if (c == '/') {
	    log (XLOG_INFO, "Job_match: '/' present in filename \"%s\"", data);
	    return (False);
	}
    }

    if ((i > CFNAMELEN)
	/* @@@ Fix from Dave Atkin */
	|| (i < STARTFR)
	|| (strchr ("tcde", data[0]) == NULL) /* Allow data files beginning 
					       * with "e" too for SGI clients 
					       */
	/* end fix */
	/*     || (i <= STARTFR) 
	       || (strchr ("tcd", data[0]) == NULL) */
	|| (data[1] != 'f')
	/* @@@ Fix from Dave Atkin */
	|| (!isalpha (data[STARTPR]) && !isdigit (data[STARTPR])) /* Allow data files 
								   * df0-df9 for Macs */

	/* end fix */
	/* || (!isalpha (data[STARTPR])) */
	|| (i != strlen (control))
	|| (strncmp (control + STARTID, data + STARTID, 3))
	|| (hostcmp (control + STARTFR, data + STARTFR)))
	{
	    log (XLOG_INFO, "Job_match: bad match control '%s', data '%s'",
		 control, data);
	    return (False);
	}

    for (j = 0; j < IDLEN; ++j) {
	if (!isdigit (data[j + STARTID])) {
	    log (XLOG_INFO, "Job_match: bad sequence control '%s', data '%s'",
		 control, data);
	    return (False);
	}
    }

    return (True);
}

/* get the fully-qualified domain name for a host.
 * usage: strcpy (fqhostname, get_fqdn (hostname));
 * NOTE: this returns a pointer to static data, so copy the result!!
 */

#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 256
#endif

const char 
*get_fqdn (const char *shorthost) {
    static char *returnspc = NULL;	/* dataspace to return ptr to (if needed) */
    struct hostent *host_ent;		/* host entry from data base */
    char shorthost_cpy[SHORTHOSTLEN + 1];
    char **ali, pattern[HOSTLEN + 1];
    int patlen;

    if (strchr (shorthost, '.') != NULL) {
	/* it's already fully-qualified, just return itself */
	return (shorthost);
    }

    /* this is needed in case gethostbyname is destructive */
    assert (strlen(shorthost)<sizeof(shorthost_cpy));
    (void) strcpy (shorthost_cpy, shorthost);

    DomainNameUnset = False;
    DEBUG4 ("Hostname '%s': trying gethostbyname", shorthost);
    host_ent = gethostbyname (shorthost_cpy);

    if (host_ent == 0) {
	DEBUG4 ("Hostname '%s': can't resolve", shorthost);
	if (Host_domain && *Host_domain) {
	    DEBUG4 ("Hostname '%s': appending domain name", shorthost);
	    if (returnspc == NULL) { malloc_or_die (returnspc, HOSTLEN); }
	    (void) sprintf (returnspc, "%s.%s", shorthost, Host_domain);
	    return (returnspc);

	} else {
	    DomainNameUnset = True;
	    return (shorthost);
	}
    }
    
    if (strchr (host_ent->h_name, '.') == NULL) {
	DEBUG4 ("Hostname '%s': h_name is not qualified", shorthost);

	/* we'll assume that if one of the aliases matches /^Shorthost\./,
	 * then it's the fully-qualified domain name. (ugh)
	 * Purify: if we get an ABR here, it's gethostbyname()'s fault, not
	 * ours.
	 */
	(void) strcpy (pattern, shorthost);
	patlen = strlen (pattern);
	pattern[patlen++] = '.';
	pattern[patlen] = '\0';
	if (returnspc == NULL) { malloc_or_die (returnspc, HOSTLEN); }

	*returnspc = '\0';
	if (host_ent->h_aliases) {
	    for (ali = host_ent->h_aliases; *ali; ali++) {
		DEBUG5 ("checking '%s' against pattern '%s' (len=%d)",
			*ali, pattern, patlen);
		if (!strncmp (*ali, pattern, patlen)) {
		    DEBUG4 ("Hostname '%s': h_aliases is qualified", shorthost);
		    return (*ali);
		}
	    }
	}

	if (!*returnspc) {
	    DEBUG4 ("Hostname '%s': h_aliases is not qualified - trying gethostbyaddr",
			shorthost);
	    {
		/* all this copying is ridiculous, but necessary, due to the
		 * resolver's use of a single static hostent structure.
		 */
		char h_first_addr[8];
		int h_length_cpy;
		int h_addrtype_cpy;

		h_addrtype_cpy = host_ent->h_addrtype;
		h_length_cpy = host_ent->h_length;
		h_first_addr[0] = host_ent->h_addr_list[0][0];
		h_first_addr[1] = host_ent->h_addr_list[0][1];
		h_first_addr[2] = host_ent->h_addr_list[0][2];
		h_first_addr[3] = host_ent->h_addr_list[0][3];
		host_ent = gethostbyaddr (h_first_addr, h_length_cpy, h_addrtype_cpy);
	    }
	    
	    if (host_ent && strchr (host_ent->h_name, '.') != NULL) {
		DEBUG4 ("Hostname '%s': ptr lookup is qualified", shorthost);
		return (host_ent->h_name);

	    } else {
		DEBUG4 ("Hostname '%s': ptr lookup is non-qualified", shorthost);
		if (Host_domain && *Host_domain) {
		    DEBUG4 ("Hostname '%s': appending domain name", shorthost);
		    (void) sprintf (returnspc, "%s.%s", shorthost, Host_domain);
		    return (returnspc);
		} else {
		    DomainNameUnset = True;
		    return (shorthost);
		}
	    }
	}
    }
    DEBUG4 ("Hostname '%s': h_name is qualified", shorthost);
    return (host_ent->h_name);
}

/***************************************************************************
 * Get_Local_fqdn()
 * Get the hostname and fully-qualified hostname of the local host.
 * Tricky this; needs to be run after the config file has been read;
 * also, it depends a lot on the local DNS/NIS/hosts-file/etc. setup.
 ***************************************************************************/

#ifndef HAVE_GETHOSTNAME
char *
gethostname (char *nbuf, unsigned long nsiz) {
    struct utsname unamestuff;	/* structure to catch data from uname */

    if (uname (&unamestuff) < 0) {
	return -1;
    }
    (void) strncpy (nbuf, unamestuff.nodename, nsiz - 1);
    return nbuf;
}
#endif

void 
Get_Local_fqdn (void) {
    const char *fqdn;

     /*
      * get the Host computer Name
      */
    if (gethostname (ShortHost, sizeof (ShortHost)) < 0) {
	logerr_die (XLOG_INFO, "Get_Local_fqdn: gethostname failed");
    }
    ShortHost[sizeof (ShortHost) - 1] = '\0';	/* trim to size (just in case) */

    fqdn = get_fqdn (ShortHost);
    if (strlen (fqdn) > sizeof (Host)) {
	log (XLOG_INFO,
	     "fully-qualified host name (%s) is too long, using short version",
	     fqdn);
	/*@@@	(void) strcpy (Host, ShortHost); */
	(void) strncpy (Host, ShortHost, sizeof (Host));
	Host[sizeof(Host) - 1] = '\0';
    } else {
      /*    (void) strcpy (Host, fqdn); */
        (void) strncpy (Host, fqdn, sizeof (Host));
	Host[sizeof(Host) - 1] = '\0';
    }

    /* check in case someone's setup is very sick:
     * ShortHost is fully-qualified and Host isn't.
     */
    if ((strchr (ShortHost, '.') != NULL) && (strchr (Host, '.') == NULL)) {
      /*	(void) strcpy (Host, ShortHost); */
        (void) strncpy (Host, ShortHost, sizeof (Host));
	Host[sizeof(Host) - 1] = '\0';
    }
    
    DEBUG3 ("Hostname: short=%s, fqdn=%s", ShortHost, Host);
    return;
}

/***************************************************************************
 * Std_environ()
 * Make sure that fd 0, 1, 2 exist by opening /dev/null for them if they
 * have not been provided; set up the host information; also, sanitize
 * potentially insecure environment variables.
 ***************************************************************************/

void 
Std_environ (int argc, char **argv, char **envp) {
    char *c, *p;
    int i, j;
#ifdef SETPROCTITLE
    extern char **Argv;
    extern char *LastArgv;
#endif

    /*
     * from sendmail 8.6.4: copy the environment out of the way for use in the
     * secure_system() code (setuid.c).
     */
    for (i = j = 0; j < MAXUSERENVIRON && (p = envp[i]) != NULL; i++) {
	/*
	 * skip EVERYTHING except the timezone; we'll regenerate the essentials
	 * ourselves. Add other exceptions as needed.
	 */
	if (strncmp (p, "TZ=", 3))
	    continue;

	c = strdup (p);
	SafeEnviron[j++] = c;
    }
    SafeEnviron[j] = NULL;

#ifdef SETPROCTITLE
    /* Save start and extent of argv for setproctitle. */
    Argv = argv;
    if (i > 0) {
	LastArgv = envp[i - 1] + strlen (envp[i - 1]);
    } else {
	LastArgv = argv[argc - 1] + strlen (argv[argc - 1]);
    }
#endif				/* SETPROCTITLE */

    setup_clean_fds ();
    umask (077);		/* rwx for user, --- for everyone else */
    return;
}

/***************************************************************************
 * Rename_cf (char *tf)
 * rename the temporary file to a control file
 ***************************************************************************/

void 
Rename_cf (char *tf) {
    char from[MAXPATHLEN + 1];
    char to[MAXPATHLEN + 1];
    int i;

    (void) sprintf (from, "%s/%s", SD, tf);
    (void) sprintf (to, "%s/c%s", SD, tf + 1);
    DEBUG4 ("renaming %s to %s", from, to);

    user_to_daemon ();
    i = rename (from, to);
    daemon_to_user ();

    if (i < 0) {
	logerr_die (XLOG_INFO, "Rename_cf: rename %s to %s failed", from, to);
    }
    return;
}

/***************************************************************************
 * Log_cf (char *tf, char *logstr, controlfile_fd)
 * copy the temporary file to the controlfile logging directory
 * for later examination (debugging aid). Should be run BEFORE
 * Rename_cf(). logstr is appended to the log's filename.
 * If controlfile_fd is >= 0, it's assumed to be an open file descriptor
 * with the file on it. The seek position will be altered.
 ***************************************************************************/

/* copy a file: returns -1 if copy failed */
int 
copy_file (char *from, char *to, int in_fd) {
    int out_fd, n, in_fd_was_passed;
    char buf[BUFSIZ];

    if (in_fd < 0) {
	in_fd = open (from, O_RDONLY, 0);
	in_fd_was_passed = False;
    } else {
	(void) lseek (in_fd, 0L, SEEK_SET);
	in_fd_was_passed = True;
    }
    out_fd = open (to, O_WRONLY | O_CREAT, S_IRUSR | S_IWUSR);
    if (out_fd <= 0) {
	logerr (XLOG_INFO, "copy_file: cannot open file %s.", to);
	return (-1);
    }

    while ((n = read (in_fd, buf, sizeof (buf))) > 0) {
	if (write (out_fd, buf, n) != n) {
	    logerr (XLOG_INFO, "copy_file: cannot write to file %s.", to);
	    return (-1);
	}
    }
    if (n < 0) {
	logerr (XLOG_INFO, "copy_file: error reading from %s", from);
	return (-1);
    }

    if (!in_fd_was_passed)
	(void) close (in_fd);
    (void) close (out_fd);
    return False;
}

void 
Log_cf (char *tf, char *logstr, int ctrl_fd) {
    char from[MAXPATHLEN + 1];
    char to[MAXPATHLEN + 1];

    if (!Ctrlfile_logdir || !*Ctrlfile_logdir) {
	return;			/* control-file logging is disabled */
    }
    (void) sprintf (from, "%s/%s", SD, tf);
    (void) sprintf (to, "%s/c%s.%s", Ctrlfile_logdir, tf + 1, logstr);
    DEBUG4 ("making log copy of %s", to);

    /* ignore the return status; we don't really mind if this fails */
    (void) copy_file (from, to, ctrl_fd);
    return;
}

/* copy a list of h_addr unions. Returns a copy of the arg. */

char 
**copy_haddr_list (char **from) {
    char **i, **j;
    int jindex, arysize;

    arysize = 3; /* make room for this many addresses before resize */
    j = (char **) malloc ((unsigned) sizeof (i) * arysize);
    if (j == NULL) {
	logerr_die (XLOG_NOTICE, Malloc_failed_msg);
    }
    jindex = 0;
    for (i = from; *i; i++) {
	int k;
	chkandrealloc_list (j, jindex, arysize, 3);
	malloc_or_die (j[jindex], sizeof (**i) * IP_ADDRESS_LEN);
	for (k = 0; k < IP_ADDRESS_LEN; k++) {
	  j[jindex][k] = (*i)[k];
	}
	jindex++;
    }
    j[jindex] = NULL;
    return (j);
}

void 
free_haddr_list (char **list) {
    char **del;

    if (list == NULL) {
	return;
    }
    del = list;
    while (*del != NULL) {
	free (*del);
	del++;
    }
    free (list);
    return;
}

/*
 * host comparison function: allows hostnames to be fully-qualified or
 * not, and will work with multi-homed hosts. Also works with localhost.
 */
int 
hostcmp (char *real_s1, char *real_s2) {
    struct hostent *tmp;
    char **to_free, **i1, **i2, *s1, *s2;
    int strcmp_result;

    /* if "localhost" is mentioned, replace it with the name of the
     * (you guessed it) local host for comparison purposes.
     */
    if (strsame (real_s1, "localhost"))
	s1 = ShortHost;
    else
	s1 = real_s1;

    if (strsame (real_s2, "localhost"))
	s2 = ShortHost;
    else
	s2 = real_s2;

    /* try strcmp'ing them, in case they're both fully-qualified,
     * or both non-qualified, or numeric addresses.
     */
    if (*s1 == *s2) {
	strcmp_result = strcasecmp (s1, s2);
	if (!strcmp_result)
	    return False;

    } else {
	strcmp_result = 1;
    }

    /* If either of the names don't resolve, return a non-match;
     * we've already checked to see if the names match using
     * strcmp().
     */
    if ((tmp = gethostbyname (s1)) == NULL)
	return True;

    /* we need to copy this -- it's in tmp storage space;
     * keep a pointer to the start of the copy so we can
     * free it all later.
     */
    to_free = i1 = copy_haddr_list (tmp->h_addr_list);

    if ((tmp = gethostbyname (s2)) == NULL) {
        free_haddr_list (to_free);
	return True;
    }

    i2 = tmp->h_addr_list;	/* no need to copy here */

    for (; *i1; i1++) {
	for (; *i2; i2++) {
	    if (((*i1)[0] == (*i2)[0]) && ((*i1)[1] == (*i2)[1]) &&
		((*i1)[2] == (*i2)[2]) && ((*i1)[3] == (*i2)[3]))
	    {
		free_haddr_list (to_free);
		return False;
	    }
	}
    }
    free_haddr_list (to_free);
    return True;			/* none of the hosts' addresses matched! */
}

/*
 * fopen_path will open the first available file from a list of files, separated by :s.
 * Note that it does NOT do this as the daemon user; it's too slow, and fopen_path()
 * is only ever used to open publicly-accessible config files (such as printcaps,
 * perms files, etc.)
 */
FILE 
*fopen_path (char *path, char *flags) {
    char *tmppath, *step;
    FILE *filep;

    /*
     * we need this to work around a HP/UX feature; (can't free tmppath after using
     * fopen(tmppath, flags) on it)
     */
    static char path_filename[MAXPATHLEN];

    if (path == NULL || !*path) {
	logerr_die (XLOG_INFO, "path is unset!");
    }
    DEBUG4 ("path to open: %s", path);
    filep = NULL;

    tmppath = strdup (path);
    /* we can now rip tmppath to shreds with impunity */

    step = (char *) strtok (tmppath, ":");
    do {
	(void) strcpy (path_filename, step);

	if ((filep = fopen (path_filename, flags)) != NULL) {
	    DEBUG4 ("opened %s", step);
	    free (tmppath);
	    return (filep);
	}
    } while ((step = (char *) strtok (NULL, ":")));

    DEBUG4 ("no files could be opened.");
    free (tmppath);
    return (NULL);
}

/***************************************************************************
 * char *reallocstrcpy (to, from, fromsize)
 * strcpy with dynamic allocation. The "to" buffer is returned.
 * if you know the size of from's buffer, pass it in as the fourth arg,
 * otherwise pass in 0 (the strlen will be used).
 * use as follows: to = reallocstrcpy(to, "hello", 0);
 **************************************************************************/

char 
*reallocstrcpy (char *to, char *from, int fromspace) {
    int tospace;

    if (fromspace <= 0) {	/* is it set? */
	fromspace = strlen (from);
    }
    fromspace++;		/* copy the trailing null */

    if (!to) {
	tospace = fromspace;
	malloc_or_die (to, fromspace);
	*to = '\0';
    } else {
	tospace = strlen (to);
    }

    /* if insufficient room, dynamically allocate buffer */
    if (tospace < fromspace) {
	DEBUG8 ("reallocing buffer to %d chars", fromspace);
	to = (char *) realloc (to, (unsigned) fromspace);
	if (to == NULL) {
	    logerr_die (XLOG_NOTICE, Malloc_failed_msg);
	}
    }
    (void) bcopy (from, to, fromspace);
    return (to);
}

/**************************************************************************
 * open /dev/null on all non-open std fds (stdin, stdout, stderr).
 *************************************************************************/

static void
setup_std_fds (void) {
    int fd;
    int i;
    while ((fd = open ("/dev/null", O_RDWR, 0)) >= 0) {
	if (fd >= 3) {
	    break;	/* just 0, 1 and 2. */
	}
	DEBUG9 ("setup_std_fds: opened /dev/null on fd %d", fd);
    }
    /* let's have a try; if a lot of files are to be printed, sometimes
     * there's an error "too many open files"; we try to avoid this sleeping 
     * for a while */
    for (i = 1; i < 5; i++) {
	if (fd < 0) {
	    if (i < 4) {
		DEBUG2 ("setup_std_fds: no fd available, sleeping %d secs", i);
		sleep (i);
		continue;
	    } else {
		logerr_die (XLOG_CRIT, "setup_std_fds: cannot open /dev/null");
		/* well, this is pretty serious */
	    }
	} else {
	    (void) close (fd);
	    break;
	}
    }
    return;
}

/*
 * close all non-std file descriptors up to getdtablesize (), and
 * call setup_std_fds().
 */

void 
setup_clean_fds (void) {
    int fd;

    for (fd = 3; fd < getdtablesize (); fd++) {
	(void) close (fd);
    }
    setup_std_fds ();
    return;
}

/*
 * setup all non-std file descriptors up to getdtablesize ()
 * so they'll close before the next exec () call, and call
 * setup_std_fds ().
 * We can't simply use setup_clean_fds () because that breaks stuff
 * on IRIX 5.1.1 (connection to NIS? getpwuid() fails, anyway).
 */

void 
setup_close_on_exec (void) {
    int fd;

#ifdef HAVE_FCNTL
    for (fd = 3; fd < getdtablesize (); fd++) {
	int val;
	/* if this fails, the fd is probably not open, so skip it... */
	if ((val = fcntl (fd, F_GETFD, 0)) >= 0) {
#ifdef FD_CLOEXEC
	    val |= FD_CLOEXEC;	/* Advanced UNIX Programming */
#else
	    val |= 1;		/* SVR4 and others */
#endif
	    if (fcntl (fd, F_SETFD, val) == -1) {
		if (errno != EBADF) {
		    logerr (XLOG_INFO, "could not set close-on-exec on fd %d");
		}
	    }
	}
    }
    setup_std_fds ();

#else
    /* if we don't have fcntl(), just use setup_clean_fds (). */
    setup_clean_fds ();
#endif
    return;
}

/* provide us with an fd map */
void 
debug_fds (void) {
    int fd;

    for (fd = 0; fd < getdtablesize (); fd++) {
	if (fcntl (fd, F_GETFL, 0) < 0) {
	    log (XLOG_DEBUG, "debug_fds: fd %d is closed", fd);
	} else {
	    log (XLOG_DEBUG, "debug_fds: fd %d is open", fd);
	}
    }
    return;
}

/* Check to see if the nonqualified hostname "host" is valid
 * when the domain "domain" is appended to it. Returns the
 * "host.domain" string if it is, NULL otherwise.
 * We cache the last set of values, as Checkperm() calls
 * us once for each line of the perms file.
 */

char 
*host_in_domain (char *host, char *domain) {
    int hlen;
    char *fq_hostname;
    static char *last_fq_host = NULL, *last_result = NULL;

    assert (host != NULL && domain != NULL);

    hlen = strlen (host);

    if (*domain == '.') {
	domain++;		/* skip a leading . if it's there */
    }

    /* check to see if the domain name is already there */
    if (strstr (host, domain) != NULL) {
	if (last_fq_host) {
	    free (last_fq_host);
	}
	last_fq_host = strdup (host);
	return (last_result = last_fq_host);
    }

    malloc_or_die (fq_hostname, hlen + strlen (domain) + 2);
    (void) strcpy (fq_hostname, host);
    fq_hostname [hlen] = '.';
    (void) strcpy (fq_hostname + hlen + 1, domain);

    if (last_fq_host) {
	if (strsame (last_fq_host, fq_hostname)) {
	    return (last_result);
	} else {
	    free (last_fq_host);
	    last_fq_host = strdup (fq_hostname);
	}
    } else {
	last_fq_host = strdup (fq_hostname);
    }

    /* DON'T use res_options here! PLP will usually be used in an
     * internal setting as opposed to accross the internet, and
     * relative domains, such as eg. "tst.com.iona.ie", are actually
     * more likely than domains such as "tst.com".
     */
    if (gethostbyname (fq_hostname) != NULL) {
	DEBUG2 ("host_in_domain: host %s exists in domain %s", host, domain);
	return (last_result = fq_hostname);

    } else {
	DEBUG2 ("host_in_domain: host %s does not exist in domain %s", host, domain);
	free (fq_hostname);
	return (last_result = NULL);
    }
}

inline void
lower_case (char *s) {
    if (s) {
	while (*s) {
	    if (isupper (*s))
		*s = tolower (*s);
	    s++;
	}
    }
    return;
}

void 
empty_cfparms (void) {
    int i;
    for (i = 0; i < 26; ++i) {
	CFparm[i][0] = 0;
    }
    return;
}

/**************************************************************************
 * int
 * update_history (void)
 * makes an update of the history file, depending on the 
 * variable "interval" 
 * returns 
 *      2 if any I/O error occured; 
 *      1 if there is no history file; 
 *      0 when successful
 **************************************************************************/
int 
update_history (void) {
    time_t now, history_time;
    char buf[255];
    char history_file[255];
    char *tmpf;			/* tmp file name */
    FILE *tmp_fp, *history_fp;
    int saved, discarded;

    saved = discarded = 0;
    /* Get the current time */
    now = (unsigned long) time ((time_t *) 0);
    /* subtract the interval */
    now = now - (interval * 3600);
    /* create a temporary file */
    tmpf = tmpnam (NULL);
    /* construct the history filename */
    (void) sprintf (history_file, "%s/.history", SD);

    DEBUG3 ("updating history file %s", history_file);

    if ((tmp_fp = fopen_daemon (tmpf, "w")) == NULL) {
	logerr_die (XLOG_INFO, "cannot open %s", tmpf);
	return 2;
    }
    if ((history_fp = fopen (history_file, "r")) == NULL) {
	(void) fclose (tmp_fp);
	(void) unlink (tmpf);
	return 1;
    }
    while (fscanf (history_fp, "%ld", &history_time) != EOF) {
	(void) fgets (buf, 128, history_fp);
	if (now >= history_time) { /* obsolete entry, discard it */
	    discarded++;
	    continue;
	} else { /* save the content in the tmp-file */
	    (void) fprintf (tmp_fp, "%lu", history_time);
	    (void) fputs (buf, tmp_fp);
	    (void) fflush (tmp_fp);
	    saved++;
	}
    }
    (void) fclose (history_fp);
    (void) fclose (tmp_fp);
    (void) unlink (history_file);
    if (copy_file (tmpf, history_file, -1) < 0) {
	(void) fprintf (stdout, "lpc: update_history: Warning: error during rename history for printer %s.\n", Printer);
	(void) fprintf (stdout, "reason: %s\n", Errormsg (errno));
	(void) fflush (stdout);
	(void) unlink (tmpf);
	return 2;
    }
    (void) unlink (tmpf);
    (void) chmod (history_file, S_IRUSR | S_IWUSR);
    (void) chown (history_file, DaemonUID, DaemonGID);
    DEBUG3 ("update_history: history-file changed in Uid %d, Gid %d.", DaemonUID, DaemonGID);
    DEBUG3 ("update_history: %d history records discarded, %d saved", discarded, saved);
    return 0;
}


/**************************************************************************
 * int
 * create_history_entry (error, q)
 * creates an history entry for the printed job(s)
 * if an error occurs during creation/update of the history file, errno will be set 
 * the different types of errors are in lpq.c, line 653ff
 * returns 
 *      1 if there is an error;
 *      0 when successful
 **************************************************************************/



int 
create_history_entry (int error, struct plp_queue *q) {
    FILE *history_fp;
    char history_file[255];
    char priority[2];
    char hostname[128];
    char jobname[64];
    unsigned long secs = 0;

    if (RP) {
	/* get the time when job was spooled */
	secs = spooled_time;
    } else {
	/* get the time when job was printed */
	secs = printed_time;
    }
    if (secs == 0) {
	secs = time ((time_t *) NULL);
    }
    /* construct the filename */
    (void) sprintf (history_file, "%s/.history", SD);
    
    /* open the file and write the entry */
    if ((history_fp = fopen (history_file, "a")) == NULL) {
	Warnmsg ("create_history_entry: could not create/update %s - %s", history_file, Errormsg (errno));
	return (1);
   } else {
	(void) chmod (history_file, S_IRUSR | S_IWUSR);
	(void) chown (history_file, DaemonUID, DaemonGID);
	DEBUG3 ("create_history_entry: history-file changed in Uid %d, Gid %d.", DaemonUID, DaemonGID);
	priority[0] = q->q_name[2];
	priority[1] = '\0';
	DEBUG3 ("create_history_entry: generating new history record for job %s", q->q_name);
	(void) strcpy (hostname, &q->q_name[6]);
	(void) strncpy (jobname, q->q_data->name, 63);
	(void) fprintf (history_fp, "%lu %s %s %d %s %s   %d\n", 
			secs, q->q_user,
			hostname, q->q_num,
			priority, jobname, error);
	(void) fclose (history_fp);
    }
    return (0);
}
