/*
 * printers: print out contents of printcap files
 * 
 * printers [-l] [-L] [-e] [printer[@domain]] -l - long listing - some information -L - long
 * listing - all information -e - use exact match
 */

#include "lp.h"
#include "library/errormsg.h"
#include "library/readconf.h"
#include "library/utils.h"

struct sort_pr {
  char pr_name[24];
};

static int longl, verylong, exact;
extern int printers_compare ();

static void
usage (void) {
    (void) printf ("Usage:\tprinters [-e] [-l] [-L] [printer[@domain]]\n\n");
    (void) printf ("where\n");
    (void) printf ("\t-e causes the printer name to be matched exactly\n");
    (void) printf ("\t-l gives some information about the printer, e.g. alias names\n");
    (void) printf ("\t-L gives full information about the printer\n");
    (void) printf ("\t-? or -h print out this usage message\n\n");
    (void) printf ("If no printer is specified then all the printers in the domain will be listed.\n");
    (void) printf ("If the printer name is appended with @domain the printer is looked up in that\n");
    (void) printf ("domain using HESIOD or NIS. If only the @domain part is specified, all of the\n");
    (void) printf ("printers in that domain are listed.\n");
    exit (1);
}

/*
 * catpent: get the printcap entry end print it
 */
static void
catpent (char *name) {
    char *cp;
    static char names[1024];

    if (Get_pc_entry (name, All_pc_vars, All_pc_len) == 0) {
	(void) printf ("no printer %s\n", name);
	return;
    }

    cp = strrchr (All_names, '|');

    /* skip printers with a final alias beginning with TEST */
    if (cp != NULL && !strncmp (cp + 1, "TEST", 4) && !(longl || verylong))
	return;

    if (cp == NULL) {
	(void) strcpy (names, name);
    } else {
	(void) sprintf (names, "%s (%s)", name, cp + 1);
    }

    printf ("%-30s ", names);
    {
	char *s = get_pr_commnt ();
	if (strstr (names, s) == NULL) {
	    printf ("%s ", s);	/* only if we haven't already said it */
	}
    }

    if (longl || verylong) {
	(void) printf ("\n");
	if (cp != NULL)
	    *cp = '\0';
	if ((cp = strchr (All_names, '|')) != NULL) {
	    do {
		*cp = ',';
	    } while ((cp = strchr (All_names, '|')) != NULL);
	    (void) printf ("\taliases: %s\n", All_names);
	}
	(void) printf ("\tspool directory: %s\n", SD);
	if (RM && *RM)
	    (void) printf ("\tremote spooling host: %s\n", RM);
	if (RP && *RP)
	    (void) printf ("\tremote printer name: %s\n", RP);
	if (FL && *FL)
	    (void) printf ("\tvalid forms: %s\n", FL);
	if (MX > 0)
	    (void) printf ("\tfile size limit: %ld KBytes\n", (long) MX);

	if (verylong) {
	    if (LP && *LP)
		(void) printf ("\tdevice: %s\n", LP);
	    if (LF && *LF)
		(void) printf ("\tlog file: %s\n", LF);
	    if (LO && *LO)
		(void) printf ("\tlock file: %s\n", LO);
	    if (AF == 0 || *AF == 0)
		(void) printf ("\tno accounting file\n");
	    else
		(void) printf ("\taccounting file: %s\n", AF);
	    if (ST && *ST)
		(void) printf ("\tstatus file: %s\n", ST);
	    if (PS && *PS)
		(void) printf ("\tprinter status file: %s\n", PS);
	    if (EU)
		(void) printf ("\tEUCS printer spooler used\n");
	    else if (BK)
		(void) printf ("\tBSD printer spooler used\n");
	    else
		(void) printf ("\tPLP printer spooler used\n");
	    if (ED)
		(void) printf ("\texpensive device\n");
	}
    }
    (void) printf ("\n");
}

int
main (int argc, char **argv, char **envp) {
    char **prlist;
    int i, option;
    int number_of_printers = 0;
    struct sort_pr *sort_pr_array;

    longl = verylong = exact = 0;

    while ((option = Getopt (argc, argv, "D:lLe?h")) != EOF) {
	switch (option) {
	case 'D':		/* Debugging */
	    Parse_debug (Optarg);
	    break;
	case 'l':		/* long listing */
	    ++longl;
	    break;
	case 'L':		/* very long listing */
	    ++verylong;
	    break;
	case 'e':		/* exact match of printer name */
	    ++exact;
	    break;
	case '?':
	case 'h':
	default:
	    usage ();
	    break;
	}
    }
    if (Optind < argc) {
	Printer = argv[Optind];
	++Optind;
    }
    if (Optind < argc) {
	usage ();
    }
    Std_environ (argc, argv, envp);
    Readconf ();

#ifdef HESIOD
    if (Printer && *Printer == '@') {
	for (prlist = All_printers_zone (Printer + 1);
	     (prlist && *prlist); ++prlist) {
	    char buf[BUFSIZ];
	    strcpy (buf, *prlist);
	    if (out_domain) {
		strcat (buf, "@");
		strcat (buf, out_domain);
	    }
	    catpent (buf);
	}
    } else
#endif
    if (Printer && (exact || strchr (Printer, '@')))
	catpent (Printer);
    else {
	for (prlist = All_printers (); (prlist && *prlist); ++prlist) {
	  number_of_printers++;
	}

	/* "rewinding" the pointer */
	for (i = 0; i < number_of_printers; i++) {
	     prlist--;
	}
	sort_pr_array = (struct sort_pr *) malloc ((sizeof (struct sort_pr) * 
						    sizeof (char)) * 
						   number_of_printers);
        if (sort_pr_array == NULL) {
            logerr_die (XLOG_INFO, "printers: Malloc for sort_pr_array failed");
        }

	/* copy the name of the printer in the array sort_pr_array */
	for (i = 0; i < number_of_printers; i++) {
	     strcpy (sort_pr_array[i].pr_name, *prlist);
	     prlist++;
	}

	/* sort the names of the printers */
	qsort ((struct sort_pr *) sort_pr_array, (size_t) number_of_printers, 
	       (size_t) sizeof (struct sort_pr), printers_compare);

	for (i = 0; i < number_of_printers; i++) {
	    if (Printer) {
		if (strstr (sort_pr_array[i].pr_name, Printer))
		    catpent (sort_pr_array[i].pr_name);
	    } else {
		catpent (sort_pr_array[i].pr_name);
	    }
	}
    }
    return False;
}

plp_signal_t 
cleanup (int signal) 
{ 
  exit (1); 
}

int
printers_compare (struct sort_pr * l, struct sort_pr * r)
{
  return (strcmp (l->pr_name, r->pr_name));
}
 
